// vincia26.cc is a part of the VINCIA plugin to PYTHIA 8
// This is test program to illustrate the merging procedure in Vincia
// + Pythia8.
// It uses the Les Houches event files PPto{Z,ZJ,ZJJ,ZJJJ}-5k.lhe.gz.
// Copyright (C) 2018 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
#include "Pythia8Plugins/HepMC2.h"

// Use PYTHIA and VINCIA namespaces
using namespace Pythia8;
using namespace Vincia;

// Main Program.
int main( int argc, char* argv[] ) {

  // Check if command file and HepMC file name is provided.
  if (argc <= 2) {
    cout << endl
         << "Use the program like this: ./vincia26 cmndfile HepMCfilename"
         << endl << "Available cmndfiles are vincia26.cmnd and "
         << "vincia26gzip.cmnd (requires gzip)."
         << endl << "Please specify HepMCfilename (e.g. vincia26.hepmc)."
         << endl << endl
         << "For passing events directly to Rivet with a fifo pipe do: " << endl
         << "   mkfifo myfifo" << endl << "   ./vincia26 cmndfile myfifo &"
         << endl << "   rivet --analysis=ANALYSIS_NAME myfifo" << endl << endl;
    return 0;
  }

  // Check in unzipped versions of LHEF files should be used.
  if (string(argv[1]).find("gzip") == string::npos) {
    for (int i=0; i<=3; i++) {
      string fileName = "PPtoZ";
      for (int j=0; j<i; j++) fileName += "J";
      fileName += "-5k.lhe";
      if ( access(fileName.c_str(), F_OK) == -1 ) {
        cout << "gunzip -c "+fileName+".gz > "+fileName << endl;
        // Dummy to prevent compiler warnings;
        if (system(("gunzip -c "+fileName+".gz > "+fileName).c_str()) == 0) {}
      }
    }
  }

  // NOTE Due to Vincia's init call, no parameters that are overwritten in
  // in the main file can be in the command file that is read it!
  // Therefore, we define these parameters here.
  // DON'T PUT Merging:nRequested or Merging:doXSectionEstimate IN COMMAND FILE!
  // Showering, hadron level, and MPI.
  bool fsr = true;
  bool isr = true;
  bool mpi = false;
  bool had = true;
  bool rem = true;
  bool che = true;
  // Merging cut.
  string tms = "0.01";

  // Generator.
  Pythia pythia;
  VinciaPlugin vincia(&pythia);

  // Settings and Info pointer.
  Settings& settings = pythia.settings;
  Info& info         = pythia.info;

  // Set showering, hadron level, MPI, and TMS.
  if (fsr) pythia.readString("PartonLevel:FSR      = on");
  else     pythia.readString("PartonLevel:FSR      = off");
  if (isr) pythia.readString("PartonLevel:ISR      = on");
  else     pythia.readString("PartonLevel:ISR      = off");
  if (mpi) pythia.readString("PartonLevel:MPI      = on");
  else     pythia.readString("PartonLevel:MPI      = off");
  if (had) pythia.readString("HadronLevel:All      = on");
  else     pythia.readString("HadronLevel:All      = off");
  if (rem) pythia.readString("PartonLevel:Remnants = on");
  else     pythia.readString("PartonLevel:Remnants = off");
  if (che) pythia.readString("Check:event          = on");
  else     pythia.readString("Check:event          = off");
  pythia.readString("Merging:TMS = " + tms);

  // First initialization.
  vincia.init(argv[1], 0);

  // Check for restrictions on what to include in the merging.
  bool nonShowerPathsOnly = settings.flag("Merging:nonShowerPathsOnly");

  // Maximum number of additional jets.
  int nMerge = settings.mode("Merging:nJetMax");

  // HepMC Interface.
  HepMC::Pythia8ToHepMC ToHepMC;
  // Switch off warnings for parton-level events.
  ToHepMC.set_print_inconsistency(false);
  ToHepMC.set_free_parton_exception(false);
  // Do not store cross section information, as this will be done manually.
  ToHepMC.set_store_pdf(false);
  ToHepMC.set_store_proc(false);
  ToHepMC.set_store_xsec(false);
  // Specify file where HepMC events will be stored.
  HepMC::IO_GenEvent HepMCevent(argv[2], std::ios::out);

  // Number of events. Negative numbers mean all events in the LHEF 
  // will be used.
  long nEvent = settings.mode("Main:numberOfEvents");
  if (nEvent < 1) nEvent = 1000000000000000;

  // Remember TMS value.
  //string tms = static_cast<ostringstream*>( &(ostringstream() << 
  //  settings.parm("Merging:TMS")) )->str();

  //----------------------------------------------------------------------------
  // Cross section estimation

  cout << endl << endl << endl;
  cout << "Start estimating ckkwl tree level cross section" << endl;

  // Switch on cross section estimation procedure.
  settings.flag("Merging:doXSectionEstimate", true);

  // Switch off all showering, hadron level, and MPI when estimating the cross
  // section after the merging scale cut.
  pythia.readString("PartonLevel:FSR      = off");
  pythia.readString("PartonLevel:ISR      = off");
  pythia.readString("PartonLevel:MPI      = off");
  pythia.readString("HadronLevel:All      = off");
  pythia.readString("PartonLevel:Remnants = off");
  pythia.readString("Check:event          = off");

  // Save estimates in vectors.
  vector<double> xsec;
  vector<double> nAccept;
  vector<double> nSelected;
  vector<int>    strategy;
  xsec.resize(1+nMerge);
  nAccept.resize(1+nMerge);
  nSelected.resize(1+nMerge);
  strategy.resize(1+nMerge);

  // Loop over subruns with varying number of jets.
  for (int iMerge = 0; iMerge <= nMerge; ++iMerge) {
    settings.mode("Merging:nRequested", iMerge);
    // For including only those events that can not be reproduced by the shower
    // in a sample with lower multiplicity, set TMS cut to 0 for cross
    // section estimation.
    if (nonShowerPathsOnly) pythia.readString("Merging:TMS = 0.0");
    // Read in name of LHE file for current subrun and initialize.
    vincia.init(argv[1], iMerge);
    // Loop over Les Houches input events.
    for (int iEvent = 1; iEvent <= nEvent; ++iEvent) {
      if ( !pythia.next() ) {
        if ( info.atEndOfFile() ) break;
        else continue;
      }
    }
    pythia.stat();
    xsec[iMerge]      = info.sigmaGen();
    nSelected[iMerge] = info.nSelected();
    nAccept[iMerge]   = info.nAccepted();
    strategy[iMerge]  = info.lhaStrategy();
  }

  //----------------------------------------------------------------------------
  // Generation of events

  cout << endl << endl << endl;
  cout << "Start generating events" << endl;

  // Switch off cross section estimation.
  settings.flag("Merging:doXSectionEstimate", false);

  // Switch showering, hadron level, and MPI back on.
  if (fsr) pythia.readString("PartonLevel:FSR      = on");
  else     pythia.readString("PartonLevel:FSR      = off");
  if (isr) pythia.readString("PartonLevel:ISR      = on");
  else     pythia.readString("PartonLevel:ISR      = off");
  if (mpi) pythia.readString("PartonLevel:MPI      = on");
  else     pythia.readString("PartonLevel:MPI      = off");
  if (had) pythia.readString("HadronLevel:All      = on");
  else     pythia.readString("HadronLevel:All      = off");
  if (rem) pythia.readString("PartonLevel:Remnants = on");
  else     pythia.readString("PartonLevel:Remnants = off");
  if (che) pythia.readString("Check:event          = on");
  else     pythia.readString("Check:event          = off");

  // Merged total cross section, summed over subruns.
  double sigmaTotal = 0.0;
  double errorTotal = 0.0;
  vector<double> sigmaSamples;
  vector<double> errorSamples;
  sigmaSamples.resize(1+nMerge);
  errorSamples.resize(1+nMerge);
  bool includeWeightInXSEC = settings.flag("Merging:includeWeightInXsection");

  // Allow abort of run if many errors.
  int  nAbort  = settings.mode("Main:timesAllowErrors");
  int  iAbort  = 0;
  bool doAbort = false;

  // Loop over subruns with varying number of jets.
  for (int iMerge = 0; iMerge <= nMerge; ++iMerge) {

    settings.mode("Merging:nRequested", iMerge);

    sigmaSamples[iMerge] = 0.0;
    errorSamples[iMerge] = 0.0;

    // For including only those events that can not be reproduced by the shower
    // in a sample with lower multiplicity, set TMS cut to infinity or TMS for
    // generation of events.
    if (nonShowerPathsOnly) {
      // Disable TMS check on input events.
      settings.flag("Merging:enforceCutOnLHE", false);
      // For the Born and Born+1 sample set TMS to infinity.
      if (iMerge <= 1) pythia.readString("Merging:TMS = 1e+10");
      // For the Born+>=2 samples set TMS to actual value to avoid double
      // counting via merging or set to infinity to void double counting
      // by hand.
      //else pythia.readString("Merging:TMS = "+tms);
      else pythia.readString("Merging:TMS = 1e+10");
    } else pythia.readString("Merging:TMS = "+tms);
    // Read in name of LHE file for current subrun and initialize.
    vincia.init(argv[1], iMerge);

    // Begin event loop.
    for (int iEvent = 1; iEvent <= nEvent; ++iEvent) {

      // Generate next event. Break out of event loop if at end of LHE file.
      if ( !pythia.next() ) {
        if ( info.atEndOfFile() ) break;
        else if (++iAbort > nAbort) { doAbort = true; break; }
        else continue;
      }

      // Get event weight(s).
      double weight = info.weight();
      // Apply merging weight if needed.
      if (!includeWeightInXSEC) weight *= info.mergingWeight();
      // Vincia's weight, includes info.weight()
      // No need to use, as reweighting doesn't work together with merging
      // anyway
      double vinWei = 1.0;

      // Do not use zero-weight events
      if (weight == 0.0) continue;

      // Get correct cross section in mb from previous estimate.
      double normhepmc = ( (abs(strategy[iMerge]) == 4) ?
                           1.0/(1e9*nSelected[iMerge]) :
                           xsec[iMerge]/nAccept[iMerge] );

      // Add the weight of the current event to the cross section.
      sigmaTotal           += weight*normhepmc;
      sigmaSamples[iMerge] += weight*normhepmc;
      errorTotal           += pow2(weight*normhepmc);
      errorSamples[iMerge] += pow2(weight*normhepmc);

      // Fill HepMC file.
      HepMC::GenEvent* HepMCeventNow = new HepMC::GenEvent();
      // Set event weight.
      HepMCeventNow->weights().push_back( vinWei*weight*normhepmc );
      // Fill HepMC event.
      ToHepMC.fill_next_event( pythia.event, HepMCeventNow );
      // Set cross section (convert mb -> pb).
      HepMC::GenCrossSection XS;
      XS.set_cross_section( sigmaTotal*1e9, info.sigmaErr()*1e9 );
      HepMCeventNow->set_cross_section( XS );
      HepMCevent << HepMCeventNow;
      delete HepMCeventNow;

    } // End of event loop.

    if (doAbort) break;

    // Print infos.
    cout << endl << scientific << setprecision(8)
         << " Contribution of sample " << iMerge
         << " to the inclusive cross section: "
         << sigmaSamples[iMerge] << "  +-  " << sqrt(errorSamples[iMerge])
         << " mb " << endl << endl;
    cout << setprecision(3) << fixed;
    vincia.printInfo();
    pythia.stat();

  }

  cout << endl << endl << endl;
  if (doAbort)
    cout << " Run was not completed owing to too many aborted events" << endl;
  else
    cout << scientific << setprecision(8) << "Inclusive cross section: "
         << sigmaTotal << "  +-  " << sqrt(errorTotal) << " mb " << endl;
  cout << endl << endl << endl;

  // Done.
  return 0;
}
