// vincia32-root.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a test program for matrix element comparisons
// Copyright (C) 2018 Peter Skands
// Main author of this file: J.J. Lopez-Villarejo

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
#include "Vincia/MG4interface.h"
// ROOT include files, uncomment as necessary
#include "TROOT.h"
#include "TDirectory.h"
#include "TFile.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"

// VinciaRoot seems to be incompatible with offline plots with the macro
//#include "VinciaRoot.h"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Optionally specify name of directory where root files will be stored
static const string rootdir="./";


//**********************************************************************
// showerApprox
//**********************************************************************

//************************************************************************

//INPUT:
//a) pointers to some data
//TO DO: structure the function into a class or something equivalent so that
// these pointers would not be needed. 
//
//b) configuration to cluster: identities, momenta, polarizations
//c) Born configuration:  identities, momenta, polarizations
//d) Matching orders: 0 if we want to cluster all the way until Born; 
//   a positive int for the number of levels before the Born where we want 
//   to stop clustering.
//e) flags: 
//          usePari (ARIADNE factor), 
//          usePimp (smooth-ordering factor)
//f) Special: pimpQE2PreviousConf is only used when the function calls itself
//             in order to compute the Pimp factor.

double showerApprox(AntennaSetFSR* antSetPtr, 
                    int verbose, double eCM, Resolution* resolutionPtr, 
                    MG4interface* mgInterfacePtr, VinClu* vinCluPtr, 
                    int nFSpartons, vector<int> idConf, 
                    vector<Vec4> pConf, vector<int> hConf, 
                    vector<int> idBorn, vector<Vec4> pBorn,
                    vector<int> hBorn, int nMatchingOrders, bool usePari, 
                    bool usePimp, double pimpQE2PreviousConf=0.){
  
  // 0) define relevant quantities
  int nClusters = nFSpartons - 2;
  int nPartons  = idConf.size();
  int nISPartons = nPartons - nFSpartons;
  double approx = 0.0;	
  
  // 0) (For pimp) Loop over clusters to find minimum Q_E (Q_E^hat) and compute
  //    pimp factor with the Q_E given as an argument of the function.
  
  double pimp=1.;
  if (usePimp){
    double QE2Min = eCM*eCM;        
    //Tell the minimum of each ordering variable of parent configuration clusterings
    for (int partonCluIndx = nISPartons + 1 ; partonCluIndx <= nPartons - 2 ;
         ++partonCluIndx) {                 
      // Q_E is PT for gluons and Mqq for quarks
      double QE2partonClu = 0.;
      if (idConf[partonCluIndx] == 21 ) {
        QE2partonClu = resolutionPtr->Q2E(pConf[partonCluIndx-1],
                                 pConf[partonCluIndx],
                                 pConf[partonCluIndx+1],1);            
      } else if (idConf[partonCluIndx]==-1 || idConf[partonCluIndx]==-2) {
        QE2partonClu = m2(pConf[partonCluIndx],pConf[partonCluIndx+1]);
      } else if (idConf[partonCluIndx]==1 || idConf[partonCluIndx]==2) {
        QE2partonClu = m2(pConf[partonCluIndx-1],pConf[partonCluIndx]);
      }
      QE2Min = min(QE2Min,QE2partonClu);          
    }        
    pimp = QE2Min/(QE2Min+pimpQE2PreviousConf);
  }
  
  
  // 1) (For matching orders)
  if (nFSpartons == 2 + nMatchingOrders) {
    approx = mgInterfacePtr->ME2(1,idConf,pConf,hConf);
    if (usePimp) approx *= pimp;
    return approx;
  }
  
  // 2) if the number of clusters is just one the procedure is special
  //    (check that it matches with the Born)  
  if (nClusters ==1) {
    double ME2Born = mgInterfacePtr->ME2(1,idBorn,pBorn,hBorn);
    // 2.1) Cluster and compute invariants
    int partonCluIndx = (nPartons-1) -1 ;
    vector<int> idCluConf;
    vector<Vec4> pCluConf;
    vector< vector<int> > hCluConfs;
    vector<double> hConf_d;
    for (unsigned int i_d = 0; i_d < hConf.size(); i_d++) 
      hConf_d.push_back((double)hConf[i_d]);
    int antIndx;
    bool swapAB = false;
    bool pointOK = false;
    pointOK = vinCluPtr->getCluPol(idConf,pConf,
                                   partonCluIndx-1,partonCluIndx,partonCluIndx+1,
                                   idCluConf,pCluConf,antIndx,swapAB,hConf_d,hCluConfs);
    if ( !pointOK ) cout<<"Warning: problem with phase-space point"<<endl;
    // Antenna Invariants
    double sAnt = m2(pConf[partonCluIndx-1]+pConf[partonCluIndx]+pConf[partonCluIndx+1]);
    double yar = (!swapAB) ? m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) ;
    double yrb = (!swapAB) ? m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) ;
    yar /= sAnt;
    yrb /= sAnt;
    
    // 2.2) Check that the identities of the clustered configuration partons
    //      match with the Born
    // TO DO: momenta check?
    if (idCluConf != idBorn) return 0.;
    // 2.3) Loop over the different polarization configurations for same
    //      clustering (only take the configuration that matches with the
    //      Born)
    for (unsigned int hConfsIndx = 0; hConfsIndx <= hCluConfs.size()-1;
         ++hConfsIndx) {
      // 2.3.1) Check that the clustered configuration reproduces the Born
      if (hCluConfs[hConfsIndx] != hBorn) continue;          
      // 2.3.2) compute antenna ant_i_hConf          
      //virtual double antennaFunction(double yij, double yjk, double yik, 
      //                               int hA, int hB, 
      //                               int hi, int hj, int hk);      
      double ant = 0.;
      //Polarizations (special treatment when legs are reversed)
      double hA = (!swapAB) ? hCluConfs[hConfsIndx][partonCluIndx-1] 
      : hCluConfs[hConfsIndx][partonCluIndx] ;
      double hB = (!swapAB) ? hCluConfs[hConfsIndx][partonCluIndx] 
      : hCluConfs[hConfsIndx][partonCluIndx-1] ;
      double hi = (!swapAB) ? hConf[partonCluIndx-1] : hConf[partonCluIndx+1];
      double hj = hConf[partonCluIndx];
      double hk = (!swapAB) ? hConf[partonCluIndx+1] : hConf[partonCluIndx-1];
      
      ant = antSetPtr->getAnt(antIndx)->
        antennaFunction(yar,yrb,0.,0.,0.,
                        hA,hB,hi,hj,hk)/sAnt;              
      // Charge factor
      double cFac = antSetPtr->getAnt(antIndx)->getChargeFac();
      ant = cFac * ant; 
      
      // 2.3.3) add to the approximation the single element that matches
      //        the Born configuration
      //        ant_i_hConf * (ariadne)
      approx = ant * ME2Born; 
      if (usePimp) {
        approx = approx * pimp;
      }
    } // end of loop over polarization configurations
    
    // 2.4) Return the approximation        
    return approx;
    
  }	

  // 3) Placeholder for sector identification.
  
  // 4) Main Loop over clusters 
  //    Sum_i (ant_i * |M_i|^2 )
  for (int partonCluIndx = nISPartons + 1 ; partonCluIndx <= nPartons -2 ;
       ++partonCluIndx) {
    // 4.1) Cluster and compute invariants
    vector<int> idCluConf;
    vector<Vec4> pCluConf;
    vector< vector<int> > hCluConfs;
    vector<double> hConf_d;
    for (unsigned int i_d = 0; i_d < hConf.size(); i_d++) 
      hConf_d.push_back((double)hConf[i_d]);
    int antIndx;
    bool swapAB = false;
    bool pointOK = false;
    pointOK = vinCluPtr->getCluPol(idConf,pConf,
                                   partonCluIndx-1,partonCluIndx,partonCluIndx+1,
                                   idCluConf,pCluConf,antIndx,swapAB,hConf_d,hCluConfs); 
    if ( !pointOK ) cout<<"Warning: problem with phase-space point"<<endl;
    // Antenna Invariants
    double sAnt = m2(pConf[partonCluIndx-1]+pConf[partonCluIndx]+pConf[partonCluIndx+1]);
    double yar = (!swapAB) ? m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) ;
    double yrb = (!swapAB) ? m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) ;
    yar /= sAnt;
    yrb /= sAnt;
    //if gluon splitting, only return non-zero if the quark and anti-quark have opposite helicity
    if (idConf[partonCluIndx] != 21 && !swapAB && hConf[partonCluIndx] == hConf[partonCluIndx+1]) {return 0.0;}
    if (idConf[partonCluIndx] != 21 && swapAB && hConf[partonCluIndx-1] == hConf[partonCluIndx]) {return 0.0;}
    // 4.2) (For pimp)
    double QE2 = 0.;
    if (usePimp) {
      if (idConf[partonCluIndx] == 21 ) {
        QE2 = resolutionPtr->Q2E(pConf[partonCluIndx-1],
                        pConf[partonCluIndx],
                        pConf[partonCluIndx+1],1);            
      } else if (idConf[partonCluIndx]==-1 || idConf[partonCluIndx]==-2) {
        QE2 = m2(pConf[partonCluIndx],pConf[partonCluIndx+1]);
      } else if (idConf[partonCluIndx]==1 || idConf[partonCluIndx]==2) {
        QE2 = m2(pConf[partonCluIndx-1],pConf[partonCluIndx]);
      }
    }   
    // 4.3) (For Ariadne)
    // Compute the ARIADNE factor for gluon splitting
    double ariFactor = 1.0;
    if (usePari) {
      if (idConf[partonCluIndx] != 21) {
        double sP = sAnt;
        double sN = sP;
        if (!swapAB) {
          sN = m2(pCluConf[partonCluIndx]+pCluConf[partonCluIndx+1]);
        } else {
          sN = m2(pCluConf[partonCluIndx-2]+pCluConf[partonCluIndx-1]);
        }
        ariFactor = 2.0*sN/(sP + sN);
      }
    }
    // 4.4) Loop over the different polarization configurations for same
    //      clustering
    //      Sum_i ( Sum_hConfs (ant_i_hConfs * |M_i_hConfs|^2 ) )
    double approxSamePartonClu = 0.;    
    for (unsigned int hConfsIndx = 0; hConfsIndx <= hCluConfs.size()-1;
         ++hConfsIndx) {
      // 4.4.1) compute antenna ant_i_hConf
      double ant=0.;
      //Polarizations (special treatment when legs are reversed)
      double hA = (!swapAB) ? hCluConfs[hConfsIndx][partonCluIndx-1] 
      : hCluConfs[hConfsIndx][partonCluIndx] ;
      double hB = (!swapAB) ? hCluConfs[hConfsIndx][partonCluIndx] 
      : hCluConfs[hConfsIndx][partonCluIndx-1] ;
      double hi = (!swapAB) ? hConf[partonCluIndx-1] : hConf[partonCluIndx+1];
      double hj = hConf[partonCluIndx];
      double hk = (!swapAB) ? hConf[partonCluIndx+1] : hConf[partonCluIndx-1];
      
      ant = antSetPtr->getAnt(antIndx)->
        antennaFunction(yar,yrb,0.,0.,0.,
                        hA,hB,hi,hj,hk)/sAnt;              
      // Charge factor
      double cFac = antSetPtr->getAnt(antIndx)->getChargeFac();
      ant = cFac * ant;       
      
      // 4.4.2) call the own function to compute the approximation for
      //        |M_i_hConfs|^2
      double cluApprox = showerApprox(antSetPtr,  verbose,
                                          eCM, resolutionPtr, mgInterfacePtr, vinCluPtr,
                                          nFSpartons-1, idCluConf, 
                                          pCluConf, hCluConfs[hConfsIndx], 
                                          idBorn, pBorn, hBorn,nMatchingOrders,
                                          usePari, usePimp, QE2);
      // 4.4.3) add to the approximation the single element 
      //        ant_i_hConf * |M_i_hConfs|^2
      //        sum over hConfs only
      approxSamePartonClu += ant * cluApprox ;
      
    } // end of loop over polarization configurations
    
    //sum over the different clusters (index i)
    if (usePari) approxSamePartonClu = approxSamePartonClu * ariFactor;
    approx += approxSamePartonClu;
    
  } //end of main loop over clusters
  
  // 5) Return the approximation
  if (usePimp) approx = approx * pimp;
  return approx;
  
} //end of function showerApprox

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

double showerApproxUnpol(AntennaSetFSR* antSetPtr, 
	 int verbose, double eCM, Resolution* resolutionPtr,
	 MG4interface* mgInterfacePtr, VinClu* vinCluPtr,
         int nFSpartons, vector<int> idConf, 
         vector<Vec4> pConf, vector<int> idBorn, vector<Vec4> pBorn,
         int nMatchingOrders, bool usePari, 
         bool usePimp, double pimpQE2PreviousConf=0.){
  
  
  // 0) define relevant quantities
  int nClusters = nFSpartons - 2;
  int nPartons = idConf.size();
  int nISPartons = nPartons - nFSpartons;
  double approx = 0.0;
  vector <double> hConf;
  vector <int> hBorn;
  for (int i = 0; i < (int)idConf.size(); i++) hConf.push_back(9);
  for (int i = 0; i < (int)idBorn.size(); i++) hBorn.push_back(9);
  
  
  // 0) (For pimp) Loop over clusters to find minimum Q_E (Q_E^hat) and compute
  //    pimp factor with the Q_E given as an argument of the function.
  
  double pimp=1.;
  if (usePimp){
    double QE2Min = eCM*eCM;        
    //Tell the minimum of each ordering variable of parent configuration clusterings
    for (int partonCluIndx = nISPartons + 1 ; partonCluIndx <= nPartons - 2 ;
         ++partonCluIndx) {                 
      // Q_E is PT for gluons and Mqq for quarks
      double QE2partonClu = 0.;
      if (idConf[partonCluIndx] == 21 ) {
        QE2partonClu = resolutionPtr->Q2E(pConf[partonCluIndx-1],
                                 pConf[partonCluIndx],
                                 pConf[partonCluIndx+1],1);            
      } else if (idConf[partonCluIndx]==-1 || idConf[partonCluIndx]==-2) {
        QE2partonClu = m2(pConf[partonCluIndx],pConf[partonCluIndx+1]);
      } else if (idConf[partonCluIndx]==1 || idConf[partonCluIndx]==2) {
        QE2partonClu = m2(pConf[partonCluIndx-1],pConf[partonCluIndx]);
      }
      QE2Min = min(QE2Min,QE2partonClu);          
    }        
    pimp = QE2Min/(QE2Min+pimpQE2PreviousConf);
  }
  
  
  // 1) (For matching orders)
  if (nFSpartons == 2 + nMatchingOrders) {
    approx = mgInterfacePtr->ME2(1,idConf,pConf,hConf);
    if (usePimp) {
      approx = approx * pimp;
    }
    return approx;
  }
  
  // 2) if the number of clusters is just one the procedure is special
  //    (check that it matches with the Born)  
  if (nClusters ==1) {
    double ME2Born = mgInterfacePtr->ME2(1,idBorn,pBorn,hBorn);
    // 2.1) Cluster and compute invariants
    int partonCluIndx = (nPartons-1) -1 ;
    vector<int> idCluConf;
    vector<Vec4> pCluConf;
    int antIndx;
    bool swapAB = false;
    bool pointOK = false;
    pointOK = vinCluPtr->getClu(idConf,pConf,
                                partonCluIndx-1,partonCluIndx,partonCluIndx+1,
                                idCluConf,pCluConf,antIndx,swapAB); 
    if ( !pointOK ) cout<<"Warning: problem with phase-space point"<<endl;
    // Antenna Invariants
    double sAnt = m2(pConf[partonCluIndx-1]+pConf[partonCluIndx]+pConf[partonCluIndx+1]);
    double yar = (!swapAB) ? m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) ;
    double yrb = (!swapAB) ? m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) ;
    yar /= sAnt;
    yrb /= sAnt;
    
    // 2.2) Check that the identities of the clustered configuration partons
    //      match with the Born
    // TO DO: momenta check?
    if (idCluConf != idBorn) return 0.;
    // 2.3) Loop over the different polarization configurations for same
    //      clustering (only take the configuration that matches with the
    //      Born)
    // 2.3.2) compute antenna ant_i          
    //virtual double antennaFunction(double yij, double yjk, mui, muj, muk
    //                               int hA, int hB, 
    //                               int hi, int hj, int hk);      
    double ant = 0.;
    
    ant = antSetPtr->getAnt(antIndx)->antennaFunction(yar,yrb)/sAnt;              
    // Charge factor
    double cFac = antSetPtr->getAnt(antIndx)->getChargeFac();
    ant = cFac * ant; 
    
    // 2.3.3) add to the approximation the single element that matches
    //        the Born configuration
    //        ant_i_hConf * (ariadne)
    approx = ant * ME2Born; 
    if (usePimp) {
      approx = approx * pimp;
    }
    
    // 2.4) Return the approximation        
    return approx;
    
  }
  
  // 3) Placeholder for sector identification.

  // 4) Main Loop over clusterings 
  //    Sum_i (ant_i * |M_i|^2 )
  for (int partonCluIndx = nISPartons + 1 ; partonCluIndx <= nPartons -2 ;
       ++partonCluIndx) {
    // 4.1) Cluster and compute invariants
    vector<int> idCluConf;
    vector<Vec4> pCluConf;
    int antIndx;
    bool swapAB = false;
    bool pointOK = false;
    pointOK = vinCluPtr->getClu(idConf,pConf,
                                partonCluIndx-1,partonCluIndx,partonCluIndx+1,
                                idCluConf,pCluConf,antIndx,swapAB);
    if ( !pointOK ) cout<<"Warning: problem with phase-space point"<<endl;
    // Antenna Invariants
    double sAnt = m2(pConf[partonCluIndx-1]+pConf[partonCluIndx]+pConf[partonCluIndx+1]);
    double yar = (!swapAB) ? m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) ;
    double yrb = (!swapAB) ? m2(pConf[partonCluIndx+1],pConf[partonCluIndx]) 
    : m2(pConf[partonCluIndx-1],pConf[partonCluIndx]) ;
    yar /= sAnt;
    yrb /= sAnt;
    // 4.2) (For pimp)
    double QE2 = 0.;
    if (usePimp) {
      if (idConf[partonCluIndx] == 21 ) {
        QE2 = resolutionPtr->Q2E(pConf[partonCluIndx-1],
                        pConf[partonCluIndx],
                        pConf[partonCluIndx+1],1);            
      } else if (idConf[partonCluIndx]==-1 || idConf[partonCluIndx]==-2) {
        QE2 = m2(pConf[partonCluIndx],pConf[partonCluIndx+1]);
      } else if (idConf[partonCluIndx]==1 || idConf[partonCluIndx]==2) {
        QE2 = m2(pConf[partonCluIndx-1],pConf[partonCluIndx]);
      }
    }   
    // 4.3) (For Ariadne)
    // Compute the ARIADNE factor for gluon splitting
    double ariFactor = 1.0;
    if (usePari) {
      if (idConf[partonCluIndx] != 21) {
        double sP = sAnt;
        double sN = sP;
        if (!swapAB) {
          sN = m2(pCluConf[partonCluIndx]+pCluConf[partonCluIndx+1]);
        } else {
          sN = m2(pCluConf[partonCluIndx-2]+pCluConf[partonCluIndx-1]);
        }
        ariFactor = 2.0*sN/(sP + sN); 
      }
    }
    // 4.4) Loop over the different polarization configurations for same
    //      clustering
    //      Sum_i ( Sum_hConfs (ant_i_hConfs * |M_i_hConfs|^2 ) )
    double approxSamePartonClu = 0.;    
    // 4.4.1) compute antenna ant_i_hConf
    double ant=0.;
    
    ant = antSetPtr->getAnt(antIndx)->antennaFunction(yar,yrb)/sAnt;              
    // Charge factor
    double cFac = antSetPtr->getAnt(antIndx)->getChargeFac();
    ant = cFac * ant;       
    
    // 4.5.2) call the own function to compute the approximation for
    //        |M_i_hConfs|^2
    double cluApprox 
      = showerApproxUnpol(antSetPtr, verbose, eCM, resolutionPtr, 
                          mgInterfacePtr, vinCluPtr, nFSpartons-1, 
                          idCluConf, pCluConf, idBorn, pBorn, nMatchingOrders,
                          usePari, usePimp, QE2);

    // 4.5.3) add to the approximation the single element 
    //        ant_i_hConf * |M_i_hConfs|^2
    //        sum over hConfs only
    approxSamePartonClu += ant * cluApprox ;
    
    //sum over the different clusters (index i)
    if (usePari) approxSamePartonClu = approxSamePartonClu * ariFactor;
    approx += approxSamePartonClu;
    
  } //end of main loop over clusters
  
  // 5) Return the approximation
  if (usePimp) approx = approx * pimp;
  return approx;
  
} //end of function showerApproxUnpol


//************************************************************************
// Main Program
//************************************************************************

int main() {
  
  
  //************************************************************************
  // Define Pythia 8 generator with plug-in
  Pythia pythia; 
  // Add vincia32-specific settings parameters
  pythia.settings.init("vincia32.xml", true);
  // Define VinciaPlugin
  VinciaPlugin vincia(&pythia);

  // Read user settings from file
  pythia.readFile("vincia32.cmnd");

  //************************************************************************
  // Everything from here on is normal Pythia8
  
  // Shorthands
  //Matching* matchingPtr          = &vincia.matching;
  MG4interface* mgInterfacePtr   = &vincia.mg4interface;
  ParticleData*  particleDataPtr = vincia.particleDataPtr;
  Settings* settingsPtr          = vincia.settingsPtr;
  VinClu* vinCluPtr              = &vincia.vinClu;
  Resolution* resolutionPtr      = &vincia.resolution;  
  
  // Extract settings to be used in the main program.  
  double eCM        = settingsPtr->parm("Beams:eCM");
  int    nEvent     = settingsPtr->mode("Main:numberOfEvents");
  int    numberCount= settingsPtr->mode("Next:numberCount");
  int    verbose    = settingsPtr->mode("Vincia:verbose");    
  string procType   = pythia.word("Main:spareWord1");
  string procTypeIn = pythia.word("Main:spareWord2");
  int nMatchingOrders = settingsPtr->mode("Vincia32:nMatchingOrders");
  bool usePari      = settingsPtr->flag("Vincia32:usePari");
  
  //polarizations are defined by integers
  int hProcQ=-9;
  int hProc1 = -9;
  //int hProc1= pythia.mode("Main:spareMode1");
  int hProc2 = -9;
  int hProc3 = -9;  
  int hProcQbar=-9;
  //
  bool   force2    = settingsPtr->flag("Main:spareFlag1");
  double cut       = settingsPtr->parm("Main:spareParm1");

  AntennaSetFSR* antSetPtr = &vincia.antennaSetFSR;  
  
  //************************************************************************
  // Initialize
  vincia.init();
  
  vector<int> idOrg;
  vector<Vec4> pOrg;
  
  // First set ID of ancestor (determines matrix element)
  if (procTypeIn=="Z") idOrg.push_back(23);
  else if (procTypeIn=="H") idOrg.push_back(25);
  
  // Then set IDs of 3 <= n <= 6 decay daughters
  // So far appears to work including masses
  // but only for kinemaptype 3 (also for massless case)
  
  
  //process type: selected by command file.
  // There is the assumption that the color ordering is q-qbar in the physical 
  // quarks and the corresponding qbar-q in the radiated ones
  
  if (procType=="q-qbar-q-qbar") {
    idOrg.push_back(1);
    idOrg.push_back(-2);
    idOrg.push_back(2);
    idOrg.push_back(-1);
  } else if (procType=="q-g-qbar-q-qbar") {
    idOrg.push_back(1);
    idOrg.push_back(21);
    idOrg.push_back(-2);
    idOrg.push_back(2);
    idOrg.push_back(-1);
  } else if (procType=="q-qbar-q-g-qbar") {
    idOrg.push_back(1);
    idOrg.push_back(-2);
    idOrg.push_back(2);
    idOrg.push_back(21);
    idOrg.push_back(-1);
  } else if (procType=="q-g-qbar") {
    idOrg.push_back(1);
    idOrg.push_back(21);
    idOrg.push_back(-1);
  } else if (procType=="q-g-g-qbar") {
    idOrg.push_back(1);
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(-1);
  } else if (procType=="q-g-g-g-qbar") {
    idOrg.push_back(1);
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(-1);
  } else if (procType=="g-g-g") {
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(21);
  } else if (procType=="g-g-g-g") {
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(21);
  } else if (procType=="g-g-g-g-g") {
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(21);
    idOrg.push_back(21);
  } 
  
  int nOrg=idOrg.size();
  pOrg.resize(nOrg);
  // p[0][i] is the energy of the i'th parton
  double p[4][8];
  double S=pow(eCM,2);
  // Add mother in CM to list of pOrg
  pOrg[0]=Vec4(0.,0.,0.,eCM);
  // Set Z mass in MadGraph
  mgInterfacePtr->setParm("zmass",eCM);
  // Set h mass in MadGraph
  mgInterfacePtr->setParm("hmass",eCM);
  
  
  //***************************************************************************
  // ROOT
  //***************************************************************************
  
  
  // Form name for ROOT file
  ostringstream outFileGlb;
  // Put root files in subdirectory defined at beginning of file
  outFileGlb<<rootdir;
  outFileGlb<<"vincia32glb-";
  outFileGlb<<procTypeIn<<"-";
  outFileGlb<< procType;
  outFileGlb<< "-LO"<<nMatchingOrders;
  if (usePari) outFileGlb<< "-Pari";
  if (force2) outFileGlb<<"-force2";
  outFileGlb<<".root";
  
  // settings for ratio histograms  
  const double log10rlo = -2.05;
  const double log10rmin = -2.0;
  const double log10rhi = 2;
  int nbins = (log10rhi-log10rlo)*25;  
  
  vector <TH1F*> HRofflinePlotGlb; 
  
  //Loop over all spin configurations of external particles
  for (int i = 0; i < 32; i++){ 
    
    //Only loop over the correct number of spin configurations
    if (nOrg - 1 == 3 && idOrg[1] != 21 && i == 4) break;
    if (nOrg - 1 == 4 && idOrg[1] != 21 &&  i == 8) break;
    if (nOrg - 1 == 5 && idOrg[1] != 21 &&  i == 16) break;
    if (nOrg - 1 == 3 && idOrg[1] == 21 && i == 8) break;
    if (nOrg - 1 == 4 && idOrg[1] == 21 && i == 16) break;
    
    ostringstream Rhistoname, sectortitle, titGlb;
    Rhistoname<<"Rhisto_";
    titGlb<<"global_";
    
    //Set the polarizations of all particles
    hProcQ = 2*(i % 2) - 1;
    Rhistoname<<hProcQ<<"_";
    titGlb<<hProcQ<<"_";
    if (nOrg-1 >= 3) {
      hProc1 = 2*((int)((double)i/2.0) % 2) - 1;
      Rhistoname<< hProc1<<"_";
      titGlb<<hProc1<<"_";
      if ((nOrg-1 >= 4 && idOrg[1] != 21) || (nOrg-1 >= 3 && idOrg[1] == 21)) {
        hProc2 = 2*((int)((double)i/4.0) % 2) - 1;
        Rhistoname<< hProc2<<"_";
        titGlb<<hProc2<<"_";
        if ((nOrg-1 == 5 && idOrg[1] != 21) || (nOrg-1 >= 4 && idOrg[1] == 21) ) {
          hProc3 = 2*((int)((double)i/8.0) % 2) - 1;
          Rhistoname<< hProc3<<"_";
          titGlb<<hProc3<<"_";
          if (nOrg-1 == 5 && idOrg[1] == 21) {
            hProcQbar = 2*((int)((double)i/16.0) % 2) - 1;
            Rhistoname<< hProcQbar<<"_";
            titGlb<<hProcQbar<<"_";
          }
        }
      }
    }
    if (procTypeIn=="Z" && idOrg[1] + idOrg[nOrg-1] == 0) hProcQbar = -hProcQ;
    if (procTypeIn=="H" && idOrg[1] + idOrg[nOrg-1] == 0) hProcQbar = hProcQ;
    if (idOrg[1] != 21) {
      Rhistoname<< hProcQbar;
      titGlb<<hProcQbar;
    }
    
    // Book histo for off-line plots
    
    TH1F* tempHistGlb = new  TH1F(
                                Rhistoname.str().c_str(),
                                titGlb.str().c_str(),
                                nbins,
                                log10rlo,
                                log10rhi
                                );
    HRofflinePlotGlb.push_back(tempHistGlb);
    
  }//End loop over spin configurations for root histos 
  
  TH1F* tempHistGlb = new  TH1F(
                              "Rhisto_unpol",
                              "Rhisto_global_unpol",
                              nbins,
                              log10rlo,
                              log10rhi
                              );
  HRofflinePlotGlb.push_back(tempHistGlb);
  
  TH1F* tempHistGlb2 = new  TH1F(
                              "Rhisto_summed",
                              "Rhisto_global_summed",
                              nbins,
                              log10rlo,
                              log10rhi
                              );
  HRofflinePlotGlb.push_back(tempHistGlb2);
  
  // Construct Born-level configuration
  // (assuming it is unique)
  vector<int> idBorn;
  vector<Vec4> pBorn;
  idBorn.push_back(idOrg[0]);
  idBorn.push_back(idOrg[1]);
  idBorn.push_back(idOrg[nOrg-1]);
  pBorn.push_back(pOrg[0]);  
  // Construct (massive) momentum vectors 
  // (outgoing along z axis)
  for (int iBorn=1; iBorn<=2; iBorn++) {
    double m0   = particleDataPtr->m0(idBorn[iBorn]);
    double e    = eCM/2;
    double pAbs = sqrt(pow2(e)-pow2(m0));
    double pz   = (iBorn == 1) ? pAbs : -pAbs;
    pBorn.push_back(Vec4(0.,pz,0.,e));
  }
  
  //  // Get matrix element for this configuration
  //  // WARNING: calling ME2 method with integer polarizations.
  //  double ME2Born = mgInterfacePtr->ME2(1,idBorn,pBorn,hBorn);
  //  if (verbose >= 3) cout<<"Retrieved Born normalization from MG = "<<
  //    ME2Born<<endl;
  //  for (unsigned int iB=0;iB<pBorn.size();iB++) cout<<" "<<idBorn[iB]
  //    <<pBorn[iB]; 
  
  // Generate points
  cout << "vincia32 : generating "<<nEvent<<" points"<<endl;
  for (int iEvent=1;iEvent<=nEvent;iEvent++) {
    
    if (iEvent%numberCount == 0)  
      cout << " Now begin event " << iEvent << endl;
    
    // Generate set of massless random momenta
    bool acceptPoint = false;
    Vec4 pSum;
    while ( !acceptPoint ) {
      vincia.rambo.genPoint(S,nOrg-1,p);
      acceptPoint = true;
      for (int iParton=1;iParton<nOrg;iParton++) { 
        Vec4 VecP;
        VecP.px(p[1][iParton-1]);
        VecP.py(p[2][iParton-1]);
        VecP.pz(p[3][iParton-1]);
        VecP.e(p[0][iParton-1]);
        pOrg[iParton]=VecP;
      }      
    }
    
    //// Isolate special region
    if (force2) { 
      //double m2_12 = m2(pOrg[1],pOrg[2]);
      double m2_23 = m2(pOrg[2],pOrg[3]);
      double m2_34 = m2(pOrg[3],pOrg[4]);
      double m2_45 = m2(pOrg[4],pOrg[5]);
      // Require small invariant in configuration qgggq
      	//if ( m2_12 > cut && m2_23 > cut && m2_34 > cut && m2_45 > cut ) continue;
      // Require small invariant in configuration qggq      
        //if ( m2_12 > cut && m2_23 > cut && m2_34 > cut) continue;
      // Require small gluon splitting invariant in configuration q qbar q qbar
        //if (  m2_23 > cut ) continue;
      // Require small invariant in configuration q qbar q g qbar
        if (m2_23 > cut && m2_34 > cut && m2_45 > cut ) continue;
      }  
      
    
    double Rglobal;
    
    //TEST
    double ME2sum = 0.0;
    double RglobalSum = 0.0;
    double ME2globalSum = 0.0;
    
    
    //Loop over all spin configurations of external particles
    for (int i = 0; i < 32; i++){ 
      
      //Only loop over the correct number of spin configurations
      if (nOrg - 1 == 3 && idOrg[1] != 21 && i == 4) break;
      if (nOrg - 1 == 4 && idOrg[1] != 21 && i == 8) break;
      if (nOrg - 1 == 5 && idOrg[1] != 21 && i == 16) break;
      if (nOrg - 1 == 3 && idOrg[1] == 21 && i == 8) break;
      if (nOrg - 1 == 4 && idOrg[1] == 21 && i == 16) break;
        
      //Set the polarizations of all particles
      hProcQ = 2*(i % 2) - 1;
      if (nOrg-1 >= 3) {
        hProc1 = 2*((int)((double)i/2.0) % 2) - 1;
        if ((nOrg-1 >= 4 && idOrg[1] != 21) || (nOrg-1 >= 3 && idOrg[1] == 21)) {
          hProc2 = 2*((int)((double)i/4.0) % 2) - 1;
          if ((nOrg-1 == 5 && idOrg[1] != 21) || (nOrg-1 >= 4 && idOrg[1] == 21) ) {
            hProc3 = 2*((int)((double)i/8.0) % 2) - 1;
            if (nOrg-1 == 5 && idOrg[1] == 21) {
              hProcQbar = 2*((int)((double)i/16.0) % 2) - 1;
            }
          }
        }
      }
      if (procTypeIn=="Z" && idOrg[1] != 21) hProcQbar = -hProcQ;
      if (procTypeIn=="H" && idOrg[1] != 21) hProcQbar = hProcQ;
        
        
      //set polarizations for BORN. H,Z->qqbar
      vector<int> hBorn;
      hBorn.push_back(9);
      if (idOrg[1] != 21) {
        hBorn.push_back(hProcQ);
        hBorn.push_back(hProcQbar);
      }
            
      //set polarizations for the configuration
      //Z(unpolarized) or H
      vector<int> hOrg;
      hOrg.push_back(9);
      //q
      hOrg.push_back(hProcQ);
      if (nOrg-1 >= 3 ) {
        hOrg.push_back(hProc1);          
        if ((nOrg-1 >= 4 && idOrg[1] != 21) || (nOrg-1 >= 3 && idOrg[1] == 21)) {
          hOrg.push_back(hProc2);          
          if ((nOrg-1 == 5 && idOrg[1] != 21) || (nOrg-1 >= 4 && idOrg[1] == 21)) {
            hOrg.push_back(hProc3);
            if (nOrg-1 == 5 && idOrg[1] == 21) {
                hOrg.push_back(hProcQbar);
            }
          }
        }
      }
      //qbar
      if (idOrg[1] != 21) hOrg.push_back(hProcQbar);
      
//for (int it = 0; it < hOrg.size(); it++) cout << "hOrg[" << it << "] = " << hOrg[it] << endl;
//for (int it = 0; it < idOrg.size(); it++) cout << "idOrg[" << it << "] = " << idOrg[it] << endl;
//for (int it = 0; it < pOrg.size(); it++) cout << "pOrg[" << it << "].e() = " << pOrg[it].e() << endl;
        
        
      // Compute LO matrix element for this n-parton phase-space point
      double ME2 = mgInterfacePtr->ME2(1,idOrg,pOrg,hOrg);
//cout << "ME2 = " << ME2 << endl;
      //TEST
      ME2sum += ME2;
      
      
      //nOrg= npartons + 1 (parent particle)
      int nFSpartons = nOrg-1;
      
      double ME2global = 0.0;
        
      for (int jp = 0; jp < 2; ++jp) {
        
//if (jp == 1) cout << "jp = " << jp << endl;
          
        if (idBorn[1] == 21) {
          hBorn.clear();
          hBorn.push_back(9);
          hBorn.push_back(2*jp - 1);
          hBorn.push_back(2*jp - 1);
        }

//cout << "hBorn[1] = " << hBorn[1] << ", hBorn[2] = " << hBorn[2] << endl;
          
        ME2global +=
          showerApprox(antSetPtr, verbose,
                       eCM, resolutionPtr, mgInterfacePtr, vinCluPtr,
                       nFSpartons, idOrg, 
                       pOrg, hOrg, 
                       idBorn, pBorn, hBorn, nMatchingOrders,
                       usePari, true);
        
        if (idBorn[1] != 21) break;
        
      }
      
      //TEST
      ME2globalSum += ME2global;
      
      Rglobal = ME2global / ME2;
      
      // Fill off-line histogram
      
      HRofflinePlotGlb[i]->Fill(log10(max(Rglobal,pow(10,log10rmin))));
      
    }//End spin configuration loop 
    
    //nOrg= npartons + 1 (parent particle)
    int nFSpartons = nOrg-1;
    
    double ME2globalUnpol = 
    showerApproxUnpol(antSetPtr, verbose,
                      eCM, resolutionPtr, mgInterfacePtr, vinCluPtr,
                      nFSpartons, idOrg, pOrg, 
                      idBorn, pBorn, nMatchingOrders,
                      usePari, true);
    
    //set as unpolarized for BORN. H,Z->qqbar
    vector<int> hBorn;
    hBorn.push_back(9);
    hBorn.push_back(9);
    hBorn.push_back(9);
    
    //set as unpolarized for the configuration
    //Z(unpolarized) or H
    vector<int> hOrg;
    hOrg.push_back(9);
    //q
    hOrg.push_back(9);
    if (nOrg-1 >= 3 ) {
      hOrg.push_back(9);
      if (nOrg-1 >= 4) {
        hOrg.push_back(9);
        if (nOrg-1 == 5) {
          hOrg.push_back(9);
        }
      }
    }
    //qbar
    hOrg.push_back(9);
    
    double ME2unpol = mgInterfacePtr->ME2(1,idOrg,pOrg,hOrg);
    
    double RglobalUnpol = ME2globalUnpol / ME2unpol;
    
    //
    //Below: some comparisons for diagnostics
    RglobalSum = ME2globalSum / ME2sum;   
//        //TEST
//    		printf("RglobalSum = %lf, RglobalUnpol = %lf\n",RglobalSum,RglobalUnpol); 
//    		printf("ME2sum = %lf, ME2unpol = %lf\n",ME2sum,ME2unpol); 
//    		printf("ME2globalSum = %lf, ME2globalUnpol = %lf\n",ME2globalSum,ME2globalUnpol); 
    
    // Fill off-line histogram
    
    HRofflinePlotGlb[(int)HRofflinePlotGlb.size()-2]->Fill(log10(max(RglobalUnpol,pow(10,log10rmin))));
    HRofflinePlotGlb.back()->Fill(log10(max(RglobalSum,pow(10,log10rmin))));
    
  } // End of event generation loop
  
  // Open and save ROOT file
  TFile fileRoot(outFileGlb.str().c_str(),"recreate");
  // Opened root file:
  cout <<"vincia32 : saving ROOT file "<<outFileGlb.str()<<endl;  
  
  for (int i = 0; i < (int)HRofflinePlotGlb.size(); i++){
    // give the integral as an output
    cout << "Integral of " << outFileGlb.str() << "= " 
         << HRofflinePlotGlb[i]->Integral() << endl;
    if (HRofflinePlotGlb[i]->Integral() == 0.0) {cout << "ERROR. INTEGRAL 0.0";}
    HRofflinePlotGlb[i]->Scale(1./HRofflinePlotGlb[i]->Integral());
    HRofflinePlotGlb[i]->Write();
  }
  
  fileRoot.Close();
  
  //  // Wait for user keystroke before terminating (kills display)
  //  cout << endl << "Enter q to terminate..." << endl;
  //  char tmp;cin>>tmp;
  
  // Done.
  return 0;
}

