// MG5interface.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2018 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Functionality to interface to FORTRAN matrix elements generated with MadGraph 5

#ifndef VINCIA_MG5interface_H
#define VINCIA_MG5interface_H

// Include Pythia and headers
#include "Pythia8/Basics.h"
#include "Pythia8/PythiaComplex.h"
#include "Pythia8/Event.h"
#include "Pythia8/Info.h"
#include "Pythia8/ParticleData.h"
#include "Pythia8/PythiaStdlib.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"
//Used for ostream_iterator, which is used to convert a vector of integers
//into a vector of doubles
#include <iostream>
#include <iterator>

// Include Vincia Standard Library
#include "Vincia/VinciaCommon.h"

// Include VinciaMG5 wrapper headers
#include "VinciaMG5/allHeaders.inc"

namespace Vincia {

//==========================================================================

// The MG5interface class maintains a list of internal MG5 FOTRAN matrix elements,
// along with the functionality to interface them from VINCIA. It is also responsible
// for filling the MG5 mass and coupling COMMON blocks with information from VINCIA's
// and PYTHIA's parameter databases. User-defined MG5 matrix elements can also be added.
 
class MG5interface {
   
 public:
   
   // Constructor.
   MG5interface() { 
     isInitPtr=false; 
     isInit=false;
     MG5matrixList.resize(0);
     MG5matrixListUser.resize(0);
     ampMap.clear();
   };
   
   // Destructor
   virtual ~MG5interface() {
     for (int i=0; i<(int)MG5matrixList.size(); ++i)
       delete MG5matrixList[i];
     for (int i=0; i<(int)MG5matrixListUser.size(); ++i)
       delete MG5matrixListUser[i];     
   };
   
   // Set pointers to required PYTHIA 8 objects
   void initPtr(Couplings* couplingsPtrIn, ParticleData* particleDataPtrIn,
		Rndm* rndmPtrIn, Settings* settingsPtrIn,
		SusyLesHouches* slhaPtrIn, VinciaCommon* vinComPtrIn) {
     couplingsPtr    = couplingsPtrIn;
     particleDataPtr = particleDataPtrIn;
     settingsPtr     = settingsPtrIn;
     rndmPtr         = rndmPtrIn;
     slhaPtr         = slhaPtrIn;
     vinComPtr       = vinComPtrIn;
     isInitPtr       = true;
   }
   
   // Main initialization routine. 
   bool init();

   // Auxiliary init function to create the list of internal MG5 MEs
   void initAmplitudes();

   // Determine process label for a given set of particle ID codes
   // Input:
   //  vector<int> id : PDG id codes in any order
   //  iA : index of first incoming particle in id list
   //  iB : (optional) index of second incoming particle in id list
   // Output: 
   //  Incoming: ordered from negative to positive ID codes
   //  Outgoing: ordered from negative to positive ID codes
   vector<int> makeProcessLabel(vector<int> id, int iA=0, int iB=-1) {
     processLabelSav.resize(0);
     // Set incoming partons, ordered by PDG ID
     int idIn1 = id[iA];
     if (iB >= 0) {
       int idIn2 = id[iB];
       processLabelSav.push_back(min(idIn1,idIn2));
       processLabelSav.push_back(max(idIn1,idIn2));
     } else {
       processLabelSav.push_back(idIn1);
       processLabelSav.push_back(0);
     }
     // Make list of final-state PDG IDs
     list<int> listFS;
     for (int i=0; i<(int)id.size(); ++i) {
       if (i == iA || i == iB) continue;
       listFS.push_back(id[i]);       
     }
     // Sort list
     listFS.sort();
     list<int>::iterator it;
     for (it=listFS.begin(); it!=listFS.end(); ++it)
       processLabelSav.push_back(*it);
     return processLabelSav;
   }

   // Wrapper with vector of particles as input
   vector<int> makeProcessLabel(vector<Particle> p, int iA=0, int iB=-1) {
     vector<int> id;
     for (int i=0; i<(int)p.size(); ++i) id.push_back(p[i].id());
     return makeProcessLabel(id, iA, iB);
   }
   
   // Get the matrix element squared for a particle state
   double ME2( vector<Particle> pIn, int iA, int iB);

   // Use ME2 to set helicities for a state.
   // (Note: takes a reference as input and operates on it)
   bool selectHelicities( vector<Particle>& pIn, int iA, int iB);
   
   // Set verbose level
   void setVerbose(int verboseIn) {verbose = verboseIn;}
   
   // Set color depth
   void setColorDepth(int colorDepthIn) {colorDepth = colorDepthIn;}
   // Get color depth
   int getColorDepth() {return colorDepth;}
   
   // Return pointer to an MG5 matrix element by process label
   MG5matrix* getMG5matrixPtr( vector<int> processLabel ) {
     if (ampMap.find(processLabel) != ampMap.end())
       return ampMap[processLabel];
     else return 0;
   }

   // Add pointer to a user-defined MG5 matrix element 
   // (E.g., user can define a matrix element in their main program
   //  and add it to Vincia's list in this way.)
   void addMG5matrixPtr(MG5matrix* matrixPtr) {
     MG5matrixListUser.push_back(matrixPtr);
   }

   // Convert a process label to a string, e.g. for printing to stdout
   string processLabel2string(vector<int> labelIn, bool convertToNames=false);

 private:
   
   // Is initialized?
   bool isInitPtr, isInit;
   
   // Pointers to VINCIA and PYTHIA 8 objects
   Couplings*      couplingsPtr;
   ParticleData*   particleDataPtr;
   Rndm*           rndmPtr;
   Settings*       settingsPtr;
   VinciaCommon*   vinComPtr;
   SusyLesHouches* slhaPtr;
   
   // Container for Process Label
   vector<int> processLabelSav;
   
   // Amplitude map, accessed by process label
   map< vector<int>, MG5matrix* > ampMap;

   // List of internal and user-defined MG5 matrix elements
   vector<MG5matrix*> MG5matrixList;
   vector<MG5matrix*> MG5matrixListUser;
   
   // MG5 model parameters
   MG5parameters_sm pars; 

   // Color Mode (0: leading color, 1: Vincia color)
   int colorDepth;

   // Saved list of helicity components for last ME evaluated
   map< vector<int> , double > me2hel;
   
   // Verbose level
   int verbose;   
   
 };

//==========================================================================

} // end namespace Vincia

#endif

