// SubtractionTerms.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2018 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's subtraction terms
// class SubtractionTerms
// class Functions

#ifndef VINCIASUBTRACTIONTERMS_H
#define VINCIASUBTRACTIONTERMS_H

#include "Vincia/VinciaCommon.h"  
#include "math.h"

namespace Vincia {
  
// *****************************************************************
// class SubtractionTerms :
// Base class implementation for generic subtraction terms 
// Note: the base class implements the GGG (sub-)antennae;
// see Gehrmann, Gehrmann-de Ridder, Glover: JHEP 0509 (2005) 056

class SubtractionTerms {

 public:
      
  // Constructor.
    
  SubtractionTerms() {isInit=false;}

  // Destructor.

  ~SubtractionTerms() {}

  // Initialize Pointers (must be done before init)
  void initPtr(Settings* settingsPtrIn) {
    settingsPtr    = settingsPtrIn; 
    isInitPtr=true;
  }
    
  // Initialize
  bool init();

  // Differential subtraction terms (= sub-antennae)
  double AqqEmit(double yij,double yjk);
  double AqgEmit(double yij,double yjk);
  double AqgSplit(double yij,double yjk);
  // Wrapper to get an enumerated antenna
  double getA(int iAnt, double yij, double yjk) {
    if (iAnt == iQQemit) return AqqEmit(yij,yjk);
    if (iAnt == iQGemit) return AqgEmit(yij,yjk);
    if (iAnt == iQGsplit) return AqgSplit(yij,yjk);
    else return 0.;
  }
    
  // Integrated subtraction terms (over resolved region defined by
  // the two dimensionless invariants and the type of evolution variable)
  double IqqEmit(double yij,double yjk,int evType);
  double IqgEmit(double yij,double yjk,int evType,int ordType);
  double IqgSplit(double yij,double yjk,int evType,int ordType,double N,
                  bool useMqq=true);


 private:

  // Is initialized?    
  bool isInitPtr, isInit;
    
  // Pointer to PYTHIA 8 Settings database
  Settings* settingsPtr;

  // Verbose level
  int verbose;   

  // Declarations for strong pT ordering
  double yEpT(double,double);
  double DLStpT(double);
  //double SL1StpT(double);
  //double SL2StpT(double);
  double SL12StpT(double);
  double SL3StpT(double);
  double CStpT(double,double,double,double); 
  // Declarations for strong mD ordering
  double DLStmD(double);
  double SL1StmD(double);
  double SL2StmD(double);
  double SL3StmD(double);
  double CStmD(double,double,double,double);
  // Declarations for energy ordering
  static const double del;
  double yEEn(double,double);
  double DLEn(double);
  double SL1En(double);
  double SL2En(double);
  // Declarations for pT smooth ordering
  double DLSmpT(double);
  double SL1SmpT(double);
  double SL2SmpT(double);
  double SL3SmpT(double);
  double CSmpT(double,double,double,double);
  // Declarations for smooth mD ordering
  double DLSmmD(double);
  double SL1SmmD(double);
  double SL2SmmD(double);
  double SL3SmmD(double);
  double CSmmD(double,double,double,double); 
  // Declaration for mqq strong ordering
  double SL1Stmqq(double);
  double SL2Stmqq(double);
  double SL3Stmqq(double);
  double CStmqq(double,double,double,double); 
  // Declaration for mqq smooth pT ordering
  //static const double NpT;
  double SL1SmpTmqq(double,double);
  double SL2SmpTmqq(double,double);
  double SL3SmpTmqq(double,double);
  double CSmpTmqq(double,double,double,double,double);
  // Declaration for mqq smooth mD ordering
  //static const double NmD;
  double SL1Smmqq(double,double);
  double SL2Smmqq(double,double);
  double SL3Smmqq(double,double);
  double CSmmqq(double,double,double,double,double);
};

class Functions {

 public: 

  // Constructor
  Functions() {};

  // *********************************************************************
  // Auxiliary Hardcoded Matrix Elements and Antenna Functions
  // *********************************************************************
    
  // Eikonal (without color or coupling factors)
  double Eikonal(Vec4 a, Vec4 r, Vec4 b) {
    double sar = m2(a+r);
    double srb = m2(r+b);
    double sab = m2(a+b);
    return 2*sab/(sar*srb);
  }
  double Eikonal(Particle a, Particle r, Particle b) {
    return Eikonal(a.p(), r.p(), b.p());
  }
  double Eikonal(Particle* a, Particle* r, Particle* b) {
    return Eikonal(a->p(), r->p(), b->p());
  }
  double Eikonal(double yij, double yjk) {
    return 2.*(1.-yij-yjk)/yij/yjk;
  }

  // Helicity-dependent Altarelli-Parisi kernels (mu = m/Q)
  // Note: Pg2gg is written with standard factor 2 normalization convention
  double Pg2gg(double z, int hA=9, int hB=9, int hC=9); 
  double Pg2qq(double z, int hA=9, int hB=9, int hC=9, double mu=0.); 
  double Pq2qg(double z, int hA=9, int hB=9, int hC=9, double mu=0.); 
  double Pq2gq(double z, int hA=9, int hB=9, int hC=9, double mu=0.) {
    return Pq2qg(1-z,mu,hA,hC,hB);
  }

  // Wrappers to get unpolarized massive Altarelli-Pariis kernels (mu = m/Q)
  double Pg2qq(double z, double mu) { return Pg2qq(z,9,9,9,mu); }
  double Pq2qg(double z, double mu) { return Pq2qg(z,9,9,9,mu); }
  double Pq2gq(double z, double mu) { return Pq2gq(z,9,9,9,mu); }

  // Altarelli-Parisi kernels with linear in/out polarizations for gluons
  //   pol = +1 for in-plane, -1 for out-of-plane
  double Pg2ggLin(double z, int polA=9, int polB=9, int polC=9); 
  double Pg2qqLin(double z, int polA=9, int hB=9, int hC=9, double mu=0.); 
  double Pq2qgLin(double z, int hA=9, int hB=9, int polC=9,double mu=0.); 
  double Pq2gqLin(double z, int hA=9, int polB=9, int hC=9,double mu=0.) {
    return Pq2qgLin(1-z,mu,hA,hC,polB);
  }

  // Z -> 3-parton antenna functions (= tree-level GGG antennae)
  double A30(vector<Vec4>& pn, int a, int r, int b);
  // Z -> 4-parton antenna functions (= tree-level GGG antennae)
  double A40(vector<Vec4>& pn, int a, int b, int c, int d);
  double A40aux(vector<Vec4>& pn, int a,int b,int c,int d);
  double B40(vector<Vec4>& pn, int a, int b, int c, int d);
  double B40aux(vector<Vec4>& pn, int a, int b, int c, int d);

 private:

};

}

#endif
