// VinciaRootSettings.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2018 P. Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Vincia ROOT utilities

#ifndef VINCIAROOTSETTINGS_H
#define VINCIAROOTSETTINGS_H

// Main ROOT header
#include "TROOT.h"
#include "Pythia8/Settings.h"
#include "TSystem.h"              // gSystem
// 1D and 2D histograms 
#include "TH1.h"
#include "TH2.h"
#include "TPaveStats.h"
// Graphs and Graphics
#include "TColor.h"
#include "TGraphAsymmErrors.h"
#include "TLine.h"
#include "TStyle.h"
// LaTeX, Text, and Legends
#include "TLatex.h"
#include "TLegend.h"
#include "TLegendEntry.h"
// Widget windows
#include "TCanvas.h"
#include "TApplication.h"
// To read/write files
#include "TDirectory.h"
#include "TFile.h"

// Standard Libraries (same as in PYTHIA 8)
// Stdlib header files for mathematics.
#include <cmath>
#include <cstdlib>

// Stdlib header files for strings and containers.
#include <string>
#include <vector>
#include <map>
#include <deque>
#include <set>

// Stdlib header file for input and output.
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>

// Define pi if not yet done.
#ifndef M_PI
#define M_PI 3.1415926535897932385
#endif

// Put all the rest in Vincia namespace

namespace Vincia {
  
  // Include the std namespace while inside the Vincia namespace.
  using namespace std;

  // ------------------------------------------------------------------------
  // class VRsettings
  // Contains global settings for the VinciaRoot interface

  class VRsettings {

  public:

    // Constructor
    VRsettings() : isInitApp(false), logY(false),  logX(false), 
      normalize(false), showRatioPad(true), showRelUncPads(true), 
      hasLeftLabel(false), hasRightLabel(false), 
      rebinLikeData(false), zeroSkipping(false), 
      showLegendChi2(false), saveDisplays(false), saveHistograms(false),
      verbose(1), optStat(0), nUpdate(1.), iLastUpdate(0.), iLastSave(0.), fSave(1.),
      sigmaStat(1.), normFac(1.), ratioMin(0.5), ratioMax(1.5),
      MCnames(" "), signature("V I N C I A R O O T") { 
      legend.resize(0); 
      logY = false;
      logX = false;
      // Default plot sizes
      plotSizeX = 360;
      plotSizeY = 420;
      // Default uncertainty variation names = VINCIA ones
      uncertaintyLabels.resize(0);
      uncertaintyLabels.push_back(string("1"));
      uncertaintyLabels.push_back(string("2"));
      uncertaintyLabels.push_back(string("3"));
      uncertaintyLabels.push_back(string("4"));
      uncertaintyLabels.push_back(string("5"));
      uncertaintyLabels.push_back(string("6"));
      uncertaintyLabels.push_back(string("7"));
      uncertaintyLabels.push_back(string("8"));
      uncertaintyLabels.push_back(string("9"));
      uncertaintyLabels.push_back(string("10"));
      //uncertaintyLabels.push_back(string("NLC#font[122]{-}"));
    };

    // Set pointer to PYTHIA settings database
    void init() {
      nUpdate = settingsPtr->mode("VinciaRoot:nUpdate");
      nSave   = settingsPtr->mode("VinciaRoot:nSave");
      fSave   = settingsPtr->parm("VinciaRoot:fSave");
      saveDisplays   = settingsPtr->flag("VinciaRoot:saveDisplays");
      saveHistograms = settingsPtr->flag("VinciaRoot:saveHistograms"); 
    }

    // Functions to determine whether to update displays
    int nextUpdate(int offset=0) { return (iLastUpdate + offset + 1) * nUpdate; }
    bool doUpdate(double iEvent) {
      // Reset if starting over
      if (iEvent < iLastUpdate * nUpdate) iLastUpdate = 0;
      bool update = false;
      if ( iEvent >= nextUpdate() ) update = true;
      iLastUpdate = int( (iEvent+0.5)/nUpdate );
      return update;
    }
    // Functions to determine whether to save displays
    int nextSave(int offset=0) {
      double multNow  = max(1.,iLastSave + offset + 1);
      multNow = max(multNow, pow(2., max(0.,(iLastSave+offset))/fSave ));
      return multNow * nSave;
    }    
    bool doSave(double iEvent) {
      // Reset if starting over
      if (iEvent < iLastSave * nSave) iLastSave = 0;
      if ( iEvent >= nextSave() ) {
	++iLastSave;
	return true;
      } else return false;
    }
    // Data members
    bool isInitApp;
    bool logY, logX, normalize;
    bool showRatioPad, showRelUncPads;
    bool hasLeftLabel, hasRightLabel;
    bool rebinLikeData, zeroSkipping;
    bool showLegendChi2;
    bool saveDisplays, saveHistograms;
    int  verbose, optStat, nUpdate;
    double iLastUpdate, iLastSave, fSave, nSave;
    double sigmaStat, normFac, ratioMin, ratioMax;
    double plotSizeX, plotSizeY;
    string MCnames, signature, leftLabel, rightLabel;    
    vector<string> legend;
    vector<string> uncertaintyLabels;

    // Pointer to Pythia Settings database
    Pythia8::Settings* settingsPtr;
    
  };
  
}

#endif
