  // MEcorrections.cc is a part of the VINCIA plug-in to the 
// PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2018 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

#include "Vincia/MHV.h"

namespace Vincia {

//------------------------------------------------------------------------//

///////////////////Functions from class SpinorProducts//////////////////////

//------------------------------------------------------------------------//

  // Calculate sqrt(k^+) = sqrt(k[0] + k[3])
  std::complex<double> SpinorProducts::sqrt_k_plus (Vec4 k) {
    double kPos = k.pPos();
    if (kPos < 0) {
      return std::complex<double>(0, sqrt(abs((kPos))));
    }
    else {
      return std::complex<double>(sqrt(kPos),0);
    }
  }

  // Calculate sqrt(k^-) = sqrt(k[0] - k[3])
  std::complex<double> SpinorProducts::sqrt_k_minus (Vec4 k) {
    double kNeg = k.pNeg();
    if (kNeg < 0) {
      return std::complex<double> (0, sqrt((abs(kNeg))));
    }
    else {
     return std::complex<double> (sqrt(kNeg), 0);
      
    }
  }
  
  // Calculate exp(i\phi_k), used to calculate the spinor inner product
  // exp(i\phi_k) = (k^1 + ik^2)/(\sqrt(k^+))
  void SpinorProducts::phasePos (Vec4 k_i, Vec4 k_j) {
    phaseDenom_i = sqrt_k_plus (k_i);
    phaseDenom_j = sqrt_k_plus (k_j);
    if (abs(real(phaseDenom_i)) < TINY && abs(imag(phaseDenom_i)) < TINY) {
      if (abs(real(phaseDenom_j)) < TINY && abs(imag(phaseDenom_j)) < TINY) {
        phase_i = pie;
        phase_j = pie;
      }
      else {
        phase_i = pie;
        phase_j = std::complex<double> (k_j.px(), k_j.py());
        phase_j /= phaseDenom_j;
      } 
    }
    else {
      if (abs(real(phaseDenom_j)) < TINY && abs(imag(phaseDenom_j)) < TINY) {
        phase_i = std::complex<double> (k_i.px(), k_i.py());
        phase_i /= phaseDenom_i;
        phase_j = pie;
      }
      else {
        phase_i = std::complex<double> (k_i.px(), k_i.py());
        phase_i /= phaseDenom_i;
        phase_j = std::complex<double> (k_j.px(), k_j.py());
        phase_j /= phaseDenom_j;
      }
    }
  }

  // Calculate exp(-i\phi_k), used to calculate the spinor inner product
  void SpinorProducts::phaseNeg (Vec4 k_i, Vec4 k_j) {
    phaseDenom_i = sqrt_k_plus (k_i);
    phaseDenom_j = sqrt_k_plus (k_j);
    if (abs(real(phaseDenom_i)) < TINY && abs(imag(phaseDenom_i)) < TINY) {
      if (abs(real(phaseDenom_j)) < TINY && abs(imag(phaseDenom_j)) < TINY) {
        phase_i = pie;
        phase_j = pie;
      }
      else {
        phase_i = pie;
        phase_j = std::complex<double> (k_j.px(), -k_j.py());
        phase_j /= phaseDenom_j;
      } 
    }
    else {
      if (abs(real(phaseDenom_j)) < TINY && abs(imag(phaseDenom_j)) < TINY) {
        phase_i = std::complex<double> (k_i.px(), -k_i.py());
        phase_i /= phaseDenom_i;
        phase_j = pie;
      }
      else {
        phase_i = std::complex<double> (k_i.px(), -k_i.py());
        phase_i /= phaseDenom_i;
        phase_j = std::complex<double> (k_j.px(), -k_j.py());
        phase_j /= phaseDenom_j;
      }
    }
  }

  // Calculate the spinor inner product <ij>
  // <ij> = sqrt(k_j^+) exp(i\phi_i) - sqrt(k_i^+) exp(i\phi_j)
  std::complex<double> SpinorProducts::spinorProdPos (Vec4 k_i, Vec4 k_j){
    // First calculate phases
    phasePos(k_i, k_j);
    // If k_i^+ == 0
    if (phase_i == pie) { 
      if (phase_j == pie) {
        return 0;
      }
      else {
        return sqrt_k_minus (k_i) * phaseDenom_j;
      }
    }
    else if (phase_j == pie) {
      return - sqrt_k_minus (k_j) * phaseDenom_i;
    }
    else {
      return  phaseDenom_j * phase_i - phaseDenom_i * phase_j;
    }
  } 

  // Calculate the spinor inner product [ij]
  // [ij] = sqrt(k_i^+) exp(-i\phi_j) - sqrt(k_j^+) exp(-i\phi_i) 
  std::complex<double> SpinorProducts::spinorProdNeg (Vec4 k_i, Vec4 k_j){
    // First calculate phases
    phaseNeg(k_i, k_j);
    // If k_i^+ == 0
    if (phase_i == pie) { 
      if (phase_j == pie) {
        return 0;
      } 
      else {
        return - sqrt_k_minus (k_i) * phaseDenom_j;
      }
    }
    else if (phase_j == pie) {
      return sqrt_k_minus (k_j) * phaseDenom_i;
    }
    else {
      return  phaseDenom_i * phase_j - phaseDenom_j * phase_i;
    }
  }
  
//------------------------------------------------------------------------//

/////////////////////Functions from class MHVamplitudes/////////////////////

//------------------------------------------------------------------------//
  
  // Initialise couplings needed for EW processes
  bool MHVamplitudes::init() {
    // Set verbose level
    verbose = settingsPtr->mode("Vincia:verbose");

    // Verbose output
    if (verbose >= 2) printOut("MHVamplitudes::init","...");

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("MHVamplitudes::init","Cannot initialize, pointers not set.");
      return false;      
    }
    
    // Innocent until proven guilty
    isInit = true;
    
    // Set color depth
    colorDepth = (settingsPtr->flag("Vincia:matchingFullColor")) ? 1 : 0;
    
    // Set couplings and parameters for MHV amplitdues (which use 
    // different conventions to MG4)
    // CKM
    Vud = couplingsPtr->VCKMgen(1,1);
    Vus = couplingsPtr->VCKMgen(1,2);
    Vub = couplingsPtr->VCKMgen(1,3);
    Vcd = couplingsPtr->VCKMgen(2,1);
    Vcs = couplingsPtr->VCKMgen(2,2);
    Vcb = couplingsPtr->VCKMgen(2,3);
    
    // Left and right handed fermions to weak bosons
    alpha   = settingsPtr->parm("StandardModel:alphaEMmZ");
    sin2w   = settingsPtr->parm("StandardModel:sin2thetaW");
    sw      = sqrt(sin2w);
    cw      = sqrt(1.-sin2w);
    ee2     = alpha*FourPi*2;
    ee      = sqrt(ee2);
    ez      = ee/(sw*cw);
    ey      = ee*sw/cw;
    efd     = couplingsPtr->ef(1);
    efu     = couplingsPtr->ef(2);
    efl     = couplingsPtr->ef(11);
    gZul    = -ez*(0.5 -efu*sin2w);
    gZur    = efu*ey;
    gZdl    = -ez*(-0.5 -efd*sin2w);
    gZdr    = efd*ey;
    gZll    = -ez*(-0.5 - efl*sin2w);
    gZlr    = efl*ey;
    gAl     = ee*efl;
    gAu     = ee*efu;
    gAd     = ee*efd;
    gWfl    = -ee/(2*sw*sqrt(2.));
    gWflSq  = pow(gWfl,2.);
    gWfr    = 0.;
    gWfrSq  = pow(gWfr,2.);
    
    // Masses, widths and propagators
    mW      = particleDataPtr->m0(24);
    mWSq    = pow(mW,2.);
    mZ      = particleDataPtr->m0(23);
    mZSq    = pow(mZ,2.);
    widthW  = particleDataPtr->mWidth(24);
    widthZ  = particleDataPtr->mWidth(23);
    propW   = complex (-mWSq,widthW*mW);
    propZ   = complex (-mZSq,widthW*mZ);

    // Set all amplitudes, spinor products to 0
    for (int i = 0; i < nPartonsMax; i++)
      for (int j = 0; j < nPartonsMax; j++) spBag[i][j] = 0; 
    for (int i = 0; i < nAmpsMax; i++) ampBag[i] = 0;
    
    return true;
  }
  // Calculate spinor products and other values needed for ME
  bool MHVamplitudes::calcSpinorProds(vector<Vec4> k, vector<int> h, vector<int> ID, bool isOrdered) {
    //Sanity check
    if (k.size() != h.size() || k.size() != ID.size()) {
      Error = true;
      if (verbose>=1) printOut("Vincia::MHVamplitudes::calcSpinorProds",
	    "Inconsistancy with size of inputs. ");
      return false;
    }
    
    // Initialise values
    nPartons         = k.size();
    Error            = false;
    fermHelConserved = true;
    nonChiralW       = false;
    
    // Check particle number
    if (nPartons > nPartonsMax || nPartons < 4) {
      Error = true;
      if (verbose>=1) 
	cout << " Vincia::MHVamplitudes::calcSpinorProds, Check number of particles."
	     << " Maximum number of particles is " << nPartonsMax 
	     << " Minimum number of particles is 4" << endl;
      return false;
    }
        
    if (!isInit) {
      Error = true;
      printOut("Vincia::MHVamplitudes::calcSpinorProds",
	       "MHVamplitudes not initialised.");
      return false;
    }
    
    if (pos1Sav == -1 || pos2Sav == -1) {
      Error = true;
      printOut("Vincia::MHVamplitudes::calcSpinorProds","Initial partons not set.");
      return false;
    }
    
    // Create vectors to store momenta, id, hel
    vector<Vec4> p; p.clear(); p.resize(0);
    vector<int> ids; ids.clear(); ids.resize(0);
    vector<int> hel; hel.clear(); hel.resize(0);
    
    // Convert incoming to outgoing
    h[pos1Sav]    = -h[pos1Sav];
    k[pos1Sav]    = -k[pos1Sav];
    h[pos2Sav]    = -h[pos2Sav];
    k[pos2Sav]    = -k[pos2Sav];
    if      (abs(ID[pos1Sav]) < 17) ID[pos1Sav] = -ID[pos1Sav];
    if      (abs(ID[pos2Sav]) < 17) ID[pos2Sav] = -ID[pos2Sav];
    
    // Check amplitude type 
        nGlue  = 0;
    int nQuark = 0;
    int nLep   = 0;
    int nPos   = 0;
    isQCD      = true;
    isWeak     = false;
    allGlue    = false;
    twoQuark   = false;
    fourQuark  = false;
    mhvPlus    = false;
    mhvMinus   = false;
    sameFlav   = false;
    prefact2   = zero;
    
    for (int i = 0; i < nPartons; i++) {
      if (ID[i] == 21) nGlue += 1;
      else if (abs(ID[i]) < 6 && ID[i] != 0) nQuark += 1;
      else if (abs(ID[i]) <= 16 && abs(ID[i]) >= 11) {
	nLep  += 1;
	isQCD  = false;
	isWeak = true;
      }
      else {
	Error = true;
	cout << " Check particle IDs" << endl;
	return false;
      }
      if (h[i] == 1) nPos += 1;
      else if (h[i] == 9) {
	cout << " Particles must be polarised." << endl;
	Error = true;
	return false;
      }
    }
    if (verbose >= 6) {
      printOut("Vincia::mhvAmplitudes::calcSpinorProds","Printing number of positive-helicity particles");
      cout << " nPos = " <<nPos << endl; 
    }
    
    if (nPos == nPartons - 2)  mhvPlus  = true;
    else if (nPos == 2)        mhvMinus = true;
    
    //Average over incoming colours
    aveColIn = 1.;
    if (ID[pos1Sav] == 21) 	    aveColIn *= 1./8.;
    else if (abs(ID[pos1Sav]) <= 6) aveColIn *= 1./3.;
    if (ID[pos2Sav] == 21) 	    aveColIn *= 1./8.;
    else if (abs(ID[pos2Sav]) <= 6) aveColIn *= 1./3.;    
    
    if (isQCD) {
      // All-Gluon
      if (nGlue == nPartons) {
	allGlue = true;
	for (int i = 0; i < nPartons; i++) {
	  hel.push_back(h[i]);
	  ids.push_back(ID[i]);
	  p.push_back(k[i]);
	}
	diffGlue1 = -1;
	diffGlue2 = -1;
	if (mhvPlus) {
	  // Find the negative helicity gluons 
	  for (int n = 0; n < nPartons; n++) {
	    if (hel[n] < 0) {
	      diffGlue1 = n;
	      break;
	    }
	  }
	  for (int n = diffGlue1 + 1; n < nPartons; n++) {
	    if (hel[n] < 0) {
	      diffGlue2 = n;
	      break;
	    }
	  }
	}
	else if (mhvMinus) {
	  // Find the positive helicity gluons 
	  for (int n = 0; n < nPartons; n++) {
	    if (hel[n] > 0) {
	      diffGlue1 = n;
	      break;
	    }
	  }
	  for (int n = diffGlue1 + 1; n < nPartons; n++) {
	    if (hel[n] > 0) {
	      diffGlue2 = n;
	      break;
	    }
	  }
	}
      }
      else if (nQuark == 2) {
	twoQuark = true;
	if (isOrdered) {
	  vector<int> iQs;
	  for (int i = 0; i < nPartons; i++) {
	    hel.push_back(h[i]);
	    ids.push_back(ID[i]);
	    p.push_back(k[i]);
	  }
	  // If same helicity quarks return 0
	  if (hel[0]*hel[nPartons-1] > 0) {
	    fermHelConserved = false;
	    if (verbose>=4) printOut("Vincia::MHVamplitudes::calcSpinorProds",
	      "Helicity must be conserved along a fermion line");
	    return false;
	  }
	}
	else {
	  // Set default order for LC
	  vector<int> newOrder = orderParticles(h, ID);
	  if (!fermHelConserved) return false;
	  for (int i = 0; i < nPartons; i++) {
	    hel.push_back(h[newOrder[i]]);
	    ids.push_back(ID[newOrder[i]]);
	    p.push_back(k[newOrder[i]]);
	  }
	}
	
	diffGlue1 = -1;	
	// Find opposite helicity gluon, needed to calculate numerator
	if (mhvPlus) {
	  for (int n = 1; n < nPartons - 1; n++) {
	    if (hel[n] < 0) {
	      diffGlue1 = n;
	      break;
	    }
	  }
	}
	else if (mhvMinus) {
	  for (int n = 1; n < nPartons - 1; n++) {
	    if (hel[n] > 0) {
	      diffGlue1 = n;
	      break;
	    }
	  }
	}

      // End if (2 Quark)
      }
      else if (nQuark == 4) {
	fourQuark = true;
	
	if (isOrdered) {
	  for (int i = 0; i < nPartons; i++) {
	    hel.push_back(h[i]);
	    ids.push_back(ID[i]);
	    p.push_back(k[i]);
	  }
	  if (ids[2] == ids[0]) sameFlav = true;
	  if (hel[0]*hel[nPartons-1] > 0 || hel[1]*hel[2]>0) {
	    fermHelConserved = false;
	    if (verbose>=4) printOut("Vincia::MHVamplitudes::calcSpinorProds",
	      "Helicity must be conserved along a fermion line");
	    return false;
	  }
	}
	else {
	  // Set default order for LC
	  vector<int> newOrder = orderParticles(h, ID);
	  if (!fermHelConserved) return false;
	  for (int i = 0; i < nPartons; i++) {
	    hel.push_back(h[newOrder[i]]);
	    ids.push_back(ID[newOrder[i]]);
	    p.push_back(k[newOrder[i]]);
	  }
	}
		
	// 0 gluons
	if (nGlue == 0) {
	  // For ease of code reading
	  q1    = 0;
	  qBar2 = 1;
	  q2    = 2;
	  qBar1 = 3;
	  // Calculate amplitude(s)
	  complex pPBar1 = spinorProdPos(p[q1],p[qBar1]);
	  complex qQBar1 = spinorProdPos(p[q2],p[qBar2]);
	  if (pPBar1 == zero || qQBar1 == zero) {
	    if (verbose>=1) printOut("Vincia::MHVamplitudes::calcSpinorProds",
	      "Singularity encountered");
	    Error = true;
	    return false;
	  }
	  else {
	    // Calculate A_0(hq1,hq2) = A_0(hq1,hq2,hg)
	    if (hel[q1] == 1.) {
	      // (hq1,hq2) = (+,+)
	      if (hel[q2] == 1.) {
		prefact = pow(spinorProdPos(p[qBar2],p[qBar1]), 2.) 
		/(pPBar1 * qQBar1);
	      }
	      // (hq1,hq2) = (+,-)
	      else {
		prefact = pow(spinorProdPos(p[qBar2],p[q2]), 2.) 
		/(pPBar1 * qQBar1);
	      }
	    }
	    else {
	    // (hq1,hq2) = (-,+)
	      if (hel[q2] == 1.) {
		prefact = pow(spinorProdPos(p[q1],p[qBar2]), 2.) 
		/(pPBar1 * qQBar1);
	      }
	      // (hq1,hq2) = (-,-)
	      else {
		prefact = pow(spinorProdPos(p[q1],p[q2]), 2.) 
		/(pPBar1 * qQBar1);
	      }
	    }
	  }  
	  if (sameFlav) {
	    // Check helicity conservation along fermion line
	    if (hel[q2] != hel[q1]) prefact2 = zero;
	    else {
	      // swap q1, q2
	      complex pPBar2 = spinorProdPos(p[q2],p[qBar1]);
	      complex qQBar2 = spinorProdPos(p[q1],p[qBar2]);
	      if (pPBar2 == zero || qQBar2 == zero) {
		if (verbose>=1) printOut("Vincia::MHVamplitudes::calcSpinorProds",
		  "Singularity encountered");
		Error = true;
		return false;
	      }
	      else {
		// Calculate A_0(hq1,hq2) = A_0(hq1,hq2,hg)
		if (hel[q2] == 1.) {
		// (hq1,hq2) = (+,+)
		  prefact2 = pow(spinorProdPos(p[qBar1],p[qBar2]), 2.) 
		  /(pPBar2 * qQBar2);
		}
		else {
		 // (hq1,hq2) = (-,-)
		 prefact2 = pow(spinorProdPos(p[q2],p[q1]), 2.) 
		 /(pPBar2 * qQBar2);
		}
	      }
	    }
	  }
	}
	// 1 gluon
	else if (nGlue == 1) {
	  // Store quark positions for ease of reading code
	  q1    = 0;
	  qBar2 = 1;
	  q2    = 2;
	  qBar1 = 4;
	}  
      // End 4 Quarks
      }
      else {
	Error = true;
	if (verbose >= 1) printOut("Vincia::MHVamplitudes::calcSpinorProds",
		      "Wrong number of quark pairs for MHV QCD amplitude");
	return false;
      }
    
    // End isQCD
    }
    
    else if (isWeak) {
      if (isOrdered) {
	for (int i = 0; i < nPartons; i++) {
	  hel.push_back(h[i]);
	  ids.push_back(ID[i]);
	  p.push_back(k[i]);
	}
	// Check helicity conservation along a fermion line
	if (hel[0]*hel[1]>0 || hel[2]*hel[nPartons-1]>0) {
	  fermHelConserved = false;
	  if (verbose>=4) printOut("Vincia::MHVamplitudes::calcSpinorProds",
	    "Helicity not conserved along a fermion line.");
	  return false;
	}
      }
      else {
	// Set default order for LC
	vector<int> newOrder = orderParticles(h, ID);
	if (!fermHelConserved) return false;
	for (int i = 0; i < nPartons; i++) {
	  hel.push_back(h[newOrder[i]]);
	  ids.push_back(ID[newOrder[i]]);
	  p.push_back(k[newOrder[i]]);
	}
      }
      
      if (nGlue == 0) {
	// (0,1,2,3) = (l,lBar,q,qBar)
	double sLLBar = norm(spinorProdPos(p[0],p[1]));
	complex qQBar = spinorProdPos(p[2],p[3]); 
	if (qQBar == zero) {
	  cout << " MHV: Singularity qQBar = zero" << endl;          
	  Error = true;
	  return false;
	}
	else {
	  // Calculate a_0(hl,hq)/<qBar q>
	  if (hel[2] > 0) {
	    // (hl,hq)=(+,+)
	    if (hel[0] > 0) prefact = pow(spinorProdPos(p[1],p[3]),2.)
	      *spinorProdNeg(p[1],p[0])/qQBar;
	    // (hl,hq)=(-,+)
	    else prefact = pow(spinorProdPos(p[0],p[3]),2.)
	      *spinorProdNeg(p[1],p[0])/qQBar;
	  }
	  else {
	  // (hl,hq)=(+,-)
	    if (hel[0] > 0) prefact = pow(spinorProdPos(p[1],p[2]),2.)
	      *spinorProdNeg(p[1],p[0])/qQBar;
	    // (hl,hq)=(-,-)
	    else prefact = pow(spinorProdPos(p[0],p[2]),2.)
	      *spinorProdNeg(p[1],p[0])/qQBar;
	  }
	}
	// Check for type of interaction
	// TODO W Boson needs testing
	if (ids[0] ==  -ids[1] - 1 || ids[0] ==  -ids[1] + 1) {
	  if (ids[2] != -ids[3] - 1 && ids[2] !=  -ids[3] + 1) {
	    Error = true;
	    if (verbose>=4) printOut("Vincia::MHVamplitudes::calcSpinorProds",
			"Unknown weak process");
	    return false;
	  }
	  if (hel[0] != -1 || hel[2] != -1) {
	    nonChiralW = true;
	    return false;
	  }
	  if (ids[2] == 1) {
	    if (ids[3] == -2) prefact *= gWflSq*Vud;
	    else if (ids[3] == -4) prefact *= gWflSq *Vcd;
	  }
	  else if (ids[2] == 2) {
	    if (ids[3] == -1) prefact *= gWflSq*Vud;
	    else if (ids[3] == -3) prefact *= gWflSq *Vus;
	  }
	  else if (ids[2] == 3) {
	    if (ids[3] == -2) prefact *= gWflSq*Vus;
	    else if (ids[3] == -4) prefact *= gWflSq *Vcs;
	  }
	  else if (ids[2] == 4) {
	    if (ids[3] == -1) prefact *= gWflSq*Vcd;
	    else if (ids[3] == -3) prefact *= gWflSq *Vcs;
	  }
	  prefact /= (sLLBar + propW);
	}
	// Z Boson/Photon exchange. 
	else if (ids[0] == -ids[1]) {
	  if (hel[2] == 1) {
	    if (hel[0] == 1) {
	      // (hl,hq)=(+,+)
	      // Up-type quark
	      if (ids[2] % 2 == 0) 
		prefact *= (gZlr*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	      // Down-type quark
	      else prefact *= (gZlr*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	    }
	    // (hl,hq)=(-,+)
	    else {
	      // Up-type quark
	      if (ids[2] % 2 == 0) 
		prefact *= (gZll*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	      // Down-type quark
	      else prefact *= (gZll*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	    }
	  }
	  else {
	    // (hl,hq)=(+,-)
	    if (hel[0] == 1) {
	      // Up-type quark
	      if (ids[2] % 2 == 0) 
		prefact *= (gZlr*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	      // Down-type quark
	      else 
		prefact *= (gZlr*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	    }
	    // (hl,hq)=(-,-)
	    else {
	      // Up-type quark
	      if (ids[2] % 2 == 0) 
		prefact *= (gZll*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	      // Down-type quark
	      else prefact *= (gZll*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	    }
	  }
	}
	else {
	  cout << "Wrong particle configuration for weak force" << endl;
	  Error = true;
	  return false;
	}
      // end if (nGlue == 0)
      }
      else { 
	// For easy code reading
	l1    = 0;
	lBar1 = 1;
	q1    = 2;
	qBar1 = nPartons - 1;
      }
    // End (isWeak)
    }
    
    if (verbose >= 6) {
      printOut("Vincia::mhvAmplitudes::calcSpinorProds","Printing particle information");
      if (!isOrdered) cout << " After ordering by MHV.cc the particles are: ";
      else cout << " The particles were not re-ordered (isOrdered = true). They are: "; 
      cout << "\n Particle IDs are:        ";
      for (int i = 0; i < nPartons; i++) cout << ids[i] << ", ";
      cout << endl;
      cout << " Particle helicities are: ";
      for (int i = 0; i < nPartons; i++) cout << hel[i] << ", ";
      cout << endl;
      cout << " Particle momenta are: " << endl;
      for (int i = 0; i < nPartons; i++) cout << p[i];
    }
	
    
    if (nGlue != 0) {
      // Create array of all spinor products to be used. 
      if (mhvPlus) {
        for (int i = 0; i < nPartons; i++) {
	  for (int j = i + 1; j < nPartons; j++) {
            spBag[i][j] = spinorProdPos(p[i], p[j]);
            spBag[j][i] = - spBag[i][j];
	    if (spBag[i][j] == zero) {
	      if (!isWeak) {
		cout << " MHV (plus): Singularity " <<i<<" "<<j
                     << " (!isWeak) "<<endl;
                cout << i << p[i] ; 
                cout << j << p[j] ;
		Error = true;
		return false;
	      }
	      else if (i < nPartons - 2 && j < nPartons - 2) {
		cout << " MHV (plus): Singularity " <<i<<" "<<j
                     <<" (isWeak)"<< endl;
                cout << i << p[i] ; 
                cout << j << p[j] ;
		Error = true;
		return false;
	      }
	    }
	  }
        }
        // Calculate numerator for 2 quark 
        if (twoQuark) {
	  if (hel[0] < 0) num = pow(spBag[0][diffGlue1], 3.0) * 
	    spBag[nPartons - 1][diffGlue1];
	  else num = pow(spBag[nPartons - 1][diffGlue1], 3.0) * 
	      spBag[0][diffGlue1];
	}
        // Calculate prefactors for 4 quark and 2 quark 2 leptons
        if (fourQuark) {
	  if (hel[q1] == 1) {
	    // (hq1,hq2,hg) = (+,+,+)
	    if (hel[q2] == 1) {
	      prefact = pow(spBag[qBar1][qBar2], 2.) 
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
    	    // (hq1,hq2,hg) = (+,-,+)
	    else {
	      prefact = pow(spBag[qBar1][q2], 2.) 
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
	  }
	  else {
    	    // (hq1,hq2,hg) = (-,+,+)
	    if (hel[q2] == 1) {
	      prefact = pow(spBag[q1][qBar2], 2.) 
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
    	    // (hq1,hq2,hg) = (-,-,+)
	    else {
	      prefact = pow(spBag[q1][q2], 2.)
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
	  }
	  if (sameFlav) {
	    if (hel[q1] != hel[q2]) prefact2 = zero;
	    else {
	      // (hq2,hq1,hg) = (+,+,+)
	      if (hel[q2] == 1) {
		prefact2 = pow(spBag[qBar1][qBar2], 2.) 
		/(spBag[q2][qBar1]*spBag[q1][qBar2]);
	      }
	      // (hq2,hq1,hg) = (-,-,+)
	      else {
		prefact2 = pow(spBag[q2][q1], 2.)
		/(spBag[q2][qBar1]*spBag[q1][qBar2]);
	      }
	    }
	  }
	}
	else if (isWeak) {
	  double sLLBar = norm(spBag[l1][lBar1]);
	  if (hel[l1] > 0) {
	    // (hl,hq,hg)=(+,+,+)
	    if (hel[q1] > 0) prefact = pow(spBag[lBar1][qBar1],2.)
	      *conj(spBag[l1][lBar1]);
	    // (hl,hq,hg)=(+,-,+)
	    else prefact = pow(spBag[lBar1][q1],2.)*conj(spBag[l1][lBar1]);
	  }
	  else {
	  // (hl,hq,hg)=(-,+,+)
	    if (hel[q1] > 0) prefact = pow(spBag[l1][qBar1],2.)
	      *conj(spBag[l1][lBar1]);
	    // (hl,hq,hg)=(-,-,+)
	    else prefact = pow(spBag[l1][q1],2.)*conj(spBag[l1][lBar1]);
	  }
	  // Check for type of interaction
	  // TODO W Boson not yet tested 
	  if (ids[lBar1] ==  -ids[l1] - 1 || ids[lBar1] ==  -ids[l1] + 1) {
	    if (ids[qBar1] !=  -ids[q1] - 1 && ids[qBar1] !=  -ids[q1] + 1) {
	      Error = true;
	      if (verbose>=4) printOut("Vincia::MHVamplitudes::calcSpinorProds",
			"Unknown weak process");
	      return false;
	    }
	    // W decays are chiral
	    if (hel[l1] != -1 || hel[q1] != -1) {
	      nonChiralW = true;
	      return false;
	    }
	    if (ids[q1] == 1) {
	      if (ids[qBar1] == -2) prefact *= gWflSq*Vud;
	      else if (ids[qBar1] == -4) prefact *= gWflSq *Vcd;
	    }
	    else if (ids[q1] == 2) {
	      if (ids[qBar1] == -1) prefact *= gWflSq*Vud;
	      else if (ids[qBar1] == -3) prefact *= gWflSq *Vus;
	    }
	    else if (ids[q1] == 3) {
	      if (ids[qBar1] == -2) prefact *= gWflSq*Vus;
	      else if (ids[qBar1] == -4) prefact *= gWflSq *Vcs;
	    }
	    else if (ids[q1] == 4) {
	      if (ids[qBar1] == -1) prefact *= gWflSq*Vcd;
	      else if (ids[qBar1] == -3) prefact *= gWflSq *Vcs;
	    }
	    prefact /= (sLLBar + propW);
	  }
	  // Z Boson/photon exchange
	  else if (ids[lBar1] == -ids[l1]) {
	    if (hel[q1] == 1) {
	      if (hel[l1] == 1) {
		// Up-type quark
		if (ids[q1] % 2 == 0)
		  prefact *= (gZlr*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZlr*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	      else {
		// Up-type quark
		if (ids[q1] % 2 == 0)  
		  prefact *= (gZll*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZll*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	    }
	    else {
	      if (hel[l1] == 1) {
		// Up-type quark
		if (ids[q1] % 2 == 0) 
		  prefact *= (gZlr*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZlr*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	      else {
		// Up-type quark
		if (ids[q1] % 2 == 0) 
		  prefact *= (gZll*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZll*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	    }
	  }
	  else {
	    cout << "Wrong particle configuration for weak force" << endl;
	    return false;
	  }
	}
      }
      
      // If mostly negative helicities, create correct spinor bag
      else if (mhvMinus) {
        for (int i = 0; i < nPartons; i++) {
	  for (int j = i + 1; j < nPartons; j++) {
	    spBag[i][j] = spinorProdNeg(p[i], p[j]);
	    spBag[j][i] = - spBag[i][j];
	    if (spBag[i][j] == zero) {
	      if (!isWeak) {
		cout << " MHV (minus): Singularity " <<i<<" "<<j<<" (!isWeak)"
                     << endl;
                cout << i << p[i] ; 
                cout << j << p[j] ;
		Error = true;
		return false;
	      }
	      else if (i < nPartons - 2 && j < nPartons - 2) {
		cout << " MHV (minus): Singularity " <<i<<" "<<j<<" (isWeak)"
                     << endl;
                cout << i << p[i] ; 
                cout << j << p[j] ;
		Error = true;
		return false;
	      }
	    }
          }
        }
        // Calculate numerator for 2 quark
        if (twoQuark) {
	  if (hel[0] > 0) num = pow(spBag[0][diffGlue1], 3.0) * 
	    spBag[nPartons - 1][diffGlue1];
	  else num = pow(spBag[nPartons - 1][diffGlue1], 3.0) * 
	    spBag[0][diffGlue1];
	}
        // Calculate prefactors for 4 quark and 2 quark 2 leptons
        if (fourQuark) {
	  if (hel[q1] == 1.) {
	    // (hq1,hq2,hg) = (+,+,-)
	    if (hel[q2] == 1.) {
	      prefact = pow(spBag[q1][q2], 2.) 
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
    	    // (hq1,hq2,hg) = (+,-,-)
	    else {
	      prefact = pow(spBag[q1][qBar2], 2.) 
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
	  }
	  else {
    	    // (hq1,hq2,hg) = (-,+,-)
	    if (hel[q2] == 1.) {
	      prefact = pow(spBag[qBar1][q2], 2.) 
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
    	    // (hq1,hq2,hg) = (-,-,-)
	    else {
	      prefact = pow(spBag[qBar1][qBar2], 2.) 
	      /(spBag[q1][qBar1]*spBag[q2][qBar2]);
	    }
	  }
	  if (sameFlav) {
	    if (hel[q1] != hel[q2]) prefact2 = zero;
	    else {
	      // (hq2,hq1,hg) = (+,+,-)
	      if (hel[q2] == 1) {
		prefact2 = pow(spBag[q2][q1], 2.)
		/(spBag[q2][qBar1]*spBag[q1][qBar2]);
	      }
	      // (hq2,hq1,hg) = (-,-,+)
	      else {
		prefact2 = pow(spBag[qBar1][qBar2], 2.) 
		/(spBag[q2][qBar1]*spBag[q1][qBar2]);
	      }
	    }
	  }
	}
	else if (isWeak) {
	  double sLLBar = norm(spBag[l1][lBar1]);
	  if (hel[l1] > 0) {
	    // (hl,hq,hg)=(+,+,-)
	    if (hel[q1] > 0) prefact = pow(spBag[l1][q1],2.)
	      *conj(spBag[l1][lBar1]);
	    // (hl,hq,hg)=(+,-,-)
	    else prefact = pow(spBag[l1][qBar1],2.)*conj(spBag[l1][lBar1]);
	  }
	  else {
	  // (hl,hq,hg)=(-,+,-)
	    if (hel[q1] > 0) prefact = pow(spBag[lBar1][q1],2.)
	      *conj(spBag[l1][lBar1]);
	    // (hl,hq,hg)=(-,-,+)
	    else prefact = pow(spBag[lBar1][qBar1],2.)*conj(spBag[l1][lBar1]);
	  }
	  // Check for type of interaction
	  // TODO W Boson not yet tested
	  if (ids[lBar1] ==  -ids[l1] - 1 || ids[lBar1] ==  -ids[l1] + 1) {
	    if (ids[qBar1] !=  -ids[q1] - 1 && ids[qBar1] !=  -ids[q1] + 1) {
	      Error = true;
	      if (verbose>=4) printOut("Vincia::MHVamplitudes::calcSpinorProds",
			"Unknown weak process");
	      return false;
	    }
	    if (hel[l1] != -1 || hel[q1] != -1) {
	      nonChiralW = true;
	      return false;
	    }
	    if (ids[q1] == 1) {
	      if (ids[qBar1] == -2) prefact *= gWflSq*Vud;
	      else if (ids[qBar1] == -4) prefact *= gWflSq *Vcd;
	    }
	    else if (ids[q1] == 2) {
	      if (ids[qBar1] == -1) prefact *= gWflSq*Vud;
	      else if (ids[qBar1] == -3) prefact *= gWflSq *Vus;
	    }
	    else if (ids[q1] == 3) {
	      if (ids[qBar1] == -2) prefact *= gWflSq*Vus;
	      else if (ids[qBar1] == -4) prefact *= gWflSq *Vcs;
	    }
	    else if (ids[q1] == 4) {
	      if (ids[qBar1] == -1) prefact *= gWflSq*Vcd;
	      else if (ids[qBar1] == -3) prefact *= gWflSq *Vcs;
	    }
	    prefact /= (sLLBar + propW);
	  }
	  // Z Boson/photon exchange
	  else if (ids[lBar1] == -ids[l1]) {
	    if (hel[q1] == 1) {
	      if (hel[l1] == 1) {
		// Up-type quark
		if (ids[q1] % 2 == 0) 
		  prefact *= (gZlr*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZlr*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	      else {
		// Up-type quark
		if (ids[q1] % 2 == 0) 
		  prefact *= (gZll*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZll*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	    }
	    else {
	      if (hel[l1] == 1) {
		// Up-type quark
		if (ids[q1] % 2 == 0) 
		  prefact *= (gZlr*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZlr*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	      else {
		// Up-type quark
		if (ids[q1] % 2 == 0) 
		  prefact *= (gZll*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
		// Down-type quark
		else prefact *= (gZll*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	      }
	    }
	  }
	  else {
	    cout << " Wrong particle configuration for weak force" << endl;
	    return false;
	  }
	}
      }
      
      // Not MHV
      else {
        Error = true;
	if (verbose >= 1) {
	  printOut("Vincia::MHVamplitudes::calcSpinorProds",
		   "Not in MHV configuration. Check helicities");
	  if (verbose >= 4) {
	    cout << " Printing helicities" << endl;
	    for (int i = 0; i < nPartons; i++) cout << " hel["<<i<<"] = " << hel[i];
	  }
	}
	return false;
      }
      
      if (verbose >= 6) {
	cout << " Printing array of all spinor products. Entry i,j corresponds"
	     << " to <ij> (or [ij]). " << endl;
	printSPbag();
      }
    }
    
    return true;
    
  // End calcSpinorProds function
  }

  double MHVamplitudes::ME2() {
    if (Error) return -1.;
    else if (!fermHelConserved || nonChiralW) return 0.;
    if (colorDepth != 0 && colorDepth != 1 && colorDepth != 2){ 
      Error = true;
      printOut("Vincia::MHVamplitudes::ME2",
	       "Please set colorDepth to 0 (leading colour), 1 (Vincia colour), or 2 (full colour).");
      ampBag[0] = pie;
      return -1;
    }
    
    double matrixSquare;
    complex denom;
    
    if (allGlue) {
      num = pow(spBag[diffGlue1][diffGlue2], 4.);
      
      // Calculate LC ME2
      if (colorDepth == 0) {
	denom = spBag[nPartons - 1][0];
	for (int i = 0; i < nPartons - 1; i++) denom *= spBag[i][i+1];
	if 	(nPartons == 4) return norm(num/denom)*LCAG4*aveColIn;
	else if (nPartons == 5) return norm(num/denom)*LCAG5*aveColIn;
	else if (nPartons == 6) return norm(num/denom)*LCAG6*aveColIn;
	else if (nPartons == 7) return norm(num/denom)*LCAG7*aveColIn;
      }
      
      else if (colorDepth == 1) {
	if (nPartons == 4 || nPartons == 5) {
	  denom = spBag[nPartons - 1][0];
	  for (int i = 0; i < nPartons - 1; i++) denom *= spBag[i][i+1];
	  if (nPartons == 4) return norm(num/denom)*FCAG4*aveColIn;
	  else 		     return norm(num/denom)*FCAG5*aveColIn;
	}
	else if (nPartons == 6) {
	  // Create list of gluons to permute
	  vector<int> list;
	  list.resize(nPartons);
	  for (int i = 0; i< nPartons; i++) list[i] = i;
	  
	  // Calculate Amplitudes and store in ampBag
	  int ampBagNum = 0;
	  do {
	    // Only want (n-1)! non-cyclic permutations
	    if (list[0] == 1) break;
	    else {
	      // Calculate the denominator for each amplitude.
	      denom = spBag[list[nPartons - 1]][0];
	      for (int i = 0; i < nPartons -1 ; i++) 
		denom *= spBag[list[i]][list[i+1]];
	      // Calculate amplitude
	      ampBag[ampBagNum] = num/denom;
	      ampBagNum += 1;
	    }
	  }
	  while (std::next_permutation(list.begin(),list.end()));
	
	  double LCSum = 0;
	  int amp1, amp2, amp3;
	  for (int i = 0; i < 120; i++) LCSum += norm(ampBag[i]);
	  // Reset list
	  for (int i = 0; i < nPartons; i++) list[i] = i;
	  // ampBagNum = sequential ampBag index
	  ampBagNum = 0;
	  double NLC = 0;
	  do {
	    // Only want (n-1)! non-cyclic permutations
	    if (list[0] == 1) break;
	    else {
	      amp1 = getAmpNum(list[2], list[4], list[1], list[5], list[3]);
	      amp2 = getAmpNum(list[4], list[2], list[5], list[1], list[3]);
	      amp3 = getAmpNum(list[2], list[5], list[3], list[1], list[4]);
	      NLC += real(conj(ampBag[ampBagNum]) * 
		    (ampBag[amp1] + ampBag[amp2] + ampBag[amp3]));
	      ampBagNum += 1;
	    }
	  }
	  while (std::next_permutation(list.begin(), list.end()));
	  return  norm(ampBag[0])*FCAG6*(LCSum + 2.*NLC/9.)*aveColIn/LCSum;
	}
	else if (nPartons == 7) {
	  if (verbose>=2) printOut("Vincia::MHVamplitudes::ME2",
		"Warning!! Cannot calculate Vincia-colour 7-gluon amplitude. Calculating leading colour instead");
	  return norm(num/denom)*LCAG7*aveColIn;
	}
	else {
	  cout << " Cannot calculate all-gluon Vincia-Colour matrix element "
	       << " for 8 or more particles. " << endl;
	  return -1;
	}
      // End if (colorDepth == 1)
      }
	
      // Calculate FC ME2
      else if (colorDepth == 2) { 
	// Create list of gluons to permute
	vector<int> list;
	list.resize(nPartons);
	for (int i = 0; i< nPartons; i++) list[i] = i;
	  
	// Calculate Amplitudes and store in ampBag
	int ampBagNum = 0;
	do {
	  // Only want (n-1)! non-cyclic permutations
	  if (list[0] == 1) break;
	  else {
	    // Calculate the denominator for each amplitude.
	    denom = spBag[list[nPartons - 1]][0];
	    for (int i = 0; i < nPartons -1 ; i++) 
	      denom *= spBag[list[i]][list[i+1]];
            // Calculate amplitude
            ampBag[ampBagNum] = num/denom;
	    ampBagNum += 1;
	  }
	}
	while (std::next_permutation(list.begin(),list.end()));
	     
	// Return the ME2
        if (nPartons == 4) {
	  matrixSquare = 0;
	  for (int i = 0; i < 6; i++) matrixSquare += norm(ampBag[i]);
	  return matrixSquare * FCAG4 * aveColIn; 
	}
	else if (nPartons == 5) {
	  matrixSquare = 0;
	  for (int i = 0; i < 24; i++) matrixSquare += norm(ampBag[i]);
	  return matrixSquare * FCAG5 * aveColIn; 
	}
	else if (nPartons == 6) {
	  matrixSquare = 0;
	  int amp1, amp2, amp3;
	  for (int i = 0; i < 120; i++) matrixSquare += norm(ampBag[i]);
	  // Reset list
	  for (int i = 0; i < nPartons; i++) list[i] = i;
	  // ampBagNum = sequential ampBag index, NLC = next-to-leading colour factor
	  ampBagNum = 0;
	  double NLC = 0;
	  do {
	    // Only want (n-1)! non-cyclic permutations
	    if (list[0] == 1) break;
	    else {
	      amp1 = getAmpNum(list[2], list[4], list[1], list[5], list[3]);
	      amp2 = getAmpNum(list[4], list[2], list[5], list[1], list[3]);
	      amp3 = getAmpNum(list[2], list[5], list[3], list[1], list[4]);
	      NLC += real(conj(ampBag[ampBagNum]) * 
		    (ampBag[amp1] + ampBag[amp2] + ampBag[amp3]));
	      ampBagNum += 1;
	    }
	  }
	  while (std::next_permutation(list.begin(), list.end()));
	  return  (matrixSquare + 2./9 * NLC) * FCAG6 * aveColIn;
	}
	else if (nPartons == 7) {
	  if (verbose>=2) printOut("Vincia::MHVamplitudes::ME2",
		"Warning!! Cannot calculate full-colour 7-gluon amplitude. Calculating leading colour instead");
	  return norm(num/denom)*LCAG7*aveColIn;
	}

	else {
          cout << "\nHave not yet added 7 or more particles" << endl;
          return -1;
	}
	  
      // End if (colorDepth == 2)  
      }
	
    // End if (allGlue)
    }
      
    else if (twoQuark) {
      // Calculate LC ME2
      if (colorDepth == 0) {
	// Calculate denominator. First calculate <qbar q>
        denom = spBag[nPartons - 1][0];
        for (int i = 0; i < nPartons - 1; i++) denom *= spBag[i][i+1];
	// Calculate amplitude
        if 	(nPartons == 4) return norm(num/denom) * LC2Q4 * aveColIn;
	else if (nPartons == 5) return norm(num/denom) * LC2Q5 * aveColIn; 
	else if (nPartons == 6) return norm(num/denom) * LC2Q6 * aveColIn;
	else if (nPartons == 7) return norm(num/denom) * LC2Q7 * aveColIn;    
      }
	
      // Calculate FC or VC ME2
      else if (colorDepth >= 1) {
	double LCSum;
	// Create list to permute gluons 
	vector<int> list;
	list.resize(nGlue);
	for (int i = 0; i < nGlue; i++) list[i] = i + 1;
	// Calculate Amplitudes and store in ampBag
	// ampBagNum = ampBag index
	int ampBagNum = 0;
	do {
	  // Next calculate <q 1>*<qbar q>
	  denom = spBag[nPartons - 1][0] * spBag[0][list[0]];
	  // Next calculate <n qbar>
	  denom *= spBag[list[nGlue-1]][nPartons-1];
          // Calculate rest of denominator
          for (int i = 0; i < nGlue -1 ; i++) {
	    denom *= spBag[list[i]][list[i+1]];
	  }
          // Calculate amplitude
          ampBag[ampBagNum] = num/denom;
	  ampBagNum += 1;
	}
	while (std::next_permutation(list.begin(), list.end()));
	  
	// Sum together for FC ME2
	if (nPartons == 4) {
          matrixSquare = 0.;
          for (int i = 0; i < 2; i++) matrixSquare += norm(ampBag[i]);
	  LCSum = matrixSquare;
	  matrixSquare *= LC2Q4; 
	  for (int i = 0; i < 2; i++) {
            for (int n = 0; n < i; n++) {
  	      matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
	      imag(ampBag[i])*imag(ampBag[n])) * qQBar2g[i][n]; 
            }
          }
          if (colorDepth == 1) return norm(ampBag[0])*matrixSquare/LCSum*aveColIn;
// 	  if (colorDepth == 3) return LCSum;
          else return matrixSquare*aveColIn;
        }
        else if (nPartons == 5) {
          matrixSquare = 0.;
          for (int i = 0; i < 6; i++) matrixSquare += norm(ampBag[i]);
	  if (colorDepth == 1) LCSum = matrixSquare;
	  matrixSquare *= LC2Q5;
	  for (int i = 0; i < 6; i++) {
            for (int n = 0; n < i; n++) {
  	      matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
	      imag(ampBag[i])*imag(ampBag[n])) * qQBar3g[i][n]; 
            }
          }
          if (colorDepth == 1) return norm(ampBag[0])*matrixSquare*aveColIn/LCSum;
          else return matrixSquare*aveColIn;
        }
        else if (nPartons == 6) {
          matrixSquare = 0.;
          for (int i = 0; i < 24; i++) matrixSquare += norm(ampBag[i]);
	  if (colorDepth == 1) LCSum = matrixSquare;
	  matrixSquare *= LC2Q6;
	  for (int i = 0; i < 24; i++) {
            for (int n = 0; n < i; n++) {
              matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
	      imag(ampBag[i])*imag(ampBag[n])) * qQBar4g[i][n]; 
            }
          }
          if (colorDepth == 1) return norm(ampBag[0])*matrixSquare*aveColIn/LCSum;
          else return matrixSquare*aveColIn;
        }
        else if (nPartons == 7) {
	  matrixSquare = 0.;
	  for (int i = 0; i < 120; i++) matrixSquare += norm(ampBag[i]);
	  if (colorDepth == 1) LCSum = matrixSquare;
	  matrixSquare *= LC2Q7;
	  for (int i = 0; i < 120; i++) {
            for (int n = 0; n < i; n++) {
              matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
	      imag(ampBag[i])*imag(ampBag[n])) * qQBar5g[i][n]; 
            }
          }
	  if (colorDepth == 1) return norm(ampBag[0])*matrixSquare*aveColIn/LCSum;
          else return matrixSquare*aveColIn;
        }
        else {
	  cout << " Maximum 7 partons for q qBar pluse gluons " << endl;
	  return -1.;
	}
	
      // End if (colorDepth >= 1)
      }
	
    // End if (twoQuark)
    }
      
    else if (fourQuark) {
      if (nGlue == 0) {
	if (prefact2 == zero || colorDepth == 0) 
	  return LC4Q*norm(prefact)*aveColIn;
	// sameFlav
	else if (colorDepth == 1) 
	  return aveColIn*norm(prefact)*( LC4Q + 16./3.*(real(prefact)*
	    real(prefact2) + imag(prefact)*imag(prefact2))/(norm(prefact) + 
	      norm(prefact2)) );
	else 
	 return aveColIn*(FC4Q*(norm(prefact) + norm(prefact2)) +16./3.* 
	   (real(prefact)*real(prefact2) + imag(prefact)*imag(prefact2)));
      }
      else if (nGlue == 1) {
	
	if (colorDepth == 0) {
	  ampBag[0] = spBag[q2][qBar1]/(spBag[q2][3] * spBag[3][qBar1]);
	  return aveColIn*LC4Q1*norm(prefact*ampBag[0]);
	}
	else if (colorDepth >= 1) {
	  ampBag[0] = spBag[q1][qBar2]/(spBag[q1][3] * spBag[3][qBar2]);
	  ampBag[1] = spBag[q2][qBar1]/(spBag[q2][3] * spBag[3][qBar1]);
	  ampBag[2] = spBag[q1][qBar1]/(spBag[q1][3] * spBag[3][qBar1]);
	  ampBag[3] = spBag[q2][qBar2]/(spBag[q2][3] * spBag[3][qBar2]);
	  double amp01Sq = norm(ampBag[0]) + norm(ampBag[1]);
	  double amp23Sq = norm(ampBag[2]) + norm(ampBag[3]);
	  if (prefact2 == zero) {
	    // return full colour squared amp
	    matrixSquare = 0;
	    // interference terms first
	    for (int i = 0; i < 2; i++) {
	      for (int j = 2; j < 4; j++) {
		matrixSquare += real(ampBag[i])*real(ampBag[j])
		  +imag(ampBag[i])*imag(ampBag[j]);
	      }
	    }
	    matrixSquare *= -16./3;
	    // Leading colour
	    matrixSquare += 24.*amp01Sq;
	    // Sub-leading colour, non-interference
	    matrixSquare += 8./3*amp23Sq;
	    matrixSquare *= norm(prefact);
	    if (colorDepth == 1) 
	      return aveColIn*matrixSquare*norm(ampBag[1])/amp01Sq;
	    // if (colorDepth == 2)
	    return matrixSquare*aveColIn;
	  }
	  
	  // if (sameFlav) is equiv to if (prefact2 != zero)
	  else if (prefact2 != zero) {
	  
	    double  ampInterfere  = 0;
	    complex ampInterfere1 = zero;
// 	    complex ampInterfere2 = zero;
	    for (int i = 0; i < 2; i++) {
	      for (int j = 2; j < 4; j++) {
		ampInterfere += real(ampBag[i])*real(ampBag[j])
		  +imag(ampBag[i])*imag(ampBag[j]);
  		ampInterfere1 += ampBag[i]*conj(ampBag[j]);
// 		ampInterfere2 += conj(ampBag[i])*ampBag[j];
	      }
	    }
// 	    cout << scientific << " ampInterfere1 = " << ampInterfere1
// 	         << " ampInterfere2 = " << ampInterfere2<<endl;
// 	    cout << " prefact*conj(prefact2) = " << prefact*conj(prefact2)<<endl;
	    matrixSquare = norm(prefact)*(24*amp01Sq + 8/3*amp23Sq
	      -16/3*ampInterfere);
	    matrixSquare += norm(prefact2)*(8/3*amp01Sq + 24*amp23Sq
	      -16/3*ampInterfere);
	    matrixSquare += 16*real(prefact*conj(prefact2)*(amp01Sq + amp23Sq
	      - ampInterfere1 - 1./9.*conj(ampInterfere1)));
//     	    matrixSquare += 16*real(prefact*conj(prefact2)*(amp01Sq + amp23Sq
// 	      - real(ampInterfere1) - 1./9.*real(ampInterfere2)));
	    if (colorDepth == 1)
	      return aveColIn*matrixSquare*norm(ampBag[1])/(amp01Sq+amp23Sq);
	    // if (colorDepth == 2)
	    return matrixSquare*aveColIn;
	  } 
	
	// End if (colorDepth == 2)
	}
	
      // End if (nGlue == 1)
      }
      
      // nGlue != 0
      else {
	if (verbose >= 1)
	  printOut("Vincia::MHVamplitudes::ME",
		 "MHV amplitude with 4 quarks + >=2 gluons not yet available");
	return -1;
	
	// Remnant of old 4 quark + gluon code. Leave here for now
	/*
	if (colorDepth == 0) {
	  // Calculate denominator = <q2 g><g qBar1>
	  denom = one;
	  for (int i = 2; i < nPartons - 1; i++) denom *= spBag[i][i+1];
	  if (nPartons == 5) 
	    return LC4Q1 * aveColIn * norm(prefact*spBag[q1][qBar2]/denom);
	  else if (nPartons == 6) 
	    return LC4Q2 * aveColIn * norm(prefact*spBag[q1][qBar2]/denom);
	}*/
      }
	
    // End if (fourQuark)
    }
    
    else if (isWeak) {
      if (nGlue == 0) return LC2Q0*norm(prefact);
      else {
	if (colorDepth == 0) {
	  denom = one;
	  for (int i = 2; i < nPartons - 1; i++) 
	    denom *= spBag[i][i+1];
	  if 	  (nPartons == 5) return norm(prefact/denom) * LC2Q1 * aveColIn;
	  else if (nPartons == 6) return norm(prefact/denom) * LC2Q4 * aveColIn; 
	  else if (nPartons == 7) return norm(prefact/denom) * LC2Q5 * aveColIn;
	  else if (nPartons == 8) return norm(prefact/denom) * LC2Q6 * aveColIn;
	  else if (nPartons == 9) return norm(prefact/denom) * LC2Q7 * aveColIn;
	}
      
	else if (colorDepth >= 1) {
	  double LCSum = -1;
	  vector<int> list;
	  list.resize(nGlue);
	  for (int i = 0; i < nGlue; i++) list[i] = i + 3;
	  // Calculate Amplitudes and store in ampBag
	  int ampBagNum = 0;
	  do {
	    // First calculate <q 1><n qBar>
	    denom = spBag[q1][list[0]]*spBag[list[nGlue-1]][qBar1];
	    for (int i = 0; i < nGlue -1 ; i++) 
	      denom *= spBag[list[i]][list[i+1]];
	    // Calculate amplitude
	    ampBag[ampBagNum] = one/denom;
	    ampBagNum += 1;
	  }
	  while (std::next_permutation(list.begin(), list.end()));
	  if 	  (nPartons == 5) return norm(prefact)*norm(ampBag[0]) * LC2Q1;
	  else if (nPartons == 6) {
	    matrixSquare = 0.;
	    for (int i = 0; i < 2; i++) matrixSquare += norm(ampBag[i]);
	    if (colorDepth == 1) LCSum = matrixSquare;
	    matrixSquare *= LC2Q4; 
	    for (int i = 0; i < 2; i++) {
	      for (int n = 0; n < i; n++) {
		matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
		imag(ampBag[i])*imag(ampBag[n])) * qQBar2g[i][n]; 
	      }
	    }
	  }
	  else if (nPartons == 7) {
	    matrixSquare = 0.;
	    for (int i = 0; i < 6; i++) matrixSquare += norm(ampBag[i]);
	    if (colorDepth == 1) LCSum = matrixSquare;
	    matrixSquare *= LC2Q5;
	    for (int i = 0; i < 6; i++) {
	      for (int n = 0; n < i; n++) {
		matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
		imag(ampBag[i])*imag(ampBag[n])) * qQBar3g[i][n]; 
	      }
	    }
	  }
	  else if (nPartons == 8) {
	    matrixSquare = 0.;
	    for (int i = 0; i < 24; i++) matrixSquare += norm(ampBag[i]);
	    if (colorDepth == 1) LCSum = matrixSquare;
	    matrixSquare *= LC2Q6;
	    for (int i = 0; i < 24; i++) {
	      for (int n = 0; n < i; n++) {
		matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
		imag(ampBag[i])*imag(ampBag[n])) * qQBar4g[i][n]; 
	      }
	    }
	  }
	  else if (nPartons == 9) {
	    matrixSquare = 0.;
	    for (int i = 0; i < 120; i++) matrixSquare += norm(ampBag[i]);
	    if (colorDepth == 1) LCSum = matrixSquare;
	    matrixSquare *= LC2Q7;
	    for (int i = 0; i < 120; i++) {
	      for (int n = 0; n < i; n++) {
		matrixSquare += 2 * (real(ampBag[i]) * real(ampBag[n]) + 
		imag(ampBag[i])*imag(ampBag[n])) * qQBar5g[i][n]; 
	      }
	    }
	  }
	  else return -1;
	  if (colorDepth == 1) return norm(prefact)*norm(ampBag[0])*matrixSquare*aveColIn/LCSum;
	  else return norm(prefact) * matrixSquare * aveColIn;
	
	// End if (colorDepth >= 1)
	}
      // End if (nGlue != 0)
      }
    // End if (isWeak)
    }
    
    // If you get this far there is an error
    if (verbose>=1) printOut("Vincia::MHVamplitudes::ME2",
	     "No amplitude calculated! Unknown Error. Returning -1");
    return -1;
    
  // End ME2 function 
  }
  
//-----------------------------------------------------------------------//
  
  // Which amplitude from ampBag do I need
  int MHVamplitudes::getAmpNum(int iIn1, int iIn2, int iIn3, int iIn4, int iIn5) {
    // getAmpNum gives the position of A(iIn0,iIn1,iIn2,iIn3,iIn4,iIn5)
    // in ampBag. It is used to calculate 6-gluon amplitude
    int ampNum = (iIn1-1) * 24;
    if (iIn2 > iIn1) {
      ampNum += (iIn2-2) * 6;
      if (iIn3 > iIn4) {
	if (iIn3 > iIn5) {
	  ampNum += 4;
	  if (iIn4 > iIn5) ampNum += 1;
	}
	// if iIn3 > iIn4 and iIn3 < iIn5
	else ampNum += 2;
      }
      // if iIn3 < iIn4
      else {
	if (iIn3 > iIn5) ampNum += 3;
	else if (iIn4 > iIn5) ampNum += 1;
      }
    }
    // if (iIn2 < iIn1)
    else { 
      ampNum += (iIn2 - 1) * 6;
      if (iIn3 > iIn4) {
	if (iIn3 > iIn5) {
	  ampNum += 4;
	  if (iIn4 > iIn5) ampNum += 1;
	}
	// if iIn3 > iIn4 and iIn3 < iIn5
	else ampNum += 2;
      }
      // if iIn3 < iIn4
      else {
	if (iIn3 > iIn5) ampNum += 3;
	else if (iIn4 > iIn5) ampNum += 1;
      }
    }
    return ampNum;
  }
  
  // Decide whether a configuration is MHV
  // -2  : unpolarised
  // -1  : unphysical
  //  0  : MHV
  //  1  : NMHV (for future code)
  //  2  : NNMHV (for future code)
  //  3  : > NNMHV
  int MHVamplitudes::helAmpType(vector<int> hIn) {
    // sanity
    if (pos1Sav == -1 || pos2Sav == -1) {
      printOut("Vincia::MHVamplitudes::helAmpType","Initial partons not set");
      return -1;
    }
    int helSum = 0;
    for (int i = 0; i < (int)hIn.size(); i++) {
      if (abs(hIn[i]) == 1) 
	helSum += ( (i==pos1Sav || i==pos2Sav) ? -hIn[i] : hIn[i] );
      else if (hIn[i] == 9) return -2;
      else {
	cout << " helicities must be +-1, or 9. Please check. " << endl;
	return -1;
      }
    }
    if (verbose>=4) cout << " helSum = " << helSum << endl;
    if      (abs(helSum) >= (int)hIn.size()-3) return -1;
    else if (abs(helSum) == (int)hIn.size()-4) return  0;
    else if (abs(helSum) == (int)hIn.size()-6) return  1;
    else if (abs(helSum) == (int)hIn.size()-8) return  2;
    else return 3;
  }

//Proposed function. Leaving it here in case it is to be used later
/*  int MHVamplitudes::getAmpID(vector<int> idsIn) {
    // All particles are in the outgoing formalism
    // Check amplitude type 
        nGlue        =    0;
        nQuark       =    0;
    int nElec        =    0;
    int nNeut        =    0;
    int nDiffQuarks  =    0;
    int lepCode      =    0;
    vector<int> quarkType;
    vector<int> lepType;
    for (int i = 0; i < (int)idsIn.size(); i++) {
      if (idsIn[i] == 21) nGlue += 1;
      else if (abs(idsIn[i]) < 6 && idsIn[i] != 0) {
	nQuark += 1;
	quarkType.push_back(idsIn[i]);
      }
      else if (abs(idsIn[i]) == 11 || abs(idsIn[i]) == 13 || abs(idsIn[i]) == 15) {
	nElec += 1;
	lepType.push_back(idsIn[i]);
      }
      else if (abs(idsIn[i]) == 12 || abs(idsIn[i]) == 14 || abs(idsIn[i]) == 16) {
	nNeut += 1;
	lepType.push_back(idsIn[i]);
      }
    }
    if ((int)quarkType.size() == 2) nDiffQuarks = 1;
    else if ((int)quarkType.size() == 4) {
      // Check if same flavour or diff flavour
      for (int i = 1; i < 3; i++) 
	if (abs(quarkType[i]) == quarkType[0]) nDiffQuarks += 1;
      // Put a flag for which colour line the gluon appears off
    }
    if (nElec == 2) lepCode = 1;
    if (nElec == 1 && nNeut == 1) {
      if (abs(lepType[0]) % 2 == 0 && lepType[0] > 0) lepCode = 2;
      else if (abs(lepType[0]) % 2 == 0 && lepType[0] < 0) lepCode = 3;
      else if (abs(lepType[0] + 1) % 2 == 0 && lepType[0] > 0) lepCode = 3;
      else if (abs(lepType[0] + 1) % 2 == 0 && lepType[0] < 0) lepCode = 2;
      else lepCode = -1;       
    }
    // NEED TO CREATE A VECTOR TO STORE THESE IDS
    return nGlue + 10*nDiffQuarks + 1000*lepCode;
  }*/

  // Get helicity-dependent part of MHV amplitude to polarise a process
  // Assuming that particles are already in colour-order/standard order for 
  // MHV amplitudes
  vector<double> MHVamplitudes::getMHVhelAmps(vector<Vec4> k, vector<int> ID, int ampType) {
    // Initialise values
    nPartons         = k.size();
    Error            = false;
    fermHelConserved = true;

    //Sanity check
    if (k.size() != ID.size()) {
      Error = true;
      if (verbose>=1) printOut("Vincia::MHVamplitudes::helInfoME2",
            "Inconsistancy with size of inputs.");
    }
    
    // Check particle number
    if (nPartons > nPartonsToPolMax || nPartons < 4) {
      Error = true;
      if (verbose>=1) 
	cout << " Vincia::MHVamplitudes::helInfoME2, Check number of particles."
	     << " Maximum number of particles is " << nPartonsToPolMax 
	     << " Minimum number of particles is 4" << endl;
    }
    
    if (!isInitPtr) {
      Error = true;
      printOut("Vincia::MHVamplitudes::helInfoME2",
	       "Pointers not set. Must initialise pointers");
    }
    
    if (!isInit) {
      Error = true;
      printOut("Vincia::MHVamplitudes::helInfoME2",
	       "Couplings not initialised");
    }
    
    if (pos1Sav == -1 || pos2Sav == -1) {
      Error = true;
      printOut("Vincia::MHVamplitudes::helInfoME2","Initial partons not set.");
    }

    // Create vectors to store momenta, id, hel, ME2, return vector
    vector<Vec4> p; p.clear(); p.resize(0);
    vector<int> ids; ids.clear(); ids.resize(0);
    vector<double> hel; hel.clear(); hel.resize(0);
    vector<double> helME2; helME2.clear(); helME2.resize(0);
    
    // Convert incoming to outgoing 
    k[pos1Sav]    = -k[pos1Sav];
    k[pos2Sav]    = -k[pos2Sav];
    if      (abs(ID[pos1Sav]) < 6 && ID[pos1Sav] != 0) 
      ID[pos1Sav] = -ID[pos1Sav];
    else if (abs(ID[pos1Sav]) <= 16 && abs(ID[pos1Sav]) >= 11) 
      ID[pos1Sav] = -ID[pos1Sav];
    if      (abs(ID[pos2Sav]) < 6 && ID[pos2Sav] != 0) 
      ID[pos2Sav] = -ID[pos2Sav];
    else if (abs(ID[pos2Sav]) <= 16 && abs(ID[pos2Sav]) >= 11) 
      ID[pos2Sav] = -ID[pos2Sav];
    
    if (Error) {
      helME2.push_back(-1.0);
      return helME2;
    }
    if (nPartons == 4) {
      if (ampType == 1) { // All-Gluon amplitudes
	// Store positions of final-state partons
	int final1 = -1, final2 = -1;
	for (int i=0; i<nPartons; i++) {
	  if (i != pos1Sav && i != pos2Sav) {
	    if (final1 == -1) final1 = i;
	    else final2 = i;
	  }
	}
	// (++++) (iIn1, iIn2, iOut1, iOut2) (uncrossed)
	helME2.push_back( norm( pow( spinorProdPos(k[pos1Sav], k[pos2Sav]), 4. ) ) );
	// (+-+-)
	helME2.push_back( norm( pow( spinorProdPos(k[pos1Sav], k[final2]), 4. ) ) );
	// (+--+)
	helME2.push_back( norm( pow( spinorProdPos(k[pos1Sav], k[final1]), 4. ) ) );
	// Use parity inversion for rest of helAmps
	for (int i=2; i>=0; i--) helME2.push_back(helME2[i]);
      }
      else if (ampType == 2) { // 2-quark plus gluons
	// (--++) (order = (q,g1,g2,qBar) all-outgoing)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[1]), 3.) * 
			      spinorProdPos(k[3], k[1]) ) );
	// (-+-+)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[2]), 3.) * 
			      spinorProdPos(k[3], k[2]) ) );
	// Use parity inversion for rest of helAmps
	for (int i=1; i>=0; i--) helME2.push_back(helME2[i]);
      }
      else if (ampType == 3) { //4-quark
	// (--++) (order = (q1,qBar2,q2,qBar1) all outgoing)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[1]), 2.) ) );
	// (-+-+)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[2]), 2.) ) );
	// Use parity inversion for rest of helAmps
	for (int i=1; i>=0; i--) helME2.push_back(helME2[i]);
      }
      else if (ampType == 4) { 
	// 2 quarks, 2 leptons (so far only combined photon/Z exchange, no W exchange)
	if (ID[2] == -ID[3]) {
	  double sLLBar = norm( spinorProdPos(k[0], k[1]) );
	  
	  // (-+-+) (order = (l,lBar,q,qBar) all outgoing)
	  complex helAmp = pow( spinorProdPos(k[0], k[2]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZll*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZll*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (-++-)
	  helAmp = pow( spinorProdPos(k[0], k[3]), 2. );
	  // Up-type quark
	  if (ID[2] %2 == 0) 
	    helAmp *= (gZll*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZll*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (+--+)
	  helAmp = pow( spinorProdPos(k[1], k[2]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZlr*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZlr*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (+-+-)
	  helAmp = pow( spinorProdPos(k[1], k[3]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZlr*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZlr*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );     
	}
	else helME2.push_back(-1.);
      }
    // End if (nPartons == 4) 
    }
    else if (nPartons == 5) {
      if (ampType == 1) { // All-Gluon amplitudes
	// Store positions of final-state partons
	int final1 = -1, final2 = -1, final3 = -1;
	for (int i=0; i<nPartons; i++) {
	  if (i != pos1Sav && i != pos2Sav) {
	    if (final1 == -1) final1 = i;
	    else if (final2 == -1) final2 = i;
	    else final3 = i;
	  }
	}
	// (+++++) (iIn1, iIn2, iOut1, iOut2) (uncrossed)
	helME2.push_back( norm( pow( spinorProdPos(k[pos1Sav], k[pos2Sav]), 4. ) ) );
	// (+-++-)
	helME2.push_back( norm( pow( spinorProdPos(k[pos1Sav], k[final3]), 4. ) ) );
	// (+-+-+)
	helME2.push_back( norm( pow( spinorProdPos(k[pos1Sav], k[final2]), 4. ) ) );
	// (+--++)
	helME2.push_back( norm( pow( spinorProdPos(k[pos1Sav], k[final1]), 4. ) ) );
	// (-+++-)
	helME2.push_back( norm( pow( spinorProdPos(k[pos2Sav], k[final3]), 4. ) ) );
	// (-++-+)
	helME2.push_back( norm( pow( spinorProdPos(k[pos2Sav], k[final2]), 4. ) ) );
	// (-+-++)
	helME2.push_back( norm( pow( spinorProdPos(k[pos2Sav], k[final1]), 4. ) ) );
	// Use parity inversion for rest of helAmps
	for (int i=6; i>=0; i--) helME2.push_back(helME2[i]);
      }
      else if (ampType == 2) {
	// (--+++) (order = (q,g1,g2,,g3,qBar) all-outgoing)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[1]), 3.) * 
			      spinorProdPos(k[4], k[1]) ) );
	// (-+-++)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[2]), 3.) * 
			      spinorProdPos(k[4], k[2]) ) );
	// (-++-+)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[3]), 3.) * 
			      spinorProdPos(k[4], k[3]) ) );
	// (+-++-)
	helME2.push_back( norm( pow( spinorProdPos(k[4], k[1]), 3.) *
			      spinorProdPos(k[0], k[1]) ) );
	// (++-+-)
	helME2.push_back( norm( pow( spinorProdPos(k[4], k[2]), 3.) *
			      spinorProdPos(k[0], k[2]) ) );
	// (+++--)
	helME2.push_back( norm( pow( spinorProdPos(k[4], k[3]), 3.) *
			      spinorProdPos(k[0], k[3]) ) );
	// Use parity inversion for rest of helAmps
	for (int i=5; i>=0; i--) helME2.push_back(helME2[i]);
      }
      else if (ampType == 3) {
	// (+-++-) (order = (q1,qBar2,q2,g,qBar1) all outgoing)
	helME2.push_back( norm( pow( spinorProdPos(k[4], k[1]), 2.) ) );
	// (++-+-)
	helME2.push_back( norm( pow( spinorProdPos(k[4], k[2]), 2.) ) );
	// (-+-++)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[2]), 2.) ) );
	// (--+++)
	helME2.push_back( norm( pow( spinorProdPos(k[0], k[1]), 2.) ) );
	// Use parity inversion for rest of helAmps
	for (int i=3; i>=0; i--) helME2.push_back(helME2[i]);
      }
      else if (ampType == 4) {
	// 2 quarks, 2 leptons + gluon 
	// (so far only combined photon/Z exchange, no W exchange)
	if (ID[2] == -ID[4]) {
	  double sLLBar = norm( spinorProdPos(k[0], k[1]) );
	  
	  // (+-++-) (order = (l,lBar,q,g,qBar) all outgoing)
	  complex helAmp = pow( spinorProdPos(k[1], k[4]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZlr*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZlr*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (+--++)
	  helAmp = pow( spinorProdPos(k[1], k[2]), 2. );
	  // Up-type quark
	  if (ID[2] %2 == 0) 
	    helAmp *= (gZlr*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZlr*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (-+-++)
	  helAmp = pow( spinorProdPos(k[0], k[2]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZll*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZll*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (-+++-)
	  helAmp = pow( spinorProdPos(k[0], k[4]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZll*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZll*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );     
	  
	  // (+---+) 
	  helAmp = pow( spinorProdNeg(k[0], k[4]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZlr*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZlr*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar); 
	  helME2.push_back( norm( helAmp ) );
	  
	  // (+-+--)
	  helAmp = pow( spinorProdNeg(k[0], k[2]), 2. );
	  // Up-type quark
	  if (ID[2] %2 == 0) 
	    helAmp *= (gZlr*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZlr*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (-++--)
	  helAmp = pow( spinorProdNeg(k[1], k[2]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZll*gZur/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZll*gZdr/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );
	  
	  // (-+--+)
	  helAmp = pow( spinorProdNeg(k[1], k[4]), 2. );
	  // Up-type quark
	  if (ID[2] % 2 == 0) 
	    helAmp *= (gZll*gZul/(sLLBar + propZ) + gAl*gAu/sLLBar);
	  // Down-type quark
	  else helAmp *= (gZll*gZdl/(sLLBar + propZ) + gAl*gAd/sLLBar);
	  helME2.push_back( norm( helAmp ) );     
	}
	else helME2.push_back(-1.);	
      }
      else helME2.push_back(-1.);
    // End if (nPartons == 5)  
    }
    
    return helME2;
    
  }


  // Order particles if not already ordered by Vincia::MEcorrections
  // Andrew Notes: return the order as vector of numbers (use a map?)
  vector<int> MHVamplitudes::orderParticles(vector<int> h, vector<int> ID){
    vector<int> MHVorder; MHVorder.clear(); MHVorder.resize(0);
    if (twoQuark) {
      q1    = -1;
      qBar1 = -1;
	
      // Find quark, put it first
      for (int i = 0; i < nPartons; i++) {
        if (abs(ID[i]) < 6 && ID[i] != 0) {
	  q1 = i;
	  MHVorder.push_back(i);
	  break;
        }
      }

      // Find gluons, put them in sequentially
      for (int i = 0; i < nPartons; i++)
	if (ID[i] == 21) MHVorder.push_back(i);
    
      // Find anti-quark, put it last
      for (int i = 0; i < nPartons; i++) {
        if (ID[i] == - ID[q1] && h[i] == -h[q1]) {
	  qBar1 = i;
	  MHVorder.push_back(i);
	  break;
        }
      }
        
      // If quark helicity error
      if (q1 == -1 || qBar1 == -1 ) {
	fermHelConserved = false;
	cout << " Check number of quarks/antiquarks and their helicities. ";
	cout << " Helicity must be conserved along a fermion line" << endl;
      }
    }
    
    else if (fourQuark) {
      if (nGlue == 0) {
	q1 = -1;
	q2 = -1;
	qBar1 = -1;
	qBar2 = -1;
	
	// Order partons for LC
	// Find 1st quark, put it first
	for (int i = 0; i < nPartons; i++) {
	  if (ID[i] > 0 && ID[i] < 6) {
	    q1 = i;
	    MHVorder.push_back(i);
	    break;
	  }
	}

	// Find the anti-quark corresponding to the first quark, 
	for (int i = 0; i < nPartons; i++) {
	  if (ID[i] == -ID[q1] && h[i] == -h[q1]) {
	    qBar1 = i;
	    break;
	  }
	}
	
	// Find 2nd quark
	for (int i = q1 + 1; i < nPartons; i++) {
	  if (ID[i] > 0 && ID[i] < 6) { 
	    q2 = i;
	    if (ID[q2] == ID[q1]) sameFlav = true;
	    break;
	  }
	}
		
	// Find the anti-quark corresponding to the second quark, 
	// Put it second
	for (int i = 0; i < nPartons; i++) {
	  if (i == q1 || i == q2 || i == qBar1) continue;
	  else if (ID[i] == -ID[q2] && h[i] == -h[q2]) {
	    qBar2 = i;	    
	    MHVorder.push_back(i);
	    break;
	  }
	}
	//Put other two particles in order q2, qBar1
	MHVorder.push_back(q2);
	MHVorder.push_back(qBar1);
      }
      else if (nGlue == 1) {
	q1 = -1;
	q2 = -1;
	qBar1 = -1;
	qBar2 = -1;
	
	// Order partons for LC
	// Find 1st quark, put it first
	for (int i = 0; i < nPartons; i++) {
	  if (ID[i] > 0 && ID[i] < 6) {
	    q1 = i;
	    MHVorder.push_back(i);
	    break;
	  }
	}

	// Find the anti-quark corresponding to the first quark, 
	for (int i = 0; i < nPartons; i++) {
	  if (ID[i] == -ID[q1] && h[i] == -h[q1]) {
	    qBar1 = i;
	    break;
	  }
	}      
     
	// Find 2nd quark
	for (int i = q1 + 1; i < nPartons; i++) {
	  if (ID[i] > 0 && ID[i] < 6) {
	    q2 = i;
	    if (ID[q2] == ID[q1]) sameFlav = true;
	    break;
	  }
	}
	
	// Find the anti-quark corresponding to the second quark,
	// Put it second
	for (int i = 0; i < nPartons; i++) {
	  if (i == q1 || i == q2 || i == qBar1) continue;
	  else if (ID[i] == -ID[q2] && h[i] == -h[q2]) {
	    qBar2 = i;	
	    MHVorder.push_back(i);
	    break;
	  }
	}
	
	//Order q1,qBar2,q2,g,qBar1
	MHVorder.push_back(q2);
	// Find gluon
	for (int i = 0; i < nPartons; i++) {
	  if (ID[i] == 21) {
	    MHVorder.push_back(i);
	    break;
	  }
	}
	MHVorder.push_back(qBar1);
      }
      if (q1 == -1 || q2 == -1 || qBar1 == -1 || qBar2 == -1) {
	fermHelConserved = false;
	if (verbose >= 1) printOut("Vincia::MHVamplitudes::orderParticles",
	  "Helicity must be conserved along a fermion line");
      }
    }
      
    else if (isWeak) {
      q1    = -1;
      qBar1 = -1;
      l1    = -1;
      lBar1 = -1;
	     
      // Find quark, antiquark and leptons, put lepton first
      for (int i = 0; i < nPartons; i++) {
        if (ID[i] > 0 && ID[i] < 6) 		  q1 = i;
        else if (ID[i] > -6 && ID[i] < 0)      qBar1 = i;
	else if (ID[i] <= 16 && ID[i] >= 11) {
	  l1 = i;
	  MHVorder.push_back(i);
	}
	else if (ID[i] >= -16 && ID[i] <= -11) lBar1 = i;
      }
      
      if (l1 == -1 || lBar1 == -1 || q1 == -1 || qBar1 == -1) {
	if (verbose>=1) printOut("Vincia::MHVamplitudes::orderParticles",
	  "Error in weak 2->n. Check fermion types");
	Error = true;
      }
      
      // Put anti-lepton second
      if (h[lBar1] == -h[l1]) MHVorder.push_back(lBar1);
      else {
	if (verbose>=1) printOut("Vincia::MHVamplitudes::orderParticles",
	  "Error in weak 2->n. Helicity must be conserved along a fermion line");
	fermHelConserved = false;
      }

      // Put quark next
      MHVorder.push_back(q1);

      // Find gluons, put them in sequentially
      if (nGlue != 0) {
	for (int i = 0; i < nPartons; i++)
	  if (ID[i] == 21) MHVorder.push_back(i);
      }

      // Put anti-quark after gluons 
      if (h[qBar1] == -h[q1]) MHVorder.push_back(qBar1);
      else {
	if (verbose>=1) printOut("Vincia::MHVamplitudes::orderParticles",
	  "Error in weak 2->n. Helicity must be conserved along a fermion line");
	fermHelConserved = false;
      }

    }
    
    return MHVorder;
  }

}
