// MatrixElementCorrections.cc is a part of the VINCIA plug-in to the 
// PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2018 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// NOTE Only works for Vincia:hyperjet = off (default)
//                     Vincia:massCorrectionSplitF < 2 (default)
//                     Vincia:massCorrectionEmitF = 0 (default)

#include "Vincia/MatrixElementCorrections.h"

namespace Vincia {

  //************************************************************************

  // Some helper functions

  // Extract particles from event
  vector<Particle> convertEventToParticles(Event& event, int MEclass,
    map<int,int>& convMap) {
    // Clear map for convert particle numbers.
    convMap.clear();
    int counter = 0;
    vector<Particle> parts;
    // For pp processes extract beam daughters.
    if ( (MEclass == pureQCD) || (MEclass == singleBosonProd) ) {
      int iBeam[2] = { 1, 2 };
      int iBeamDaughter[2] = { event[iBeam[0]].daughter1(),
                               event[iBeam[1]].daughter1() };
      if ( (iBeamDaughter[0] == 0) || (iBeamDaughter[1] == 0) ) 
        cout << " iBeamDaughters = 0" << endl;
      else {
        parts.push_back(event[iBeamDaughter[0]]);
        parts.push_back(event[iBeamDaughter[1]]);
        convMap[iBeamDaughter[0]] = counter;
        counter++;
        convMap[iBeamDaughter[1]] = counter;
        counter++;
      }
    }
    // For single boson processes extract last boson.
    if ( (MEclass == singleBosonProd) || (MEclass == singleBosonDecay) ) {
      int iBosonLast = -1;
      for (int i=0; i<(int)event.size(); i++) 
        if ( (event[i].idAbs() > 22) && (event[i].idAbs() < 26) )
          iBosonLast = i;
      if (iBosonLast == -1) cout << " iBosonLast = -1" << endl;
      else {
        parts.push_back(event[iBosonLast]);
        convMap[iBosonLast] = counter;
        counter++;
      }
    }
    // At last include all final state partons.
    for (int i=0; i<(int)event.size(); i++) if ( event[i].isFinal() &&
      ( (event[i].idAbs() == 21) || (event[i].idAbs() < 6) ) ) {
      parts.push_back(event[i]);
      convMap[i] = counter;
      counter++;
    }
    // Done.
    return parts;
  }
  vector<Particle> convertEventToParticles(Event& event, int MEclass) {
    map<int,int> convMap;
    return convertEventToParticles(event, MEclass, convMap);
  }

  // Functions to count specific partons
  int countNrQuarks(vector<Particle> parts, int emi1 = -1, int emi2 = -1) {
    int countQ = 0;
    for (int i=0; i<(int)parts.size(); i++) {
      if ( (i != emi1) && (i != emi2) ) if (parts[i].idAbs()<6) countQ++;
    }
    return countQ;
  }
  int countNrGluons(vector<Particle> parts) {
    int countG = 0;
    for (int i=0; i<(int)parts.size(); i++)
      if (parts[i].id()==21) countG++;
    return countG;
  }
  int countNrPartons(vector<Particle> parts) {
    int countP = 0;
    for (int i=0; i<(int)parts.size(); i++)
      if ( (parts[i].idAbs()<6) || (parts[i].id()==21) ) countP++;
    return countP;
  }
  int countNrLeptons(vector<Particle> parts) {
    int countL = 0;
    for (int i=0; i<(int)parts.size(); i++)
      if (abs(parts[i].id())>=11 && abs(parts[i].id())<=16) countL++;
    return countL;
  }

  // Extract colour chains from input
  vector< vector<int> > getColChains(vector<Particle> parts) {
    vector< vector<int> > cChains;
    bool wrongColors = false;
    for (int i=0; i<(int)parts.size(); i++) {
      int col = parts[i].col(), acol = parts[i].acol();
      int id = parts[i].id();
      if (abs(id) < 6) {
             if ( (id > 0) && ( col == 0) ) wrongColors = true;
        else if ( (id < 0) && (acol == 0) ) wrongColors = true;
      }
      else if ( (id == 21) && (col == 0 || acol == 0) ) wrongColors = true;
    }
    if (wrongColors) {
      cout << "wrong colours in getColChains " << endl;
      for (int i=0; i<(int)parts.size(); i++)
        cout <<" id"<<i<<(i>9 ? " = " : "  = ")<<num2str(parts[i].id(),7)
             <<", col = "<<num2str(parts[i].col(),3)<<", acol = "
             <<num2str(parts[i].acol(),3)<<endl;
      return cChains;
    }
    vector<int> iQs;
    // Colour Map.
    map<int,int> indexOfAcol, indexOfCol;
    for (int i=0; i<(int)parts.size(); i++) {
      int  col  = 0;
      int  acol = 0;
      bool isIS = !parts[i].isFinal();
      if (isIS) { col = parts[i].acol(); acol = parts[i].col();  }
      else      { col = parts[i].col();  acol = parts[i].acol(); }
      if (col  != 0) indexOfCol[col]   = i;
      if (acol != 0) indexOfAcol[acol] = i;
      int id = parts[i].id();
      if (abs(id) < 6) {
             if ( (id > 0) && !isIS ) iQs.push_back(i);
        else if ( (id < 0) &&  isIS ) iQs.push_back(i);
      }
    }
    // Gluons only, pick random one to start with.
    if (iQs.size()==0) {
      int iG1 = -1;
      int col = 0;
      for (int i=0; i<(int)parts.size(); i++) if (parts[i].id() == 21) {
        iG1 = i;
        col = (parts[iG1].isFinal() ? parts[iG1].col() : parts[iG1].acol());
        break;
      }
      vector<int> chain; chain.clear(); chain.resize(0);
      chain.push_back(iG1);
      bool done  = false;
      int iGnext = -1;      
      while (!done) {
        iGnext = indexOfAcol[col];
        // Found the end (gluon we started with).
        if (iGnext == iG1) done = true;
        else {
          chain.push_back(iGnext);
          // Update colour to look for.
          col = (parts[iGnext].isFinal() ? parts[iGnext].col() : 
            parts[iGnext].acol());
        }
      }
      cChains.push_back(chain);
    }
    // Loop over all quarks.
    else {
      for (int iQ=0; iQ<(int)iQs.size(); iQ++) {
        vector<int> chain; chain.clear(); chain.resize(0);
        chain.push_back(iQs[iQ]);
        int col = (parts[iQs[iQ]].isFinal() ? parts[iQs[iQ]].col() : 
          parts[iQs[iQ]].acol());
        bool done = false;
        int iPnext = -1;
        while (!done) {
          iPnext = indexOfAcol[col];
          chain.push_back(iPnext);
          int id = parts[iPnext].id();
          // Check if end of chain found (qbar) and update colour to look for.
          if (parts[iPnext].isFinal()) {
            if ( (id < 0) && (abs(id) < 6) ) done = true;
            col = parts[iPnext].col();
          } else {
            if ( (id > 0) && (abs(id) < 6) ) done = true;
            col = parts[iPnext].acol();
          }
        }
        cChains.push_back(chain);
      }
    }
    return cChains;
  }

  // Function to check if a q qbar pair can be clustered
  bool canClustQQB(vector<Particle> parts, vector< vector<int> > cChains,
                   int iJ, int iPQ, int nQLmin) {
    // Only forming chains if # quark pairs >= 2.
    if (cChains.size() <= 0) return true;
    if ((int)cChains.size() <= nQLmin) return false;
    // Flavour sum to check if Ws are involved.
    int flavourSum = 0;
    int icChainQ1  = 0;
    int jcChainQ1  = 0;
    int icChainQ2  = 0;
    int jcChainQ2  = 0;
    // 1) Are the two quarks in the same colour chain?
    for (int i=0; i<(int)cChains.size(); i++) {
      bool foundQ1 = false;
      bool foundQ2 = false;
      for (int j=0; j<(int)cChains[i].size(); j++) {
        if (cChains[i][j] == iJ) {
          foundQ1   = true;
          icChainQ1 = i;
          jcChainQ1 = j;
        } else if (cChains[i][j] == iPQ) {
          foundQ2   = true;
          icChainQ2 = i;
          jcChainQ2 = j;
        }
        if (parts[cChains[i][j]].id() != 21) {
          if (parts[cChains[i][j]].isFinal()) 
            flavourSum += parts[cChains[i][j]].id();
          else flavourSum -= parts[cChains[i][j]].id();
        }
      }
      if (foundQ1 && foundQ2) return false;
    }
    // 2) For >= 3 colour chains: Has merged chain and remaining chain(s) each
    // the same two but different flavours (e.g. u - g - ubar and d - dbar)?
    if (cChains.size() >= 3) {
      // List of flavours.
      map<int,int> flavours;
      for (int ifl=1; ifl<=5; ++ifl) flavours[ifl] = 0;
      // Flavours that form the colour chain, from merged one.
      // Do they have the same flavour?
      int flavEndQ1 = parts[ cChains[icChainQ1][ 
        cChains[icChainQ1].size()-1-jcChainQ1 ] ].idAbs();
      int flavEndQ2 = parts[ cChains[icChainQ2][ 
        cChains[icChainQ2].size()-1-jcChainQ2 ] ].idAbs();
      bool mergedSameCol = (flavEndQ1 == flavEndQ2);
      flavours[flavEndQ1]++;
      flavours[flavEndQ2]++;
      vector<bool> remainSameCol;
      remainSameCol.resize(0);
      // Check remaining colour chains.
      for (int i=0; i<(int)cChains.size(); i++) {
        if (i != icChainQ1 && i != icChainQ2) {
          int flavEnd1 = parts[cChains[i][0]].idAbs();
          int flavEnd2 = parts[cChains[i][cChains[i].size()-1]].idAbs();
          remainSameCol.push_back(flavEnd1 == flavEnd2);
          flavours[flavEnd1]++;
          flavours[flavEnd2]++;
        }
      }
      // Do all remaining chains have the same flavour?
      bool allRemainSameCol = true;
      for (int i=0; i<(int)remainSameCol.size(); i++)
        if (!remainSameCol[i]) allRemainSameCol = false;
      // How many different flavours do we have?
      int nrDiffFlavours = 0;
      for (int ifl=1; ifl<=5; ++ifl) if (flavours[ifl] > 0) nrDiffFlavours++;
      // Merged two colour chains -> cChains.size()-1 left over.
      // We only have colour chains that each have the same quark
      // flavour and all with different flavours -> disconnected.
      if ((nrDiffFlavours == (int)cChains.size()-1) && mergedSameCol && 
        allRemainSameCol) return false;
      // This can happen for W events.
      else if (nrDiffFlavours == (int)cChains.size()) return false;
    }
    return true;
  }

  //************************************************************************

  // MEC history helper class

  MEChist::MEChist( int depth,
                    MatrixElementCorrections* MECptr,
                    TimeAntShower* timeShowerPtr,
                    SpaceAntShower* spaceShowerPtr,
                    Event& stateIn,
                    map<string, double> stateVarsIn,
                    double probIn,
                    bool orderedIn,
                    int nQLmin,
                    clustCombi clusterIn,
                    MEChist* motherIn = 0 ) {
    stateSav     = stateIn;
    stateVarsSav = stateVarsIn;
    scaleSav     = stateVarsIn["t"];
    probSav      = probIn;
    orderedSav   = orderedIn;
    clusterSav   = clusterIn;
    motherSav    = motherIn;
    childrenSav.clear();
    childrenSav.resize(0);
    map<int,int> convMap;
    vector<Particle> particleState = 
      convertEventToParticles(stateIn,MECptr->getProcClass(),convMap);
    MEvalSav      = max(0.0,MECptr->evalME(particleState));
    MECcontribSav = 0.0;
    asEffNsav     = 0.0;
    if (depth > 0) {
      // Perform clusterings on this state.
      // Colour Map.
      map<int,int> indexOfAcol, indexOfCol;
      for (int i=0; i<(int)stateIn.size(); i++) {
        int col = stateIn[i].col(), acol = stateIn[i].acol();
        if (!stateIn[i].isFinal()) { int tmp = col; col = acol; acol = tmp; }
        // Only allow final state partons and beam daughters.
        if (stateIn[i].isFinal() || (stateIn[i].mother1() == 1) || 
          (stateIn[i].mother1() == 2) ) {
          if ( col!=0) indexOfCol[col]   = i;
          if (acol!=0) indexOfAcol[acol] = i;
        }
      }
      // Colour Chains.
      int nQuarks = countNrQuarks(particleState);
      vector< vector<int> > colChains;
      if (nQuarks >= 4) colChains = getColChains(particleState);
      // Loop over all final state partons and find branchings.
      for (int i=0; i<(int)stateIn.size(); i++) if (stateIn[i].isFinal()) {
        int iEmi    = i;
        int colEmi  = stateIn[iEmi].col();
        int acolEmi = stateIn[iEmi].acol();
        int idEmi   = stateIn[iEmi].id();
        // 1) Emission:
        if (idEmi == 21) {
          int iRad = indexOfCol[acolEmi];
          int iRec = indexOfAcol[colEmi];
          vector<string> splitNames;
          vector<double> splitProbs;
          vector< map<string, double> > stateVars;
          vector<Event> clustEvts;
          if (timeShowerPtr->isTimelike(stateIn,iRad,iEmi,iRec,"")) {
            splitNames = timeShowerPtr->getSplittingName(
              stateIn,iRad,iEmi,iRec);
            splitProbs = timeShowerPtr->getSplittingProbAll(
              stateIn,iRad,iEmi,iRec,splitNames[0]);
            stateVars  = timeShowerPtr->getStateVariablesAll(
              stateIn,iRad,iEmi,iRec,splitNames[0]);
            clustEvts  = timeShowerPtr->clusteredAll(
              stateIn,iRad,iEmi,iRec,splitNames[0]);
          } else {
            splitNames = spaceShowerPtr->getSplittingName(
              stateIn,iRad,iEmi,iRec);
            splitProbs = spaceShowerPtr->getSplittingProbAll(
              stateIn,iRad,iEmi,iRec,splitNames[0]);
            stateVars  = spaceShowerPtr->getStateVariablesAll(
              stateIn,iRad,iEmi,iRec,splitNames[0]);
            clustEvts  = spaceShowerPtr->clusteredAll(
              stateIn,iRad,iEmi,iRec,splitNames[0]);
          }
          for (int cl=0; cl<(int)splitNames.size(); cl++) {
            if (clustEvts[cl].size() <= 3) continue;
            double splitScale = stateVars[cl]["t"];
            bool   orderedNow = ( !orderedSav ? false :
              (scaleSav < (1.0+Q2DIFFMAX)*splitScale) );
            childrenSav.push_back( new MEChist(depth-1,MECptr,timeShowerPtr, 
              spaceShowerPtr,clustEvts[cl],stateVars[cl],splitProbs[cl],
              orderedNow,nQLmin,clustCombi(iRad,iEmi,iRec),this) );
          }
        }
        // 2) Splitting/Conversion:
        else if (abs(idEmi) < 6) {
          // Find the colour partner of the emission.
          int iColP = ((colEmi>0) ? indexOfAcol[colEmi] : indexOfCol[acolEmi]);
          // Initial values for radiator and recoiler.
          vector<int> iRad, iRec;
          vector<bool> isTimelike;
          // 2.1) Initial gluon splitting:
          if ( (stateIn[iColP].id() == 21) && (!stateIn[iColP].isFinal()) ) {
            // Gluon = radiator, colour partner of gluon = recoiler.
            iRad.push_back(iColP);
            iRec.push_back( (colEmi > 0) ? indexOfAcol[stateIn[iColP].acol()] :
              indexOfCol[stateIn[iColP].col()] );
            isTimelike.push_back(false);
          } 
          // Will cluster two quarks now, is that ok?
          if ((nQuarks-2) >= 2*nQLmin) {
            // Possible radiators:
            for (int j=0; j<(int)stateIn.size(); j++) {
              // 2.2) Final gluon splitting:
              // FS quarks with -idEmi and no colour connection.
              if ( stateIn[j].isFinal() && (idEmi == -stateIn[j].id()) ) {
                int colRad = stateIn[j].col(), acolRad = stateIn[j].acol();
                int colMaxEmi = max(colEmi, acolEmi);
                int colMaxRad = max(colRad, acolRad);
                if ( canClustQQB(particleState,colChains,convMap[iEmi],
                  convMap[j],nQLmin) && (colMaxEmi != colMaxRad) ) {
                  iRad.push_back(j);
                  iRec.push_back(iColP);
                  isTimelike.push_back(stateIn[iColP].isFinal());
                }
              }
              // 2.3) Initial gluon conversion:
              // IS quarks (beam daughters) with idEmi and no colour connection.
              else if ( (!stateIn[j].isFinal()) && (idEmi == stateIn[j].id())
                && (stateIn[j].mother1() <= 2) ) {
                int colRad = stateIn[j].col(), acolRad = stateIn[j].acol();
                int colMaxEmi = max(colEmi, acolEmi);
                int colMaxRad = max(colRad, acolRad);
                if ( canClustQQB(particleState,colChains,convMap[iEmi],
                  convMap[j],nQLmin) && (colMaxEmi != colMaxRad) ) {
                  // 2.3.1) FS quark is emission:
                  // colour partner of FS quark is recoiler.
                  iRad.push_back(j);
                  iRec.push_back(iColP);
                  isTimelike.push_back(false);
                  // 2.3.2) IS quark is emission:
                  // colour partner of IS quark is recoiler.
                  int iRecNow = ( (colRad > 0) ? indexOfCol[colRad] : 
                    indexOfAcol[acolRad] );
                  iRad.push_back(j);
                  iRec.push_back(iRecNow);
                  isTimelike.push_back(false);
                }
              }
            }
          }
          // Now loop over the found iRad-iRec combinations.
          for (int j=0; j<(int)iRad.size(); j++) {
            int iRadNow = iRad[j], iRecNow = iRec[j];
            bool isTimelikeNow = isTimelike[j];
            vector<string> splitNames;
            vector<double> splitProbs;
            vector< map<string, double> > stateVars;
            vector<Event> clustEvts;
            if (isTimelikeNow) {
              splitNames = timeShowerPtr->getSplittingName(
                stateIn,iRadNow,iEmi,iRecNow);
              splitProbs = timeShowerPtr->getSplittingProbAll(
                stateIn,iRadNow,iEmi,iRecNow,splitNames[0]);
              stateVars  = timeShowerPtr->getStateVariablesAll(
                stateIn,iRadNow,iEmi,iRecNow,splitNames[0]);
              clustEvts  = timeShowerPtr->clusteredAll(
                stateIn,iRadNow,iEmi,iRecNow,splitNames[0]);
            } else {
              splitNames = spaceShowerPtr->getSplittingName(
                stateIn,iRadNow,iEmi,iRecNow);
              splitProbs = spaceShowerPtr->getSplittingProbAll(
                stateIn,iRadNow,iEmi,iRecNow,splitNames[0]);
              stateVars  = spaceShowerPtr->getStateVariablesAll(
                stateIn,iRadNow,iEmi,iRecNow,splitNames[0]);
              clustEvts  = spaceShowerPtr->clusteredAll(
                stateIn,iRadNow,iEmi,iRecNow,splitNames[0]);
            }
            for (int cl=0; cl<(int)splitNames.size(); cl++) {
              if (clustEvts[cl].size() <= 3) continue;
              double splitScale = stateVars[cl]["t"];
              bool   orderedNow = ( !orderedSav ? false :
                (scaleSav < (1.0+Q2DIFFMAX)*splitScale) );
              childrenSav.push_back( new MEChist(depth-1,MECptr,timeShowerPtr,
                spaceShowerPtr,clustEvts[cl],stateVars[cl],splitProbs[cl],
                orderedNow,nQLmin,clustCombi(iRadNow,iEmi,iRecNow),this) );
            }
          }
        }
      }
    }
  }

  // Helper functions with histories as input.
  bool hasChildren( vector<MEChist*> histories ) {
    for (int iHist=0; iHist<(int)histories.size(); iHist++)
      if (histories[iHist]->childrenSav.size() > 0) return true;
    return false;
  }
  vector<MEChist*> getLeaves( MEChist* history ) {
    vector<MEChist*> leaves = history->childrenSav;
    // First order -> leaves are empty
    if (leaves.size() == 0) {
      leaves.push_back(history);
      return leaves;
    }
    bool hasChildrenNow = hasChildren(leaves);
    while (hasChildrenNow) {
      vector<MEChist*> leavesNow;
      for (int iLeaf=0; iLeaf<(int)leaves.size(); iLeaf++) {
        int nChildren = (int)leaves[iLeaf]->childrenSav.size();
        if (nChildren > 0) {
          for (int iChild=0; iChild<nChildren; iChild++)
            leavesNow.push_back(leaves[iLeaf]->childrenSav[iChild]);
        } else leavesNow.push_back(leaves[iLeaf]);
      }
      leaves         = leavesNow;
      hasChildrenNow = hasChildren(leaves);
    }
    return leaves;
  }
  vector< vector<MEChist*> > getLayers( MEChist* history, int depth ) {
    vector< vector<MEChist*> > histories;
    // Layer 0 = mother itself.
    if (depth >= 0) {
      vector<MEChist*> mother;
      mother.push_back(history);
      histories.push_back(mother);
    }
    // Layer 1 = children.
    if (depth >= 1) {
      histories.push_back(history->childrenSav);
    }
    // Layer >=2 = children of children of ...
    int depthNow = depth-1;
    vector<MEChist*> historiesNow = history->childrenSav;
    while (depthNow > 0) {
      // Fill new vector with children to get next layer.
      vector<MEChist*> nextLayer;
      for (int i=0; i<(int)historiesNow.size(); i++) {
        vector<MEChist*> children = historiesNow[i]->childrenSav;
        for (int j=0; j<(int)children.size(); j++)
          nextLayer.push_back(children[j]);
      }
      // Save this layer.
      histories.push_back(nextLayer);
      // Updated histories to investigate next.
      historiesNow = nextLayer;
      depthNow--;
    }
    return histories;
  }

  //************************************************************************

  // MatrixElementCorrections class

  // Initialization
  void MatrixElementCorrections::init() {
    verbose            = settingsPtr->mode("Vincia:verbose");

    // Matching settings
    matchingLO         = settingsPtr->mode("Vincia:matchingLO");
    meCorrections      = (matchingLO >= 1);
    matchingFullColor  = settingsPtr->flag("Vincia:matchingFullColor");
    helicityShower     = settingsPtr->flag("Vincia:helicityShower");
    useMHV             = settingsPtr->flag("Vincia:useMHVamplitudes");
    // Without helicity shower, switch MHV off
    if (!helicityShower) useMHV = false;

    // Information about merging
    doMerging          = (!(settingsPtr->word("Merging:Process").compare("void")
      == 0));
    showerPathsOnly    = settingsPtr->flag("Merging:showerPathsOnly");
    nonShowerPathsOnly = settingsPtr->flag("Merging:nonShowerPathsOnly");
    nJetMax            = settingsPtr->mode("Merging:nJetMax");
    muFSave            = settingsPtr->parm("Merging:muFac");
    unorderedScaleVariation =
      settingsPtr->parm("Merging:unorderedScaleVariation");
    incompleteScalePrescrip =
      settingsPtr->mode("Merging:incompleteScalePrescrip");

    // Set initial values
    procClass          = 0;
    nQLmin             = 0;
    helAmpUsed         = -1;
    isMHV              = false;

    // Initialise MG4 interface
    bool isMG4on = mg4interfacePtr->init();
    mg4interfacePtr->setColorDepth(matchingFullColor);
    if (meCorrections && !isMG4on && verbose >= 1) 
      printOut("Vincia::MatrixElementCorrections", 
        "Couldn't switch MG4 interface on.");

    // Initialise MG5 interface
    bool isMG5on = mg5interfacePtr->init();
    mg5interfacePtr->setColorDepth(matchingFullColor);
    if (meCorrections && !isMG5on && verbose >= 1) 
      printOut("Vincia::MatrixElementCorrections", 
        "Couldn't switch MG5 interface on.");

    // Initialise MHV interface
    if (useMHV) {
      useMHV = mhvPtr->init();
      if (!useMHV && verbose >= 1) printOut("Vincia::MatrixElementCorrections", 
        "Couldn't switch MHVamplitudes on.");
      mhvPtr->setColorDepth(matchingFullColor);
    }

    isInit             = true;
  }

  // Can we do the process
  bool MatrixElementCorrections::canDoMEC(vector<Particle> parts, 
    bool hardProcess) {
    string printText = (hardProcess ? " in hard process!" : "!");
    if (procClass < procClassMin || procClass > procClassMax) {
      if (verbose >= 4)
        printOut("Vincia::MatrixElementCorrections::canDoMEC",
          "Requested MEC for unknown process class" + printText );
      return false;
    }
    int nParticles = (int)parts.size();
    if (nParticles > nMax[procClass]) {
      if (verbose >= 4)
        printOut("Vincia::MatrixElementCorrections::canDoMEC",
          "Requested MEC order too large" + printText );
      return false;
    } else if (nParticles == 8 && procClass == 3 && !useMHV) {
      if (verbose >= 4)
        printOut("Vincia::MatrixElementCorrections::canDoMEC",
          "Requested MEC order too large" + printText );
      return false;
    }
    // Check single boson production / decay
    if (procClass == singleBosonDecay || procClass == singleBosonProd) {
      // We can do Z=23, W=24, H=25
      bool foundBoson = false;
      for (int i=0; i<nParticles; i++) {
        if (parts[i].idAbs()>22 && parts[i].idAbs()<26) {
          // Make sure it's only a single boson
          if (foundBoson) return false;
          foundBoson = true;
        }
      }
      return foundBoson;
    }
    // Check QCD
    if (procClass == pureQCD) {
      bool foundNonQCD = false;
      for (int i=0; i<nParticles; i++)
        if (parts[i].idAbs()>6 && parts[i].idAbs()!=21) foundNonQCD = true;
      if (!useMHV) return (!foundNonQCD);
      else {
        int helType = helAmpType(parts);
        if (helType == 0) isMHV = true;
        else if (helType == -1) {
          if (verbose >= 4)
            printOut("Vincia::MatrixElementCorrections::canDoMEC", 
              "Helicity configuration is unphysical");
          return false;
        } else if (nParticles == 8 && helType != 0) {
          if (verbose >= 4)
            printOut("Vincia::MatrixElementCorrections::canDoMEC", 
              "Requested MEC order too large for non-MHV");
          return false;
        }
        else isMHV = false;
        return (!foundNonQCD);
      }
    }
    // Check EW 2->n
    if (procClass == EWtwoToN) {
      bool foundNonExpected = false;
      for (int i=0; i<nParticles; i++) {
        if (parts[i].idAbs()>6 && parts[i].idAbs()<11) foundNonExpected = true;
        else if (parts[i].idAbs()>16 && parts[i].idAbs()!=21)
          foundNonExpected = true;
      }
      if (!useMHV) return (!foundNonExpected);
      else {
        int helType = helAmpType(parts);
        if (helType == 0) isMHV = true;
        else if (helType == -1) {
          if (verbose >= 4)
            printOut("Vincia::MatrixElementCorrections::canDoMEC",
              "Helicity configuration is unphysical");
          return false;
        }
        else isMHV = false;
        return (!foundNonExpected);
      }
    }
    // All good!
    return true;
  }

  // What type of amplitude are we calculating
  int MatrixElementCorrections::helAmpType(vector<Particle> parts) {
    int helSum = 0;
    for (int i = 0; i < (int)parts.size(); i++) {
      if (abs(parts[i].pol()) == 1) 
        helSum += ( (parts[i].isFinal()) ? parts[i].pol() : -parts[i].pol() );
      else if (parts[i].pol() == 9) return -2;
      else {
        cout << " helicities must be +-1, or 9. Please check. " << endl;
        return -1;
      }
    }
    if      (abs(helSum) >= (int)parts.size()-3) return -1;
    else if (abs(helSum) == (int)parts.size()-4) return  0;
    else if (abs(helSum) == (int)parts.size()-6) return  1;
    else if (abs(helSum) == (int)parts.size()-8) return  2;
    else return 3;
  }

  // Polarize particles
  bool MatrixElementCorrections::polarize(int nIn, vector<Particle>& parts) {
    // Reset helAmpUsed
    helAmpUsed = -1;
    // Check if we have the relevant matrix element
    if (!canDoMEC(parts,true)) {
      if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::polarize",
        "failed to polarize hard process. Unknown matrix element.");
      return false;      
    }

    // Make sure that particles are in colour order
    vector< vector<int> > order = orderForME(parts);
    if ((int)order[0].size() == 0) {
      if (verbose>=4) printOut("Vincia::MatrixElementCorrections::polarize",
        "Failed to polarise. Problem colour-ordering partons."); 
      return false;
    }

    // Vectors to copy particle information
    vector<int> idPartons; idPartons.clear(); idPartons.resize(0);
    vector<Vec4> pPartons; pPartons.clear(); pPartons.resize(0);
    vector<int> polPartons; polPartons.clear(); polPartons.resize(0);
    // Vectors to store all non-zero helicity amplitudes, and their helicities
    vector<double> helAmps; helAmps.clear(); helAmps.resize(0);
    vector<double> helAmps1; helAmps1.clear(); helAmps1.resize(0);
    vector< vector<int> > pols; pols.clear(); pols.resize(0);

    bool isMassive = false;

    // Resonance decays. 
    if (nIn == 1) {
      // Fill IDs and momenta in order (Boson, coloured-order)
      int iBoson = -1;
      for (int i=0; i<(int)parts.size(); i++) 
        if ( (parts[i].idAbs() > 21) && (parts[i].idAbs() < 38) ) iBoson = i;
      if (iBoson == -1) {
        if (verbose>=4) printOut("Vincia::MatrixElementCorrections::polarise",
          "No boson found in resonance decay. Not polarising partons");
        return false;
      }
      idPartons.push_back(parts[iBoson].id());
      pPartons.push_back(parts[iBoson].p());
      polPartons.push_back(9);
      for (int i=0; i<(int)order[0].size(); ++i) {
        idPartons.push_back(parts[order[0][i]].id());
        pPartons.push_back(parts[order[0][i]].p());
        polPartons.push_back(9);
        // State should already have been mapped to massless at this
        // point (if massless) so only check for numerical rounding
        // errors to see if massive here or not.
        if (i >= nIn && pow2(pPartons[i].mCalc()/pPartons[i].e()) > TINY) 
          isMassive = true; 
      }
      if (verbose >= 6) printOut("Vincia::MatrixElementCorrections::polarize",
        "isMassive = "+bool2str(isMassive));
      // Always allow decaying resonance to be polarized
      polPartons[0] = parts[iBoson].pol();
      // Get the polarization-summed ME
      mg4interfacePtr->setColorDepth(1);
      int npols = (int)pow(2.,(int)idPartons.size() - nIn);
      for (int i = 0; i < npols; i++) {
        // Calculate all helAmps by removing all but decaying resonance from 
        // polPartons, then looping through all helicity structures 
        polPartons.resize(1);
        for (int j = 1; j <= (int)idPartons.size() - nIn; ++j) {    
          if (i % (int)pow(2.,j) < (int)pow(2.,j-1)) polPartons.push_back(-1.);
          else polPartons.push_back(1.);
        }
        helAmps.push_back(mg4interfacePtr->ME2(1, idPartons, pPartons,
          polPartons));
        pols.push_back(polPartons);
        if (helAmps[i] < 0.) {
          if (verbose>=4) printOut("Vincia::MatrixElementCorrections::polarize",
            "helAmps returned negative ME2!!! returning false");
          return false;
        }
      }
      // Return colorDepth to setting requested for MECs
      mg4interfacePtr->setColorDepth(matchingFullColor);
      if (verbose>=6) {
        printOut("Vincia::MatrixElementCorrections::polarize",
          "Check calculation of all helAmps");
        for (int i=0; i < npols; i++) {
          cout << " helAmps["<<i<<"] = "<<scientific<< helAmps[i]
               <<"   pols are: ";
          for (int j=0; j<(int)polPartons.size(); j++) cout<<"   " <<pols[i][j];
          cout << endl;
        }
      }
      double sumAll = 0;
      for (int i=0; i<(int)helAmps.size(); i++) sumAll += helAmps[i];
      // Generate random number
      double sumR   = rndmPtr->flat() * sumAll;
      // Sum helAmps[i] until sumNow >= sumR
      double sumNow = 0;
      for (int i=0; i<(int)helAmps.size(); i++) {
        sumNow += helAmps[i];
        if (sumNow >= sumR) {
          helAmpUsed = i;
          break;
        }
      }
      if (helAmpUsed == -1) {
        if (verbose>=4) printOut("Vincia::MatrixElementCorrections::polarize",
          "sumNow never chose a polarization");
        return false;
      }
      if (verbose >= 5) {
        printOut("Vincia::MatrixElementCorrections::polarize",
          "Chose FS polarizations = ");
        for (int j=1; j<= int(idPartons.size()) - 1; ++j) {
          cout << pols[helAmpUsed][j] << " ";
        }
        cout << endl;
      }
      // Save the chosen polarisation (including map back to input order)
      for (int iIn=0; iIn<(int)parts.size(); iIn++) {
        for (int iOrd=0; iOrd<(int)order[0].size()+1; iOrd++) {
          if (pPartons[iOrd].pz() == parts[iIn].pz()) {
            parts[iIn].pol(pols[helAmpUsed][iOrd]);
            break;
          }
        }
      }
      // All is well
      return true;
    }

    if (nIn == 2) {
      for (int j=0; j<(int)order[0].size(); j++) {
        idPartons.push_back(parts[order[0][j]].id());
        pPartons.push_back(parts[order[0][j]].p());
        // Double check masslessness
        if (pow2(pPartons[j].mCalc()/pPartons[j].e()) > TINY) {
          isMassive = true;
          isMHV = false;
        }
      }
      if (isMHV) {
        if ((int)order[0].size()!=(int)parts.size()) {
          if (verbose>=4) printOut("Vincia::MatrixElementCorrections::polarize",
            "Failed to polarize. Problem ordering partons."); 
          return false;
        } 
        // What type of process is this?
        int nQuarks  = countNrQuarks(parts);
        int nGluons  = countNrGluons(parts);
        int nLeptons = countNrLeptons(parts);

        if ((int)parts.size() == 4) { 
          if (nGluons == 4) {
            //        helAmps[0] = (++++) (order = (in1,in2,out1,out2))
            //        helAmps[1] = (+-+-)
            //        helAmps[2] = (+--+)
            //        helAmps[3] = (-++-)
            //        helAmps[4] = (-+-+)
            //        helAmps[5] = (----)
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 1);
            int final1 = -1, final2 = -1;
            for (int i=0; i<(int)order[0].size(); i++) {
              // helAmps[0]
              polPartons.push_back(1);
              // Store positions of final-state partons
              if (i != order[1][0] && i != order[1][1]) {
                if (final1 == -1) final1 = i;
                else final2 = i;
              }
            }
            pols.push_back(polPartons);
            // helAmps[1]
            polPartons[order[1][1]] = -1;
            polPartons[ final2 ]    = -1;
            pols.push_back(polPartons);
            // helAmps[2]
            polPartons[ final1 ]    = -1;
            polPartons[ final2 ]    =  1;
            pols.push_back(polPartons);
            // End if (all-glue)  
          } else if (nQuarks == 2 && nGluons == 2) {
            //        helAmps[0] = (--++) (order = (q,g1,g2,qBar) all-outgoing)
            //        helAmps[1] = (-+-+)
            //        helAmps[2] = (+-+-)
            //        helAmps[3] = (++--)
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 2);
            polPartons.clear(); polPartons.resize(4);
            // helAmps[0]
            polPartons[0] = ( parts[order[0][0]].isFinal() ? -1 :  1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ? -1 :  1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            // helAmps[1]
            polPartons[1] = ( parts[order[0][1]].isFinal() ?  1 : -1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            // End if (2 quark QCD)    
          } else if (nQuarks == 4) {
            //        helAmps[0] = (--++) (order = (q1,qBar2,q1,qBar1) all-outgoing)
            //        helAmps[1] = (-+-+)
            //        helAmps[2] = (+-+-)
            //        helAmps[3] = (++--)
            bool sameFlav = false;
            if (parts[order[0][0]].id() == parts[order[0][2]].id()) 
              sameFlav = true;
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            if (!sameFlav)
              helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 3);
            polPartons.clear(); polPartons.resize(4);
            // helAmps[0]
            polPartons[0] = ( parts[order[0][0]].isFinal() ? -1 :  1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ? -1 :  1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            if (sameFlav) {
              mhvPtr->setColorDepth(1);
              mhvPtr->calcSpinorProds(pPartons, polPartons, idPartons, true);
              helAmps.push_back(mhvPtr->ME2());
            }
            // helAmps[1]
            polPartons[1] = ( parts[order[0][1]].isFinal() ?  1 : -1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            if (sameFlav) {
              mhvPtr->calcSpinorProds(pPartons, polPartons, idPartons, true);
              helAmps.push_back(mhvPtr->ME2());
              // Fill rest of helAmps, return colorDepth to that set by user
              for (int i=1; i>=0; i--) helAmps.push_back(helAmps[i]);
              mhvPtr->setColorDepth(matchingFullColor);
            }
            // End if (4 quark)
          } else if (nQuarks == 2 && nLeptons == 2) {
            //        helAmps[0] = (-+-+) (order = (l,lBar,q,qBar) all-outgoing)
            //        helAmps[1] = (-++-)
            //        helAmps[2] = (+--+)
            //        helAmps[3] = (+-+-)
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 4);
            polPartons.clear(); polPartons.resize(4);
            // helAmps[0]
            polPartons[0] = ( parts[order[0][0]].isFinal() ? -1 :  1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ?  1 : -1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ? -1 :  1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            // helAmps[1]
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            // End if (2 quark 2 leptons QCD)    
          }
          // End if (nPartons == 4)
        }

        else if ((int)parts.size() == 5) { 
          if (nGluons == 5) {
            //        helAmp[0]  = (+++++) (order = (in1,in2,out1,out2))
            //        helAmp[1]  = (+-++-)
            //        helAmp[2]  = (+-+-+)
            //        helAmp[3]  = (+--++)
            //        helAmp[4]  = (-+++-)
            //        helAmp[5]  = (-++-+)
            //        helAmp[6]  = (-+-++)
            //        helAmp[7]  = (+-+--)
            //        helAmp[8]  = (+--+-)
            //        helAmp[9]  = (+---+)
            //        helAmp[10] = (-++--)
            //        helAmp[11] = (-+-+-)
            //        helAmp[12] = (-+--+)
            //        helAmp[13] = (-----)
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 1);
            int final1 = -1, final2 = -1, final3 = -1;
            for (int i=0; i<(int)order[0].size(); i++) {
              polPartons.push_back(1);
              // Store positions of final-state partons
              if (i != order[1][0] && i != order[1][1]) {
                if (final1 == -1) final1 = i;
                else if (final2 == -1) final2 = i;
                else final3 = i;
              }
            }
            pols.push_back(polPartons); 
            //helAmps[1]
            polPartons[order[1][1]] = -1;
            polPartons[ final3 ]    = -1;
            pols.push_back(polPartons);
            //helAmps[2]
            polPartons[ final2 ]    = -1;
            polPartons[ final3 ]    =  1;
            pols.push_back(polPartons);
            //helAmps[3]
            polPartons[ final1 ]    = -1;
            polPartons[ final2 ]    =  1;
            pols.push_back(polPartons);
            //helAmps[4]
            polPartons[order[1][0]] = -1;
            polPartons[order[1][1]] =  1;
            polPartons[ final1 ]    =  1;
            polPartons[ final3 ]    = -1;
            pols.push_back(polPartons);
            //helAmps[5]
            polPartons[ final2 ]    = -1;
            polPartons[ final3 ]    =  1;
            pols.push_back(polPartons);
            //helAmps[6]
            polPartons[ final1 ]    = -1;
            polPartons[ final2 ]    =  1;
            pols.push_back(polPartons);
            // End if (all-glue)  
          } else if (nQuarks == 2 && nGluons == 3) {
            //        helAmps[0]  = (--+++) (order = (q,g1,g2,,g3,qBar) all-outgoing)
            //        helAmps[1]  = (-+-++)
            //        helAmps[2]  = (-++-+)
            //        helAmps[3]  = (+-++-)
            //        helAmps[4]  = (++-+-)
            //        helAmps[5]  = (+++--)
            //        helAmps[6]  = (---++)
            //        helAmps[7]  = (--+-+)
            //        helAmps[8]  = (-+--+)
            //        helAmps[9]  = (+--+-)
            //        helAmps[10] = (+-+--)
            //        helAmps[11] = (++---)
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 2);
            polPartons.clear(); polPartons.resize(5);
            // helAmps[0]
            polPartons[0] = ( parts[order[0][0]].isFinal() ? -1 :  1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ? -1 :  1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ?  1 : -1 );
            polPartons[4] = ( parts[order[0][4]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            // helAmps[1]
            polPartons[1] = ( parts[order[0][1]].isFinal() ?  1 : -1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            // helAmps[2]
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            // helAmps[3]
            polPartons[0] = ( parts[order[0][0]].isFinal() ?  1 : -1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ? -1 :  1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ?  1 : -1 );
            polPartons[4] = ( parts[order[0][4]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            // helAmps[4]
            polPartons[1] = ( parts[order[0][1]].isFinal() ?  1 : -1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            // helAmps[5]
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
          } else if (nQuarks == 4) {
            //        helAmps[0] = (+-++-) (order = (q1,qBar2,q2,g,qBar1) all-outgoing)
            //        helAmps[1] = (++-+-)
            //        helAmps[2] = (-+-++)
            //        helAmps[3] = (--+++)
            //        helAmps[4] = (++---)
            //        helAmps[5] = (+-+--)
            //        helAmps[6] = (--+-+)
            //        helAmps[7] = (-+--+)
            bool sameFlav = false;
            if (parts[order[0][0]].id() == parts[order[0][2]].id()) 
              sameFlav = true;
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            if (!sameFlav) 
              helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 3);
            polPartons.clear(); polPartons.resize(5);
            // helAmps[0]
            polPartons[0] = ( parts[order[0][0]].isFinal() ?  1 : -1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ? -1 :  1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ?  1 : -1 );
            polPartons[4] = ( parts[order[0][4]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            if (sameFlav) {
              mhvPtr->setColorDepth(1);
              mhvPtr->calcSpinorProds(pPartons, polPartons, idPartons, true);
              helAmps.push_back(mhvPtr->ME2());
            }
            // helAmps[1]
            polPartons[1] = ( parts[order[0][1]].isFinal() ?  1 : -1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            if (sameFlav) {
              mhvPtr->calcSpinorProds(pPartons, polPartons, idPartons, true);
              helAmps.push_back(mhvPtr->ME2());
            }
            // helAmps[2]
            polPartons[0] = ( parts[order[0][0]].isFinal() ? -1 :  1 );
            polPartons[4] = ( parts[order[0][4]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            if (sameFlav) {
              mhvPtr->calcSpinorProds(pPartons, polPartons, idPartons, true);
              helAmps.push_back(mhvPtr->ME2());
            }
            // helAmps[3]
            polPartons[1] = ( parts[order[0][1]].isFinal() ? -1 :  1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            if (sameFlav) {
              mhvPtr->calcSpinorProds(pPartons, polPartons, idPartons, true);
              helAmps.push_back(mhvPtr->ME2());
              // Fill rest of helAmps, return colorDepth to that set by user
              for (int i=3; i>=0; i--) helAmps.push_back(helAmps[i]);
              mhvPtr->setColorDepth(matchingFullColor);
            }
          } else if (nQuarks == 2 && nLeptons == 2) {
            //        helAmps[0] = (+-++-) (order = (l,lBar,q,g,qBar) all-outgoing)
            //        helAmps[1] = (+--++)
            //        helAmps[2] = (-+-++)
            //        helAmps[3] = (-+++-)
            //        helAmps[4] = (+---+)
            //        helAmps[5] = (+-+--)
            //        helAmps[6] = (-++--)
            //        helAmps[7] = (-+--+)
            mhvPtr->setIIpos( order[1][0], order[1][1] );
            helAmps = mhvPtr->getMHVhelAmps(pPartons, idPartons, 4);
            polPartons.clear(); polPartons.resize(5);
            // helAmps[0]
            polPartons[0] = ( parts[order[0][0]].isFinal() ?  1 : -1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ? -1 :  1 );
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[3] = ( parts[order[0][3]].isFinal() ?  1 : -1 );
            polPartons[4] = ( parts[order[0][4]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
            // helAmps[1]
            polPartons[2] = ( parts[order[0][2]].isFinal() ? -1 :  1 );
            polPartons[4] = ( parts[order[0][4]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            // helAmps[2]
            polPartons[0] = ( parts[order[0][0]].isFinal() ? -1 :  1 );
            polPartons[1] = ( parts[order[0][1]].isFinal() ?  1 : -1 );
            pols.push_back(polPartons);
            // helAmps[3]
            polPartons[2] = ( parts[order[0][2]].isFinal() ?  1 : -1 );
            polPartons[4] = ( parts[order[0][4]].isFinal() ? -1 :  1 );
            pols.push_back(polPartons);
          }
          // End if (nPartons == 5) 
        }

        // Check that MHV amplitudes were calculated correctly
        if (helAmps[0] < 0.) {
          if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::"
            "polarize", "Failed to polarise partons, MHVamplitudes::"
            "getMHVhelAmps returned -1");
          return false;
        }
        // Fill rest of polarisations with flipped helicities
        pols.resize((int)helAmps.size());
        int polNum = helAmps.size()/2 - 1;
        for (int i=(int)helAmps.size()/2; i<(int)pols.size(); i++)  {
          pols[i].resize(parts.size());
          for (int j=0; j<(int)pPartons.size(); j++)
            pols[i][j] = -pols[polNum][j];
          polNum--;
        }

        // Calculate the polarisation-summed ME
        double sumAll = 0;
        if (nLeptons == 0) {
          for (int i=0; i<(int)helAmps.size()/2; i++) sumAll += helAmps[i];
          sumAll *= 2;
        }
        // Weak processes aren't symmetric under parity flip
        else for (int i=0; i<(int)helAmps.size(); i++) sumAll += helAmps[i];
//         double sumAll1 = 0;
//         if (nLeptons == 0) {
//           for (int i=0; i<(int)helAmps1.size()/2; i++) sumAll1 += helAmps1[i];
//           sumAll1 *= 2;
//         }
//         else for (int i=0; i<(int)helAmps1.size(); i++) sumAll1 += helAmps1[i];

        if (verbose>=4) {
          printOut("Vincia::MatrixElementCorrections::polarize",
            "All helicity amplitudes (as fractions of the unpolarized "
            "amplitude) and polarisations are: ");
          for (int i=0; i<(int)helAmps.size(); i++) {
            cout << " helAmps[" << i << "]/sumAll = " << setw(5)
                 << setprecision(5) << helAmps[i]/sumAll << endl;
            for (int j=0;j<(int)idPartons.size();j++)
              cout << " " << setw(2) << idPartons[j] << " " << setw(2) 
                   << pols[i][j] << " " << pPartons[j];
          }
//           printOut("Vincia::MatrixElementCorrections::polarize",
//             "Same as above for using full MHV:");
//           for (int i=0; i<(int)helAmps1.size(); i++) {
//             cout << " helAmps1[" << i << "]/sumAll1 = " << setw(5)
//                  << setprecision(5) << helAmps1[i]/sumAll1 << endl;
//             for (int j=0;j<(int)idPartons.size();j++)
//               cout << " " << setw(2) << idPartons[j] << " " << setw(2) 
//                    << pols[i][j] << " " << pPartons[j];
//            }
        }

        // Generate a random number
        double sumR = rndmPtr->flat() * sumAll;

        // Sum over helAmps until sumNow > sumR
        double sumNow = 0;
        for (int i=0; i<(int)helAmps.size();i++) {
          sumNow += helAmps[i];
          if (sumNow > sumR) {
            helAmpUsed = i;
            break;
          }
        }
        if (helAmpUsed == -1) {
          if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::"
            "polarize", "sumNow never chose a polarization");
          return false;
        }

        if (verbose >= 4) {
          printOut("Vincia::MatrixElementCorrections::polarize",
            "Chosen polarizations: ");
          cout << " helAmpUsed = " << helAmpUsed << ", colorDepth = " 
               << mhvPtr->getColorDepth() << endl;
          for (int i=0;i<(int)idPartons.size();i++)
            cout << " " << setw(2) << idPartons[i] << " " << setw(2) 
                 << pols[helAmpUsed][i] << " " << pPartons[i];
        }
        // Save the chosen polarisation (including map back to input order)
        for (int iIn=0; iIn<(int)parts.size(); iIn++) {
          for (int iOrd=0; iOrd<(int)order[0].size(); iOrd++) {
            if (pPartons[iOrd].pz() == parts[iIn].pz()) {
              parts[iIn].pol(pols[helAmpUsed][iOrd]);
              break;
            }
          }
        }
        return true;
        // End if (isMHV)
      }

      else {
        // Fill IDs and momenta in order (coloured-order, boson)
        int iBoson = -1;
        for (int i=0; i<(int)parts.size(); i++) 
          if ( (parts[i].idAbs() > 21) && (parts[i].idAbs() < 38) ) iBoson = i;
        if (iBoson == -1 && order[0].size() != parts.size()) {
          if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::"
            "polarise", "Problem ordering partons for non-MHV. Not polarising "
            "partons");
          return false;
        }
        if (iBoson != -1) {
          idPartons.push_back(parts[iBoson].id());
          pPartons.push_back(parts[iBoson].p());
        }
        if (verbose >= 6) printOut("Vincia::MatrixElementCorrections::"
          "polarize", "isMassive = "+bool2str(isMassive));

        // Get the polarization-summed ME
        mg4interfacePtr->setColorDepth(1);
        mg4interfacePtr->setIIpos(order[1][0], order[1][1]);
        int npols = (int)pow(2.,(int)order[0].size());
//         cout << " npols = " << npols << " order.size() = " << order.size()
//              << endl;

        for (int i = 0; i < npols; i++) {
          // Calculate all helAmps by removing all but decaying resonance from 
          // polPartons, then looping through all helicity structures 
          polPartons.resize(0);
          if (iBoson != -1) {
            for (int j = 1; j < (int)idPartons.size(); ++j) {       
              if (i % (int)pow(2.,j) < (int)pow(2.,j-1)) 
                polPartons.push_back(-1.);
              else polPartons.push_back(1.);
            }
            // Allow gauge boson to be polarised
            polPartons.push_back(parts[iBoson].pol());
          } else {
            for (int j = 1; j < (int)idPartons.size() + 1; ++j) {   
              if (i % (int)pow(2.,j) < (int)pow(2.,j-1))
                polPartons.push_back(-1.);
              else polPartons.push_back(1.);
            }
          }
          pols.push_back(polPartons);
          helAmps.push_back( mg4interfacePtr->ME2(2, idPartons, pPartons,
            polPartons) );
          if (helAmps[i] < 0.) {
            if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::"
              "polarize", "helAmps returned negative ME2! returning false");
            return false;
          }
        }
        // Return colorDepth to setting requested for MECs
        mg4interfacePtr->setColorDepth(matchingFullColor);

        if (verbose>=6) {
          printOut("Vincia::MatrixElementCorrections::polarize",
            "Check calculation of all helAmps");
          for (int i=0; i < npols; i++) {
            cout << " helAmps[" << i << "] = " << scientific << helAmps[i]
                 << endl /*<<" ids are: "*/;
            for (int j=0; j<(int)pPartons.size(); j++)
              cout << " " << setw(2) << idPartons[j] << " " << setw(2) 
                   << pols[i][j] << " " << pPartons[j];
          }
        }
        double sumAll = 0;
        for (int i=0; i<(int)helAmps.size(); i++) sumAll += helAmps[i];
        // Generate random number
        double sumR   = rndmPtr->flat() * sumAll;
        // Sum helAmps[i] until sumNow >= sumR
        double sumNow = 0;
        for (int i=0; i<(int)helAmps.size(); i++) {
          sumNow += helAmps[i];
          if (sumNow >= sumR) {
            helAmpUsed = i;
            break;
          }
        }
        if (helAmpUsed == -1) {
          if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::"
            "polarize", "sumNow never chose a polarization");
          return false;
        }
      
        if (verbose >= 5) {
          printOut("Vincia::MatrixElementCorrections::polarize",
            "Chose polarizations = ");
          for (int j=0; j< int(idPartons.size()); ++j)
            cout << pols[helAmpUsed][j] << " ";
          cout << endl;
        }

        // Save the chosen polarisation (including map back to input order)
        for (int iIn=0; iIn<(int)parts.size(); iIn++) {
          for (int iOrd=0; iOrd<(int)order[0].size()+1; iOrd++) {
            if (pPartons[iOrd].pz() == parts[iIn].pz()) {
              parts[iIn].pol(pols[helAmpUsed][iOrd]);
              break;
            }
          }
        }
        // All is well
        return true;
      }
      // End if (nIn == 2 && nPartons == 4)
    }

    else if (nIn == 2 && (int)parts.size() == 5) {
      // End if (nIn == 2 && 5 partons) 
    }

    else {
      if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::polarize",
        "Not yet able to polarise this process");
      return false;
    }

    // if you got here something went wrong
    return false;
  }

  // Extract MEclass and clustering probability for input event.
  void MatrixElementCorrections::prepMECs(const Event& eventIn,
    int incomeType) {
    // TODO HintNow and nQLminNow settings!
    int nQLminNow = 1, HintNow = 2, MEclassNow = -1;
    if ((incomeType > 0) && (incomeType < 10)) MEclassNow = singleBosonDecay;
    if (incomeType <= 2) nQLminNow  = 1;
    else if (incomeType == 3) {
      nQLminNow  = 0;
      HintNow    = 2;
    } else if (incomeType == 10) {
      bool onlyQCD = true;
      // Skip system and beam.
      for (int i=2; i<(int)eventIn.size(); i++) {
        int id = eventIn[i].idAbs();
        if ( (id != 21) && (id > 6) ) onlyQCD = false;
        if ( (id > 22) && (id < 25) ) nQLminNow = 1;
      }
      nQLminNow  = (onlyQCD ? 0 : nQLminNow );
      MEclassNow = (onlyQCD ? pureQCD : singleBosonProd);
    }
    prepMECs(MEclassNow, nQLminNow, HintNow);
  }

  // Main Function: Construct all possible histories and 
  // return MEn+1 / sum ant*MEn (using the shower merging functions)
  double MatrixElementCorrections::performMECs(const Event& eventIn, 
    TimeAntShower* timeShowerPtr, SpaceAntShower* spaceShowerPtr, int nStep) {

    // Copy event.
    Event event = eventIn;

    // Safety checks.
    vector<Particle> parts = convertEventToParticles(event,procClass);
    int nParticles = (int)parts.size();
    if (nParticles > nMax[procClass]) {
      infoPtr->errorMsg("Error in MatrixElementCorrections::performMECs: "
        "Too many particles.");
      return 1.0;
    }
    if (nParticles < nMin[procClass]) {
      infoPtr->errorMsg("Error in MatrixElementCorrections::performMECs: "
        "Too few particles.");
      return 1.0;
    }
    if (!canDoMEC(parts, false)) {
      infoPtr->errorMsg("Error in MatrixElementCorrections::performMECs: "
        "Necessary matrix elements not present.");
      return 1.0;
    }

    // Setup the histories.
    int nClustStep = ( nStep > 0 ? nStep : nParticles-nMin[procClass]+1 );
    map<string, double> scales;
    scales["t"] = 0.0;
    MEChist* history = new MEChist(nClustStep, this, timeShowerPtr, 
      spaceShowerPtr, event, scales, 1.0, true, nQLmin, clustCombi(0,0,0));

    // Setup the different layers of the history tree.
    int depthNow = nClustStep-1;
    vector< vector<MEChist*> > layers = getLayers(history, depthNow);

    // If no ordered history is found (without checking factorization scale),
    // do not attempt to setup MEC calculation.
    bool foundOrderedPath = false;
    for (int i=0; i<(int)layers[depthNow].size(); i++) {
       vector<MEChist*> children = layers[depthNow][i]->childrenSav;
       for (int j=0; j<(int)children.size(); j++)
        if (children[j]->orderedSav) foundOrderedPath = true;
    }
    if (!foundOrderedPath) {
      delete history;
      return 1.0;
    }

    // Go through the layers of the tree. Start at one layer above the leaves.
    while (depthNow >= 0) {
      vector<MEChist*> layerNow = layers[depthNow];
      // Go through the histories of that layer.
      for (int i=0; i<(int)layerNow.size(); i++) {
        MEChist* state = layerNow[i];
        // ME and scale of current state.
        double stateME    = state->MEvalSav;
        double stateScale = state->scaleSav;
        // Denominator of MEC factor for the current state and mother.
        double stateDenom = 0.0, motherDenom = 0.0;
        // Build up the MEC factor by using the children.
        vector<MEChist*> children = state->childrenSav;
        for (int j=0; j<(int)children.size(); j++) {
          MEChist* child = children[j];
          // Probability: Use the contribution calculated in the previous step.
          double childProb  = child->probSav * child->MECcontribSav;
          double childScale = child->scaleSav;
          // Special case for MEC factor with only one clustering back.
          if (child->childrenSav.size() == 0) {
            // For Born the contribution is a*ME.
            childProb = child->probSav * child->MEvalSav;
            // Check shower starting scale = scale of Born state.
            double bornScale = 1.0e+15;
            if (doMerging) {
              bornScale = spaceShowerPtr->getBornScale(child->stateSav);
            } else {
              // Check if first emission was time- or spacelike.
              if (state->stateSav[child->clusterSav.iRadSav].isFinal() &&
                state->stateSav[child->clusterSav.iRecSav].isFinal())
                bornScale = timeShowerPtr->getShowerStartingScale(
                  child->stateSav);
              else bornScale = spaceShowerPtr->getShowerStartingScale(
                child->stateSav);
            }
            if (bornScale <= 0.0) bornScale = 1.0e+15;
            // If branching scale > Born scale set probability to zero.
            if ((1.0+Q2DIFFMAX)*childScale > bornScale) childProb = 0.0;
          }
          // Add a*ME (or a*MECcontrib if previous MECs included).
          stateDenom += childProb;
          // Contribution to MEC factor for mother only if ordered.
          if ((1.0+Q2DIFFMAX)*childScale > stateScale) motherDenom += childProb;
        }
        // No children: incomplete path, might be included with merging.
        if (children.size() <= 0) {
          // Set denominator to ME of current state for unity correction.
          stateDenom  = stateME;
          motherDenom = ( (doMerging && (!showerPathsOnly) &&
            (nJetMax >= nClustStep-depthNow)) ? incompleteContribution(state) :
            0.0 );
        }
        // MEC factor for current state. If stateDenom = 0, set MEC factor
        // to 1 (won't be used anyway as motherDenom = 0 as well).
        double MECfactor  = ((stateDenom==0.0) ? 1.0 : (stateME/stateDenom));
        // Contribution of current state to MEC factor of mother
        // (= full ME if all children contributed).
        double MECcontrib = motherDenom*MECfactor;
        // If we are on top of the tree, save the final result.
        if (depthNow == 0) MECcontrib = MECfactor;
        // Fully unordered events can be included with merging.
        else if ( doMerging && (!showerPathsOnly) && (stateDenom == 0.0) &&
          (nJetMax >= nClustStep-depthNow) ) {
          MECcontrib = unorderedContribution( state, nClustStep-depthNow,
              timeShowerPtr, spaceShowerPtr );
        }
        // Let this history know about the contribution.
        state->MECcontribSav = MECcontrib;
      }
      depthNow--;
    }

    // MEC contribution factor of mother history has been set to final result.
    double MEPS = history->MECcontribSav;

    // Delete histories.
    delete history;

    // Done.
    return MEPS;
 
  }

  // Return the contribution from an unordered event included with merging.
  double MatrixElementCorrections::unorderedContribution(MEChist* history,
    int nJet, TimeAntShower* timeShowerPtr, SpaceAntShower* spaceShowerPtr) {
    // We have a fully unordered + >= 1 event, included with merging.
    if (nJet == 1) {
      // For +1 take care of ordering with respect to factorization scale.
      // Check if ordered emission off this event contributes.
      // Take into account that multiple Borns might exist and only some lead
      // to current state with ordered emission.
      double num = 0.0, den = 0.0;
      for (int i=0; i<(int)history->childrenSav.size(); i++) {
        MEChist* child = history->childrenSav[i];
        den += child->probSav * child->MEvalSav;
        if ((1.0+Q2DIFFMAX)*child->scaleSav > history->scaleSav)
          num += child->probSav * child->MEvalSav;
      }
      return history->MEvalSav/den*num;
    } else if (nJet == 2) {
      // Special case: +2 event.
      map<string,double> effectiveScales = getUnorderedInfos(
        history->stateSav, timeShowerPtr, spaceShowerPtr );
      double teff  = effectiveScales["scaleEff"];
             teff *= pow2(unorderedScaleVariation);
      // Check if ordered emission off this event contributes.
      if ((1.0+Q2DIFFMAX)*teff > history->scaleSav) return history->MEvalSav;
    } else {
      // Due to how the ordering is restored and overlap removed, we only have
      // to check if ordering can be restored with an effective scale in the
      // generation below the current one. Therefore, we only have to check
      // the children.
      double num = 0.0, den = 0.0;
      for (int i=0; i<(int)history->childrenSav.size(); i++) {
        MEChist* child    = history->childrenSav[i];
        double childScale = child->scaleSav;
        map<string,double> effectiveScales = getUnorderedInfos(
          child->stateSav, timeShowerPtr, spaceShowerPtr );
        double teff  = effectiveScales["scaleEff"];
               teff *= pow2(unorderedScaleVariation);
        if ((1.0+Q2DIFFMAX)*teff > childScale) {
          // NOTE num and den are only approximations, but good enough
          den += child->probSav * child->MEvalSav;
          // If this restores ordering, check if ordered emission off this 
          // event contributes.
          if ((1.0+Q2DIFFMAX)*childScale > history->scaleSav)
            num += child->probSav * child->MEvalSav;
        }
      }
      if (den > 0.0) return history->MEvalSav/den*num;
      // If not able to restore ordering calculate effective scale of the
      // state and check if ordered emission off this event contributes.
      map<string,double> effectiveScales = getUnorderedInfos(
        history->stateSav, timeShowerPtr, spaceShowerPtr );
        double teff  = effectiveScales["scaleEff"];
               teff *= pow2(unorderedScaleVariation);
      if ((1.0+Q2DIFFMAX)*teff > history->scaleSav) return history->MEvalSav;
    }
    // Done. No contribution found.
    return 0.0;
  }

  // Return the contribution from an incomplete path included with merging.
  double MatrixElementCorrections::incompleteContribution(MEChist* history) {
    // Find starting scale.
    double tStart = pow2(muFSave);
    if (incompleteScalePrescrip == 1) {
      Vec4 pOut(0.0, 0.0, 0.0, 0.0);
      for(int k=0; k<(int)history->stateSav.size(); k++)
        if (history->stateSav[k].isFinal()) pOut += history->stateSav[k].p();
      tStart = pOut.m2Calc();
    } else if (incompleteScalePrescrip == 2) {
      tStart = pow2(history->stateSav[0].e());
    }
    // Contribution if ordered branching with respect to starting scale.
    if ((1.0+Q2DIFFMAX)*history->scaleSav > tStart) return history->MEvalSav;
    return 0.0;
  }

  // Return scales and values in case of only unordered shower paths.
  map<string, double> MatrixElementCorrections::getUnorderedInfos( 
    const Event& eventIn, TimeAntShower* timeShowerPtr, SpaceAntShower* 
    spaceShowerPtr) {

    // Copy event.
    Event event = eventIn;

    // Final result.
    map<string, double> unorderedInfos;

    // Safety checks.
    vector<Particle> parts = convertEventToParticles(event,procClass);
    int nParticles = (int)parts.size();
    if (nParticles > nMax[procClass]) {
      infoPtr->errorMsg("Error in MatrixElementCorrections::getUnorderedInfos: "
        "Too many particles.");
      return unorderedInfos;
    }
    if (nParticles < nMin[procClass]) {
      infoPtr->errorMsg("Error in MatrixElementCorrections::getUnorderedInfos: "
        "Too few particles.");
      return unorderedInfos;
    }
    if (!canDoMEC(parts, false)) {
      infoPtr->errorMsg("Error in MatrixElementCorrections::getUnorderedInfos: "
        "Necessary matrix elements not present.");
      return unorderedInfos;
    }

    // Setup the histories.
    int nClustStep = nParticles-nMin[procClass]+1;
    map<string, double> scales;
    scales["t"] = 0.0;
    MEChist* history = new MEChist(nClustStep, this, timeShowerPtr, 
      spaceShowerPtr, event, scales, 1.0, true, nQLmin, clustCombi(0,0,0));

    // Minimum and maximum of all scales appearing in the history tree.
    double tmin = 1.0e+15;
    double tmax = 0.0;

    bool foundTinyScale = false;
    double tinyScale    = pow2(0.9);

    // Setup the different layers of the history tree.
    int depthNow = nClustStep-1;
    vector< vector<MEChist*> > layers = getLayers(history, depthNow);

    // Go through the layers of the tree. Start at one layer above the leaves.
    while (depthNow >= 0) {
      vector<MEChist*> layerNow = layers[depthNow];
      // Go through the histories of that layer.
      for (int i=0; i<(int)layerNow.size(); i++) {
        // Numerator and denominator of alphaS^n+1.
        double numerator = 0.0, denominator = 0.0;
        // Go through the children = clusterings.
        vector<MEChist*> layerChildren = layerNow[i]->childrenSav;
        for (int j=0; j<(int)layerChildren.size(); j++) {
          MEChist* child = layerChildren[j];
          // Check for small scales.
          if (child->stateVarsSav["t"] < tinyScale) foundTinyScale = true;
          // ME, branching probability, and alphaS of the child.
          double probChild    = child->probSav;
          double MEchild      = child->MEvalSav;
          double asScaleChild = child->stateVarsSav["scaleASraw"];
          if (asScaleChild < tmin) tmin = asScaleChild;
          if (asScaleChild > tmax) tmax = asScaleChild;
          asScaleChild        = child->stateVarsSav["scaleAS"];
          double asChild      = vinComPtr->alphaS.alphaS(asScaleChild);
          // If clustering is not Born we need effective alphaS^n.
          double asEffN       = 1.0;
          if (child->childrenSav.size() > 0) asEffN = min( child->asEffNsav,
            pow(vinComPtr->alphaSmax,nClustStep-1-depthNow) );
          // Add contributions to alphaS^n+1.
          numerator   += (asChild*probChild*asEffN*MEchild);
          denominator += (probChild*MEchild);
        }
        // Save alphaS^n+1.
        layerNow[i]->asEffNsav = min( numerator/denominator,
           pow(vinComPtr->alphaSmax,nClustStep-depthNow) );
      }
      depthNow--;
    }

    // Final result, effective alphaS^nClustStep.
    double asEff = min( pow(history->asEffNsav, 1.0/nClustStep),
      vinComPtr->alphaSmax );

    // Effective scales: raw scale and effective alphaS scale.
    double scaleEffRaw;
    if (settingsPtr->mode("Vincia:alphaSorder") == 0) {
      // In case of constant alphaS return average.
      scaleEffRaw = pow2((sqrt(tmin)+sqrt(tmax))/2.0);
    } else if ( (tmax+vinComPtr->mu2freeze) < vinComPtr->mu2min ) {
      // If tmax+freeze smaller than minimum scale, use that scale.
      scaleEffRaw = (vinComPtr->mu2min-vinComPtr->mu2freeze);
    } else {
      // Use the Newton solver to get the effective scale.
      scaleEffRaw = vinComPtr->solveForT(tmin, tmax, asEff);
    }
    // For alphaS use freeze and minimum scale to return to merging.
    double scaleEff   = max( vinComPtr->mu2min,
                             vinComPtr->mu2freeze+scaleEffRaw);
    double asEffFound = vinComPtr->alphaS.alphaS(scaleEff);
    // Cross check.
    if (/*(!foundTinyScale) && */(abs(asEffFound-asEff) > TINY*asEff)) {
      infoPtr->errorMsg("Error in MatrixElementCorrections::getUnorderedInfos: "
        "Calculation of effective scale failed.");
      /*cout << endl << " alphaS(" << sqrt(scaleEff) << "^2) = " << asEffFound
           << " but should be alphaS(teff^2) = " << asEff << endl << " tmin = "
           << sqrt(tmin) << " and tmax = " << sqrt(tmax) << endl;*/
    }

    // Save result.
    unorderedInfos["scaleASeff"]  = scaleEff;
    unorderedInfos["asEff"]       = asEff;
    unorderedInfos["scalePDFeff"] = scaleEffRaw;
    unorderedInfos["scaleEff"]    = scaleEffRaw;

    unorderedInfos["tinyScale"]   = (foundTinyScale ? 1.0 : -1.0);

    // Delete histories.
    delete history;

    // Done.
    return unorderedInfos; 
  }

  // Get the order of the n partons for evaluating the matrix elements:
  // posII: position of incoming partons if present, smaller index first
  // order: colour chains (q - g - g - qbar etc, initial partons crossed
  //        into final state), chains ordered in size (smallest first)
  vector< vector<int> > MatrixElementCorrections::orderForME(
    vector<Particle> parts) {
    vector<int> order; order.clear(); order.resize(0);
    vector<int> posII; posII.clear(); posII.resize(0);

    bool wrongColors = false;
    for (int i=0; i<(int)parts.size(); i++) {
      int col = parts[i].col(), acol = parts[i].acol();
      int id = parts[i].id();
      if (abs(id) < 6) {
             if ( (id > 0) && ( col == 0) ) wrongColors = true;
        else if ( (id < 0) && (acol == 0) ) wrongColors = true;
      }
      else if ( (id == 21) && (col == 0 || acol == 0) ) wrongColors = true;
    }
    if (wrongColors) {
      cout << "wrong colours in orderForME " << endl;
      for (int i=0; i<(int)parts.size(); i++)
        cout <<" id"<<i<<(i>9 ? " = " : "  = ")<<num2str(parts[i].id(),7)
             <<", col = "<<num2str(parts[i].col(),3)<<", acol = "
             <<num2str(parts[i].acol(),3)<<endl;
      vector< vector<int> > combi; combi.clear(); combi.resize(0);
      combi.push_back(order); combi.push_back(posII);
      return combi;
    }

    int nQuarks  = countNrQuarks(parts);
    int nGluons  = countNrGluons(parts);
    int nLeptons = countNrLeptons(parts);

    // Check if MHVamplitudes class has required amplitude
    if (useMHV && procClass > singleBosonProd) {
      if (nQuarks == 0 && nGluons >= 4) isMHV = true;
      else if (nQuarks == 2) isMHV = true;
      else if (nQuarks == 4 && nGluons <= 1)  isMHV = true;
      else if (nQuarks == 2 && nLeptons == 2) isMHV = true;
      else isMHV = false;
    } else isMHV = false;
    if (isMHV) isMHV = (helAmpType(parts) == 0);

    //  Special case: gluons only, start with gi1
    if (nQuarks == 0) { 

      // 2 gluons only
      if (nGluons == 2) {
        for (int i=0; i<(int)parts.size(); i++)
          if (parts[i].id() == 21) order.push_back(i);
        posII.push_back(0);
        posII.push_back(1);
        vector< vector<int> > combi; combi.clear(); combi.resize(0);
        combi.push_back(order); combi.push_back(posII);
        return combi;
      }

      int iG1 = -1;
      // Find and save all colours and anticolours
      map<int,int> indexOfAcol; // Put initial colours here
      map<int,int> indexOfCol;  // Put initial anticolours here
      for (int i=0; i<(int)parts.size(); i++) {
        int col   = 0;
        int acol  = 0;
        bool isIS = !parts[i].isFinal();
        if (isIS) {
          col  = parts[i].acol();
          acol = parts[i].col();
        } else {
          col  = parts[i].col();
          acol = parts[i].acol();
        }
        if (col > 0) indexOfCol[col] = i;
        else if (col < 0) indexOfAcol[-col] = i;
        if (acol > 0) indexOfAcol[acol] = i;
        else if (acol < 0) indexOfCol[-acol] = i;
        if ( (parts[i].id() == 21) && (isIS) && (iG1 == -1) ) iG1 = i;
      }

      int col = 0;
      if (iG1 > -1) {
        // 1) Initial state gluons present
        col = parts[iG1].acol();
        posII.push_back(0);
      } else {
        // 2) No initial state gluons
        // Pick random gluon
        for (int i=0; i<(int)parts.size(); i++)  if (parts[i].id() == 21) {
          iG1 = i;
          col = parts[iG1].col();
          break;
        }
      }      

      order.push_back(iG1);
      bool done  = false;
      int iGnext = -1;
      int count  = 0;
      while (!done) {
        count++;
        iGnext = indexOfAcol[col];
        if (iGnext == iG1) done = true; // Found the end (gluon we started with)
        else {
          order.push_back(iGnext);
          if (!parts[iGnext].isFinal()) {
            posII.push_back(count);
            col = parts[iGnext].acol(); // Update colour to look for
          } else col = parts[iGnext].col();
        }
      }

      vector< vector<int> > combi; combi.clear(); combi.resize(0);
      combi.push_back(order); combi.push_back(posII);
      return combi;

    }

    vector< vector<int> > cChains;
    vector<int> iQs;
    vector<int> iLs;
    // Colour Map
    map<int,int> indexOfAcol;
    map<int,int> indexOfCol;
    for (int i=0; i<(int)parts.size(); i++) {
      int  col  = 0;
      int  acol = 0;
      bool isIS = !parts[i].isFinal();
      if (isIS) { col = parts[i].acol(); acol = parts[i].col();  }
      else      { col = parts[i].col();  acol = parts[i].acol(); }
      if (col  != 0) indexOfCol[col]   = i;
      if (acol != 0) indexOfAcol[acol] = i;
      int id = parts[i].id();
      if (abs(id) < 6) {
             if ( (id > 0) && !isIS ) iQs.push_back(i);
        else if ( (id < 0) &&  isIS ) iQs.push_back(i);
      }
      //TODO Might need to be more general. Currently assuming that if
      // leptons and quarks, it is OK for IS leptons to be first
      else if ( abs(id) > 10 && abs(id) < 17 ) iLs.push_back(i);
    }

    // Loop over all quarks
    for (int iQ=0; iQ<(int)iQs.size(); iQ++) {
      vector<int> chain; chain.clear(); chain.resize(0);
      chain.push_back(iQs[iQ]);
      int col    = (parts[iQs[iQ]].isFinal() ? parts[iQs[iQ]].col() : 
        parts[iQs[iQ]].acol());
      bool done  = false;
      int iPnext = -1;
      while (!done) {
        iPnext = indexOfAcol[col];
        chain.push_back(iPnext);
        int id = parts[iPnext].id();
        if (parts[iPnext].isFinal()) {
          // Found the end of the chain (antiquark)
          if ( (id < 0) &&  (abs(id) < 6) ) done = true;
          col = parts[iPnext].col(); // Update colour to look for
        } else {
          // Found the end of the chain (antiquark)
          if ( (id > 0) &&  (abs(id) < 6) ) done = true;
          col = parts[iPnext].acol(); // Update colour to look for
        }
      }
      cChains.push_back(chain);
    }

    // Sort vectors in size
    sort(cChains.begin(), cChains.end(), compVecSizeHelper);

    //TODO Currently implicitly assuming only 2 leptons maximum, both incoming. 
    // Need to generalise?
    int Lep = -1;
    for (int i=0; i<(int)iLs.size(); i++) {
      if ( parts[iLs[i]].id() < 0 ) {
        order.push_back(iLs[i]);
        if ( !parts[iLs[i]].isFinal() ) posII.push_back(0);
      } else if ( parts[iLs[i]].id() > 0 ) Lep = i;
    }
    if (Lep>=0) {
      order.push_back(iLs[Lep]);
      if ( !parts[iLs[Lep]].isFinal() ) posII.push_back(1);
    }
    // Loop over the chains and put nr in result
    int count = 0;
    if ((int)iLs.size()!=0) count = (int)iLs.size();
    for (int i=0; i<(int)cChains.size(); i++) {
      for (int j=0; j<(int)cChains[i].size(); j++) {
        order.push_back(cChains[i][j]);
        if (!parts[cChains[i][j]].isFinal()) posII.push_back(count);
        count++;
      }
    }

    vector< vector<int> > combi; combi.clear(); combi.resize(0);
    combi.push_back(order); combi.push_back(posII);
    return combi;
  }

  // Evaluate a matrix element for the defined process classes
  double MatrixElementCorrections::evalME(vector<Particle> parts) {
    // Safety check
    if (!canDoMEC(parts, false)) {
      if (verbose >= 4) printOut("Vincia::MatrixElementCorrections::evalME",
        "Requested ME that we don't have! Returning 0!");
      return 0.0;
    }

    // IDs, input momenta, and polarizations
    vector<int> MGids; MGids.clear(); MGids.resize(0);
    vector<Vec4> p; p.clear(); p.resize(0);
    vector<int> pol; pol.clear(); pol.resize(0);
    vector< vector<int> > order = orderForME(parts);
    if (order[0].size()==0) return -1.0;

    // Fill IDs and momenta, check if we have single boson
    int nIn      = ( decayME() ? 1 : 2 );
    int iBoson   = -1;
    if (decayME() || prodME()) {
      for (int i=0; i<(int)parts.size(); i++) 
        if ( (parts[i].idAbs() > 21) && (parts[i].idAbs() < 38) ) iBoson = i;
    }
    if ( decayME() && (iBoson != -1) ) {
      MGids.push_back(parts[iBoson].id());
      p.push_back(parts[iBoson].p());
      pol.push_back((int)parts[iBoson].pol());
    }
    for (int j=0; j<(int)order[0].size(); j++) {
      MGids.push_back(parts[order[0][j]].id());
      p.push_back(parts[order[0][j]].p());
      pol.push_back((int)parts[order[0][j]].pol());
    }
    if ( prodME() && (iBoson != -1) ) {
      MGids.push_back(parts[iBoson].id());
      p.push_back(parts[iBoson].p());
      pol.push_back((int)parts[iBoson].pol());
    }

    // Position of incoming guys
    if (twoToN()) {
      if (useMHV && isMHV) mhvPtr->setIIpos( order[1][0], order[1][1] );
      mg4interfacePtr->setIIpos( order[1][0], order[1][1] );
    }

    if (verbose >= 5) {
      printOut("Vincia::MatrixElementCorrections::evalME",
        "Printing ids, momenta, and polarizations:");
      for (int i=0; i<(int)MGids.size(); i++)
        cout << "  " << setw(2) << MGids[i] << " " << setw(2) << pol[i]
             << " " << p[i];
      if (twoToN()) 
        cout << "  ipos1 = " << order[1][0] << " and ipos2 = " 
             << order[1][1] << endl;
      cout << " colourDepth = " << ((useMHV && isMHV) ? mhvPtr->getColorDepth()
        : mg4interfacePtr->getColorDepth());
      cout << " isMHV = " << isMHV << " twoToN = " << twoToN();
    }

    if (useMHV && isMHV) {
      mhvPtr->calcSpinorProds(p, pol, MGids, true);
      return mhvPtr->ME2();
    }
    if (helicityShower) return mg4interfacePtr->ME2(nIn,MGids,p,pol);
    return mg4interfacePtr->ME2(nIn,MGids,p);
  }

}
